// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   23. Jul 1999
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: ARTE constructor for RecoSegment
//
// changes:
// 23 Jul 1999 creation by I. Siccama, C. Borgmeier
// 22 Nov 2000 (TG) added namespace CLUE
// 12 Jan 2001 (TG) removed obsolete include files
// 09 Jul 2001 (TG) renamed l to valid_
// 11 Jul 2001 (TG) removed Ecal member
// 13 Sep 2001 (TG) call sPinvToP() in convert()
//
// ********************************************************************

#include "clue/cluearte.hh"
#include "clue/RecoSegment.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  //==============================================================================
  // convert from RSEG or RTRA
  //==============================================================================
  template<class T>
  void RecoSegment::convert( T rtra ) {
    if( rtra == 0 ) {
      X = Y = Z = Tx = Ty = P = Chi2 = 0.;
      Charge = 0;
      for(unsigned int i = 0; i<15; ++i) Cov[i] = 0.;
    } else {
      
      X    = rtra->xf;
      Y    = rtra->yf;
      Z    = rtra->zf;
      Tx   = rtra->txf;
      Ty   = rtra->tyf;
      Chi2 = rtra->chi;


      if(rtra->pf == 0)
	Charge = 0;
      else
	if( rtra->pf < 0 ) 
	  Charge = -1;
	else
	  Charge = 1;
      
      P = Charge * 1./rtra->pf; // rtra->pf = Q/p => P is without charge
      
      for(unsigned int i=0; i<15; ++i) Cov[i] = rtra->cvf[i];

      // convert sigma(Q/p) to sigma(p) (needed by vertexing)
      sPinvToP();
    }
    
    return;
  }


  //==============================================================================
  // Constructors
  //==============================================================================
  RecoSegment::RecoSegment( const ArtePointer<RTRA>& rtra ) :
    pmatch (),
    X      (0),
    Y      (0),
    Z      (0),
    Zorig  (0),
    Tx     (0),
    Ty     (0),
    P      (0),
    Chi2   (0),
    Charge (0),
    Cmp    ( rtra->cmp ),
    Rtra   ( getIDX<RTRA>( rtra ) ),
    Nf     ( 0 ), 
    valid_ ( true ),
    COV_   (0),
    CINV_  (0)
  {
    convert( rtra );

    // in case rtra->cmp is not properly filled
    if(rtra->cmp == 0) {
      if( rtra->hca > 0 ) Cmp |= Rsegc::ecal;
      if( rtra->hin > 0 ) Cmp |= Rsegc::msg;
      if( rtra->hmu > 0 ) Cmp |= Rsegc::muon;
      if( rtra->hou > 0 ) { Cmp |= Rsegc::patt; Cmp |= Rsegc::magt; }
      if( rtra->hri > 0 ) Cmp |= Rsegc::rich;
      if( rtra->hsi > 0 ) Cmp |= Rsegc::vxd;
      if( rtra->htr > 0 ) Cmp |= Rsegc::trd;
    } else {
      // add missing flags
      if( rtra->hmu > 0 ) Cmp |= Rsegc::muon;
    }
  }

  RecoSegment::RecoSegment( const ArtePointer<RSEG>& rseg ) :
    pmatch (),
    X      (0),
    Y      (0),
    Z      (0),
    Zorig  (0),
    Tx     (0),
    Ty     (0),
    P      (0),
    Chi2   (0),
    Charge (0),
    Cmp    ( (rseg!=0) ? rseg->cmp : 0 ),
    Rtra   ( getIDX<RSEG>( rseg ) ),
    Nf     ( 0 ),
    valid_ ( true ),
    COV_   (0),
    CINV_  (0)
  {
    convert( rseg );
  }

#ifdef NAMESPACE
} // namespace CLUE
#endif
