// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   03. Jan 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: class to fill a BEE event for Single Lepton data analysis
//
// changes:
// 03 Jan 2001 (TG) creation
// 11 Feb 2001 (TG) modified fillRecoVertex()
// 16 Mar 2001 (TG) FillRecoTrack(): added call to maximizeRTRA()
// 27 Apr 2001 (TG) removed maximizeRTRA(), as this is already done in RTRA
// 27 Apr 2001 (TG) exchanged anl_TrSelect() by Cloneremove()
// 08 Jun 2001 (TG) mark VDS only, VDS-ECAL, VDS-ECAL-RICH invalid before Cloneremove
// 09 Jul 2001 (TG) removed filling of user field, fill RecoTrack-RecoVertex relations
// 11 Jul 2001 (TG) fill new RecoTrack-EcalHit relation
// 28 Sep 2001 (TG) added Dmain() cut, added trSelect optionally to
//                  Cloneremove(), because with Cloneremove() strange particle id
//                  spectrum is observed
// 01 Okt 2001 (TG) exchanged EcalHit filling by a more intelligent one, added
//                  hit cut for RecoTracks, fill only events with 1 or 2 good
//                  primary vertices.
//
// *****************************************************************************
#include <vector>
#include <map>
#include <set>
#include "cluearte/SlTableFill.hh"
#include "cluearte/CloneMap.hh"
#include "cluearte/fillEcalHit.hh"

#include "clue/cluearte.hh"
#include "geometry/Point3D.hh" // needed by RecoSegment.hh
#include "clue/RecoTrack.hh"
#include "clue/RecoVertex.hh"
#include "clue/RtEvent.hh"

#include "analysis/anl_evselect.hh"

int  trSelect(int SELMASK); // from MS
bool Dmain(ArtePointer<RTRA> rtra1); // from MS
int  Cloneremove(void);
int  GroverEvent(void);
bool findBremsstrahlung(const ArtePointer<RTRA>& r, ArtePointer<RCCL>&);

extern RtEvent* clueEvent;

// RTRA <-> RecoTrack map
static map<ArtePointer<RTRA>,ArtePointer<RecoTrack> > rtramap;

static const float ZMAINav   = -4.85;   // zentral z-position of a candidate main vertex
static const float ZMAINcut  =  0.2;    // allowed window around ZMAINav for a main vertex


//==============================================================================
// Constructors
//==============================================================================
SlTableFill::SlTableFill() {}

SlTableFill::~SlTableFill() {}

//==============================================================================
// FillMisc
//==============================================================================
bool SlTableFill::FillMisc() const {

  selectRCCL();  // coarse RCCL selection

  return true;
}

//==============================================================================
// EventSelect
//==============================================================================
bool SlTableFill::EventSelect() const {

#ifdef XXX
  // select only events with 1 primary vertex
  unsigned int NPV = 0;
  ArteTableIterator<RVER> rend = ArteTable<RVER>::end();
  for(ArteTableIterator<RVER> it = ArteTable<RVER>::begin();
      it != rend; ++it) {
    if((*it)->type == 1) { ++NPV; }
  }
  if(NPV != 1) return false;
#endif

  if(ArteTable<RTAR>::size() == 0) return false;

  // now run usual event selection
  anl_evselect_info info;
  const int selectFlag = anl_evselect(EVSEL_INNER2, info);
  if( selectFlag != 0 ) return false;

  // check that at least one good primary vertex has been found
  // ==========================================================
  unsigned int nprim = 0;
  const ArteTableIterator<RVER> end = ArteTable<RVER>::end();
  for( ArteTableIterator<RVER> it = ArteTable<RVER>::begin();
       it != end; ++it ) {
    if(it->type!=1 || fabs(it->zv - ZMAINav)>ZMAINcut) continue;
    ++nprim;
  }

  if(nprim == 0) return false;

  return true;
}

//==============================================================================
// FillRecoTrack
//==============================================================================
bool SlTableFill::FillRecoTrack() const {

  //    clueEvent->user.push_back(ArteTable<RTRA>::size());
  //    clueEvent->user.push_back(ArteTable<RVER>::size());

  typedef vector<ArtePointer<RTRA> >                  RTRA_v;
  typedef vector<ArtePointer<RTRA> >::iterator        RTRA_vit;
  typedef vector<ArtePointer<RTRA> >::const_iterator  RTRA_vcit;

  rtramap.clear();

  const ArteTableIterator<RTRA> end = ArteTable<RTRA>::end();
#ifndef CLONEREMOVE
  // before we call Cloneremove, run over RTRA and
  // mark VDS only, VDS-ECAL and VDS-ECAL-RICH tracks
  // as invalid (they are removed later anyhow, but can
  // kill VDS-MAIN tracks in Clonremove())
  for(ArteTable<RTRA>::iterator itrtra=ArteTable<RTRA>::begin();
      itrtra != end; ++itrtra) {
    const ArtePointer<RTRA> rtra = *itrtra;

    // mark tracks with too few hits,
    // VDS only, VDS-ECAL, VDS-RICH-ECAL and
    // reconstructed particles as invalid for Cloneremove()
    if( rtra->hsi < 6                || 
	(rtra->hou + rtra->hin) < 16 ||
	rtra->cmp == 0x10001         ||
	rtra->cmp == 0x10011         ||
	rtra->cmp == 0x4010011       ||
	rtra->fit != 0 ) 
      rtra->fit = 1;
  } // for RTRA

  Cloneremove();
#else
  const int ntrsel  = trSelect(Rsegc::clonerm);
#endif

  // define clone maps
  typedef R_trait<RecoTrack>::Cl_map Cl_map;
  static Cl_map svdclone;
  static Cl_map mainclone;
  static Cl_map richclone;
  static Cl_map ecalclone;
  static Cl_map muonclone;
  svdclone.clear();
  mainclone.clear();
  richclone.clear();
  ecalclone.clear();
  muonclone.clear();

  for( ArteTableIterator<RTRA> rtit = ArteTable<RTRA>::begin();
       rtit != end; ++rtit ) {
    const ArtePointer<RTRA> rtra = *rtit;

#ifdef XXX
    // select clone-free track set, discard VDS only tracks and
    // VDS-ECAL matches
    // (cmp == 1114113 == 0x110001)
    if( (rtra->cmp & Rsegc::clonerm) == 0 ||
	(rtra->cmp == 0x110001)         ||
	(rtra->cmp == 0x110011)
	) continue;
#endif

    // select only tracks which are selected by Cloneremove
    if( (rtra->cmp & Rsegc::clonerm) == 0 ) continue;
    // good tracks are not far from targets
    if(Dmain(rtra) == true) continue;

    //    const ArtePointer<RecoTrack> recotr;
    const ArtePointer<RecoTrack> recotr = 
      ArteTable<RecoTrack>::push_back( RecoTrack(rtra) );

    // fill RTRA <-> RecoTrack map
    rtramap[rtra] = recotr;

    // calc eoverp by assigning clusters: warning: eoverp is stored in mass!
    assignClusters(recotr);

    // fill SVD clone map
    const int csvd = rtra->csvd;
    if( csvd > 0) {
      fillCloneMap<RecoTrack>(svdclone, csvd, recotr);
    }

    // fill OTR clone map
    const int cmain = rtra->cmain;
    if( cmain > 0 ) {
      fillCloneMap<RecoTrack>(mainclone, cmain, recotr);
    }

    // fill RICH clone map
    const int crich = rtra->crich;
    if( crich > 0 ) {
      fillCloneMap<RecoTrack>(richclone, crich, recotr);
    }

    // fill ECAL clone map
    const int cecal = rtra->cecal;
    if( cecal > 0 ) {
      fillCloneMap<RecoTrack>(ecalclone, cecal, recotr);
    }

    // fill Muon clone map
    const int cmuon = rtra->cmuon;
    if( cmuon > 0) {
      fillCloneMap<RecoTrack>(muonclone, cmuon, recotr);
    }

  } // for RTRA
  

  //----------------------------------------------------------------------------
  // Create Clone Relations
  //----------------------------------------------------------------------------
  createCloneRelation<RecoTrack>(svdclone,  Rsegc::vxd );
  createCloneRelation<RecoTrack>(mainclone, Rsegc::patt | Rsegc::magt );
  createCloneRelation<RecoTrack>(richclone, Rsegc::rich );
  createCloneRelation<RecoTrack>(ecalclone, Rsegc::ecal );
  createCloneRelation<RecoTrack>(muonclone, Rsegc::muon );

#ifdef ClueDEBUG
  cout << " clones again: " << endl;
  for(ArteTableIterator<RecoTrack> it = ArteTable<RecoTrack>::begin(); it !=
	ArteTable<RecoTrack>::end(); ++it) {
    if(it->clones.size() > 0) {
      cout << *it << " rtrack->clones: " << it->clones << endl;
    }
  }
#endif

  //----------------------------------------------------------------------------
  // find Bremsstrahlung and fill EcalHit table
  //----------------------------------------------------------------------------
  findBremsstrahlung();
  fillEcalHitTable();

  return true;
}

//==============================================================================
// FillRecoVertex
//==============================================================================
bool SlTableFill::FillRecoVertex() const {

  // first of all clear RVER ARTE table
  ArteTable<RVER>::clear();

  // now, call Grover to find primary vertices
  // to make sure that only selected tracks were used, use the
  // /GROVER/STEERING/TSELECTION  7  data card
  GroverEvent();

  const ArteTableIterator<RVER> end = ArteTable<RVER>::end();
  for( ArteTableIterator<RVER> rvit = ArteTable<RVER>::begin();
       rvit != end; ++rvit ) {
    const ArtePointer<RVER> rver = *rvit;

    // save only good primary vertices (type = 1)
    if( rver->type != 1 ||
	fabs(rver->zv - ZMAINav)>ZMAINcut
	) continue;

    const ArtePointer<RecoVertex> rVertex = 
      ArteTable<RecoVertex>::push_back( RecoVertex(rver) );

    //    double chi2 = 0;
    for( ArteRelationIterator<RREL> rrit = rver->rrel.begin();
	 rrit != rver->rrel.end(); ++rrit ) {
      const ArtePointer<RREL> rrel = *rrit;

      // set track relation by considering clone removal!
      if( rrel->rtra != 0 ) {
	//	const ArtePointer<RecoTrack> rTrack = getPTR<RecoTrack,RTRA>( rrel->rtra );
	if(rtramap.find(rrel->rtra) != rtramap.end()) {
	  createRelation( rVertex, rtramap[rrel->rtra] );
	} else {
	  cout << "Error: could not find RTRA " << rrel->rtra 
	       << " in RecoTrack table!" << endl;
	}
      }
      //      chi2 += rrel->chi;
    }
    //    rVertex->set_chi2(chi2);
  }

  return true;
}

