#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"

#include "grover/grover.hh" // - class defs, f.prototypes and global variables

// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
// 00-11-01  M.Sang   No RTRA corresponding to primary (makes no sense!)
// 00-11-08  M.Sang   Writes chi2 of vertex into RVER; bugfix in cov matrix
// 00-11-13  M.Sang   Now writes vtx 'parent' into RTRA if secondary.
// 00-11-13  M.Sang   Bug fix in momentum writing
// 00-11-15  M.Sang   Bug fix in safety check
// 00-12-11  M.Sang   Initialise all meaningless fields to zero (cleaner)
// 00-12-13  M.Sang   'type' argument is now provided by the user. 1=primary
//                    2 or 3 = secondary vertex (written to RTRA). Also, there
//                    is now an optional third argument which forces the RTRA.topp 
//                    code to this value, regardless of any fitting hypotheses.
// 00-12-14  M.Sang   Bugfix in RTRA covariance elements, reported by Mauro Villa
// 00-12-18  M.Sang   RVER fields 'type' and 'ngdf' now filled
// 01-02-28  M.Sang   Now works with a single vertex OR a list
// 01-03-13  Max      Added preprocessor directives



int GroverAddToRVER(const GroverVertex* ivert,
		    const int vtype = 1, 
		    const ArtePointer<TOPP> tp = 0) 
{
  ////////////////////////////////////////////////////////////////////////
  //
  // Append a GroverVertex to RVER with optional TOPP flag set by hand
  // 
  // 1) Transform to global coordinates if necessary
  // 2) Try to fit unfitted vertices
  // 3) Write RVER and parent RTRA. Manual TOPP choice overrides internal
  //    pointer in the vertex.
  // 4) Set RTRA relations if possible
  //
  ////////////////////////////////////////////////////////////////////////

  int type = vtype%100;
  const bool primary = (type == 1);
  const bool secondary=(type == 2 || type == 3);
  if (ivert == 0) return FAILURE;
  if (!(primary || secondary)) return FAILURE-1;

  bool verbose = false;//(groverCuts.Verbose == 1);
  RVER vertex;
  RTRA track;
  int charge = 0;

  // set all the other stuff to zero
  track.bsi
    = track.cecal = track.cmain  = track.cmp  = track.cmuon
    = track.crich = track.csvd   = track.fid  = track.fit
    = track.fle   = track.flr    = track.flt  = track.hin
    = track.hmu   = track.hou    = track.hri  = track.hsi 
    = track.htr   = track.mommask= track.hca  = track.flm 
    = 0;
  track.cer
    = track.le    = track.lee    = track.leh  = track.lemi 
    = track.lk    = track.lmh    = track.lmmu = track.lmu 
    = track.lp    = track.lpi    = track.lre  = track.lrk 
    = track.lrme  = track.lrmu   = track.lrp  = track.lrpi 
    = track.lte   = track.lth    = track.txr  = track.tyr
    = 0.0;
  
  // first change the coordinate system
  GroverVertex tempvtx = *ivert;
  tempvtx.LocalToGlobal();
  
  // for secondaries, check they've been fit !
  bool skip = false;
  if (secondary) {
    if (tempvtx.Momentum() == FAILURE) {
      if (tempvtx.Size()>1 && tempvtx.Size()<11) {
	tempvtx.Fit(); 
	cout << "*** GroverAddToRVER: WARNING - refitting secondary "
	  "vertex with current constraints!" << endl;
      }
      else {
	cout <<"*** GroverAddToRVER: WARNING - "
	  "strange unfitted secondary passed - ignored" << endl;
	skip = true;
      }
    }
  }
  if (skip) return FAILURE;
  
  // Simple RVER stuff first
  vertex.xv=tempvtx.X();   vertex.yv=tempvtx.Y();   vertex.zv=tempvtx.Z();
  for (int i=0;i<6;i++) vertex.cv[i] = tempvtx.Cov(i);
  vertex.nt   = tempvtx.Size();
  vertex.mass = tempvtx.Mass();
  vertex.merr = tempvtx.MassError();
  vertex.wire = tempvtx.Wire();
  vertex.chi2 = tempvtx.Chi2();
  if (primary) {
    vertex.type = 1;
  }
  else {
    if (tempvtx.Nconstraints()==0) vertex.type = 10;
    else vertex.type = 20 + tempvtx.Nconstraints();
  }
  if (type>99) type = 99;
  vertex.ndgf = tempvtx.Ndof();
  
  // RTRA stuff (secondaries only)
  ArtePointer<RTRA> my_btra = 0;
  if (secondary) {
    float mom = tempvtx.Momentum();
    if (mom <= 0.0) track.pf = 0.0;
    else {
      track.pf = 1.0/mom;
      if (charge<0) track.pf = -1.0*track.pf;
    }
    GroverParticle ttrk(&tempvtx);
    track.xf = ttrk.BeginX();  track.yf = ttrk.BeginY();  track.zf = ttrk.BeginZ();
    track.txf = ttrk.BeginSlopeX();  track.tyf = ttrk.BeginSlopeY();
    // off-diagonals
    for (int i=0; i<14; ++i) track.cvf[i] = ttrk.BeginCov(i);
    track.cvf[14] = (mom>0.0) ? ttrk.BeginCov(14)/(mom*mom*mom*mom) : 1000000.0;
    track.chi = tempvtx.Chi2();
    // end vertex pointer
    ArtePointer<RVER> lastonlist = ArtePointer<RVER>(ArteTable<RVER>::size());
    track.ever= ++lastonlist;
    track.fit = vtype;// changed 13/12/00 MS !
    if (track.chi==0.0) track.fit += 100;
    // Option 1 for TOPP
    track.topp = 0;
    // Option 2 for TOPP
    if (tempvtx.Nconstraints()==1) {
      // special hack for single constraints (22/11/00)
      track.topp = tempvtx.ConstraintTOPP(0);
    }
    // Option 3 for TOPP
    if (tp != 0) track.topp = tp;
    
    ArteTable<RTRA>::push_back(track);
    my_btra = ArtePointer<RTRA>(ArteTable<RTRA>::size());
  }


  vertex.btra = my_btra;
  ArteTable<RVER>::push_back(vertex);
  
  
  // Next, fill the relations between RVER and RTRAs (if possible)
  ArtePointer<RVER> pvertex(ArteTable<RVER>::size()); 
  for (GroverParticleList::const_iterator ipar = ivert->particles.begin();
       ipar != ivert->particles.end(); ++ipar) {
    charge += ipar->Charge();
    if (ipar->Type() == 1) {
      // This GroverParticle was made from an entry in RTRA. 
      // Good! we can fill the relation
      ArtePointer<RTRA> parent = ipar->ParentRTRA();
      if (parent != 0) {
	if (verbose) cout << "*** GroverAddToRVER - setting relation" << endl;
	RREL relation;
	relation.chi  = ipar->Chi2();
	relation.flag = 1;
	relation.topp = ipar->Topp();
	ArteTable<RREL>::push_back(relation);
	ArtePointer<RREL> prelation(ArteTable<RREL>::size());
	createRelation(pvertex, prelation, &RVER::rrel);
	createRelation(parent,  prelation, &RTRA::rrel);
      }
      else {
	if (verbose) 
	  cout << "*** GroverAddToRVER - WARNING: RTRA parent of a vertex track "
	       << " was not set!" << endl;
      }
    }
  }

  return SUCCESS;

}




/////////////////////////////////////////////////////////////////////////////////




int GroverAddToRVER(const GroverVertexList* vertices,
		    const int vtype = 1, 
		    const ArtePointer<TOPP> tp = 0) 
{
  int err = SUCCESS;

  for (GroverVertexList::const_iterator ivert = vertices->begin();
       ivert != vertices->end(); ++ivert)  
    err += GroverAddToRVER(ivert, vtype, tp);

  return err;
}



#endif



