

#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"
#endif

#include "grover/grover.hh"
#include <math.h>

#ifdef USE_ROOT
#ifndef GROVER_WITHOUT_ROOT
#include "TROOT.h"        
#include "TApplication.h" 
#include "TCanvas.h"      
#include "TCONE.h"        
#include "TTUBE.h"        
#include "TBRIK.h"        
#include "TNode.h"        
#include "TRotMatrix.h"   
#endif
#endif
//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
// 00-10-11 Max    First version - works OK, track drawing
//                 disabled (still buggy and I'm away next week)
// 00-11-01 Max    Works fine now (thanks Sven)
// 00-11-08 Max    Added facility to only draw tracks with certain 
//                 minimum p, pt, Nhits, or maximum chi2/dof
// 00-11-14 Max    chi2 cut bug fix
// 01-01-25 Max    Added axis labels
// 01-02-28 Max    const safer
// 01-03-13 Max    Added preprocessor directives
//

bool ggRootRunning;

extern int xargc;    // # of command line arguments of main()
extern char** xargv; // list of -"-

extern int ConvertAngles(const float& tx, const float& ty, 
			 double& theta1, double& phi1,
			 double& theta2, double& phi2,
			 double& theta3, double& phi3);

extern void rnorml(float&);

int GroverDraw3D(const GroverParticleList* Particles,
		 const GroverWireList*     Wires,
		 const GroverVertexList*   Primaries,
		 const GroverVertexList*   Secondaries,
		 const GroverVertexList*   MCPrimaries = 0,
		 const GroverVertexList*   MCSecondaries = 0,
		 const GroverParticleList* Specials = 0,
		 const GroverCuts*         usercuts = 0)
{
#ifdef USE_ROOT
#ifndef GROVER_WITHOUT_ROOT
  
  ///////////////////////////////////////////////////////////////
  //
  // GroverDraw3D
  // 11 Oct 00
  //
  // Simple 3D event display function for Grover.
  // Uses ROOT TGeometry classes and X3D interface. 
  // See http://root.cern.ch for more information.
  //
  ///////////////////////////////////////////////////////////////


  // 1) Cuts and temporary particle
  static const GroverCuts localcuts;
  const GroverCuts* cuts = &localcuts;
  if (usercuts != 0) {
    cuts = usercuts;
  }
  bool verbose          = (cuts->Verbose == 1);
  bool drawprim         = (cuts->TDPrimaries == 1);
  bool drawsec          = (cuts->TDSecondaries == 1);
  bool drawwires        = (cuts->TDWires == 1);
  bool drawaxes         = (cuts->TDAxes == 1);
  bool drawtr           = (cuts->TDTargetRegion == 1);
  bool drawmcprim       = (cuts->TDMCPrimaries == 1);
  bool drawmcsec        = (cuts->TDMCSecondaries == 1);
  bool drawprimtracks   = (cuts->TDPrimaryTracks == 1);
  bool drawsectracks    = (cuts->TDSecondaryTracks == 1);
  bool drawunusedtracks = (cuts->TDUnusedTracks == 1);
  bool drawspecialtracks= (cuts->TDSpecialTracks == 1);
  if ( !(drawprim         || drawsec       || drawwires        || 
	 drawaxes         || drawmcprim    || drawmcsec        || 
	 drawprimtracks   || drawsectracks || drawunusedtracks || 
	 drawspecialtracks                                       )
      ) return SUCCESS;

  GroverParticle  tempgp;
  GroverParticle* p_tempgp = &tempgp;




  // 2) Constants
  const float nsig                 =   3.5; // number of sigma for box size
  const float SCALE                = 500.0; // for X3D step size
  const float TRACKCENTRE          = -3.0;
  const float PRIMARYTRACKLENGTH   = (cuts->TDPrimaryTrackLength)*SCALE;
  const float SECONDARYTRACKLENGTH = (cuts->TDSecondaryTrackLength)*SCALE;
  const float UNUSEDTRACKLENGTH    = (cuts->TDUnusedTrackLength)*SCALE;
  const float SPECIALTRACKLENGTH   = (cuts->TDSpecialTrackLength)*SCALE;
  const float XAXISLENGTH          = (cuts->TDXAxisLength)*SCALE;
  const float YAXISLENGTH          = (cuts->TDYAxisLength)*SCALE;
  const float ZAXISLENGTH          = (cuts->TDZAxisLength)*SCALE;
  const float CONESIZE             = 0.1 *SCALE; // Axis marker size
  const float AXISXPOS             = (cuts->TDAxisX)*SCALE; // Position of axis cross
  const float AXISYPOS             = (cuts->TDAxisY)*SCALE;
  const float AXISZPOS             = (cuts->TDAxisZ)*SCALE;
  const float AXISRADIUS           = (cuts->TDAxisRadius)*SCALE;
  const int   TARGETCOLOUR         = (cuts->TDWireColour);
  const int   PRIMARYCOLOUR        = (cuts->TDPrimaryColour);
  const int   SECONDARYCOLOUR      = (cuts->TDSecondaryColour);
  const int   MCPRIMARYCOLOUR      = (cuts->TDMCPrimaryColour);
  const int   MCSECONDARYCOLOUR    = (cuts->TDMCSecondaryColour);
  const int   PRIMARYTRACKCOLOUR   = (cuts->TDPrimaryTrackColour);
  const int   SECONDARYTRACKCOLOUR = (cuts->TDSecondaryTrackColour);
  const int   UNUSEDTRACKCOLOUR    = (cuts->TDUnusedTrackColour);
  const int   SPECIALTRACKCOLOUR   = (cuts->TDSpecialTrackColour);
  const int   TRACKSMOOTHNESS      = (cuts->TDTrackSmoothness);
  const int   MINNH                = (cuts->TDTrackMinNh);
  const float PCUT                 = (cuts->TDTrackMinP);
  const float PTCUT                = (cuts->TDTrackMinPt);
  const float CHI2CUT              = (cuts->TDTrackMaxChi2);


  const Double_t ut1 = 90.0, ut2 = 90.0, ut3 = 0.0;
  const Double_t up1 =  0.0, up2 = 90.0, up3 = 0.0;


  // 3) Initialisation
  static bool firstcall = true;
  if (firstcall)  cout  << "*** GroverDraw3D:" << endl;
  if (firstcall)  cout  << "    Setting up ROOT/X3D display..." << endl;
  if (!ggRootRunning) {
    if (firstcall) cout << "    ...Grover ROOT object..."<<endl;
    static TROOT GroverRoot("GroverRoot", "Grover ROOT object");
    ggRootRunning = true;
  }
  if (firstcall)  cout  << "    ...Grover TApplication object..."<<endl;
  static TApplication* GroverTApp = 
    new TApplication("GroverTApp",&xargc,xargv);
  if (firstcall)  cout  << "    ...Grover TCanvas..."<<endl;
  static TCanvas* GroverCanvas = 
    new TCanvas("GroverCanvas","Grover Vertex Display",200,10,700,500);




  // 4) Define mother volume
  TBRIK Grover_tr("","","void",
		  XAXISLENGTH*5.0,YAXISLENGTH*5.0,ZAXISLENGTH*5.0);
  Grover_tr.SetVisibility(((drawtr) ? 1 : 0));
  TNode MotherVolume("Mother","Mother volume",&Grover_tr);
  //MotherVolume.cd(); // all other volumes contained in here now




  // 5)  Define volumes for axis arrows and letters
  TRotMatrix Grover_45mat("Grover_45mat","Grover_45mat",
			  ut1,             up1+45,
			  ut2,             up2+45,
			  up3,             up3);
  TRotMatrix Grover_m45mat("Grover_m45mat","Grover_m45mat",
			   ut1,             up1-45,
			   ut2,             up2-45,
			   up3,             up3);
  // X //
  const float XL_LEFTBAR_X  = 0.100 *SCALE, 
              XL_LEFTBAR_Y  = 0.020 *SCALE, 
              XL_LEFTBAR_Z  = 0.020 *SCALE;
  const float XL_RIGHTBAR_X = 0.020 *SCALE, 
              XL_RIGHTBAR_Y = 0.100 *SCALE, 
              XL_RIGHTBAR_Z = 0.020 *SCALE;
  TTUBE Grover_xaxis("","","void",
		     0,AXISRADIUS ,XAXISLENGTH /2.0);
  TCONE Grover_xmrkr("", "","void",
		     CONESIZE ,0.0,CONESIZE ,0.0,0.0);
  TBRIK Grover_xltr_leftbar("", "", "void",
			   XL_LEFTBAR_X, XL_LEFTBAR_Y, XL_LEFTBAR_Z);
  TBRIK Grover_xltr_rightbar("", "", "void",
			   XL_RIGHTBAR_X, XL_RIGHTBAR_Y, XL_RIGHTBAR_Z);
  // Y //
  const float YL_SHORTBAR_X = 0.020 *SCALE, 
              YL_SHORTBAR_Y = 0.063 *SCALE, 
              YL_SHORTBAR_Z = 0.020 *SCALE;
  const float YL_LONGBAR_X  = 0.125 *SCALE, 
              YL_LONGBAR_Y  = 0.020 *SCALE, 
              YL_LONGBAR_Z  = 0.020 *SCALE;
  TTUBE Grover_yaxis("","","void",
		     0,AXISRADIUS ,YAXISLENGTH /2.0);
  TCONE Grover_ymrkr("", "","void",
		     CONESIZE ,0.0,CONESIZE ,0.0,0.0);
  TBRIK Grover_yltr_shortbar("", "", "void",
			   YL_SHORTBAR_X, YL_SHORTBAR_Y, YL_SHORTBAR_Z);
  TBRIK Grover_yltr_longbar("", "", "void",
			   YL_LONGBAR_X, YL_LONGBAR_Y, YL_LONGBAR_Z);

  // Z //
  const float ZL_TOPBAR_X   = 0.100 *SCALE, 
              ZL_TOPBAR_Y   = 0.020 *SCALE, 
              ZL_TOPBAR_Z   = 0.020 *SCALE;
  const float ZL_BOTBAR_X   = 0.100 *SCALE, 
              ZL_BOTBAR_Y   = 0.020 *SCALE, 
              ZL_BOTBAR_Z   = 0.020 *SCALE;
  const float ZL_MIDBAR_X   = 0.100 *SCALE, 
              ZL_MIDBAR_Y   = 0.020 *SCALE, 
              ZL_MIDBAR_Z   = 0.020 *SCALE;
  TTUBE Grover_zaxis("","","void",
		     0,AXISRADIUS ,ZAXISLENGTH /2.0);
  TCONE Grover_zmrkr("", "","void",
		     CONESIZE ,0.0,CONESIZE ,0.0,0.0);
  TBRIK Grover_zltr_topbar("", "", "void",
	        	   ZL_TOPBAR_X, ZL_TOPBAR_Y, ZL_TOPBAR_Z);
  TBRIK Grover_zltr_botbar("", "", "void",
			   ZL_BOTBAR_X, ZL_BOTBAR_Y, ZL_BOTBAR_Z);
  TBRIK Grover_zltr_midbar("", "", "void",
			   ZL_MIDBAR_X, ZL_MIDBAR_Y, ZL_MIDBAR_Z);


  // cosmetic
  Grover_xaxis.SetNumberOfDivisions(4);
  Grover_yaxis.SetNumberOfDivisions(4);
  Grover_zaxis.SetNumberOfDivisions(4);
  Grover_xaxis.SetLineColor(1); Grover_xmrkr.SetLineColor(2);
  Grover_yaxis.SetLineColor(1); Grover_ymrkr.SetLineColor(3);
  Grover_zaxis.SetLineColor(1); Grover_zmrkr.SetLineColor(4);
  // rotation matrices
  TRotMatrix Grover_unit(  "Grover_unit",  "Grover_unit",
			   ut1,       up1,
			   ut2,       up2,
			   ut3,       up3);
  TRotMatrix Grover_zaxmat("Grover_zaxmat","Grover_zaxmat",
			   ut1,       up1,
			   ut2,       up2,
			   ut3,       up3);
  TRotMatrix Grover_xaxmat("Grover_xaxmat","Grover_xaxmat",
			   ut1+90.0,  up1,
			   ut2,       up2,
			   ut3+90.0,  up3);
  TRotMatrix Grover_yaxmat("Grover_yaxmat","Grover_yaxmat",
			   ut1+90.0,  up1,
			   ut2,       up2+90.0,
			   ut3+90.0,  up3+90.0);
  if (drawaxes) {
    TNode* Grover_node2 = new TNode("","",&Grover_zaxis,
				    AXISXPOS ,
				    AXISYPOS ,
				    AXISZPOS ,
				    &Grover_zaxmat);
    TNode* Grover_node3 = new TNode("","",&Grover_xaxis,
				    AXISXPOS ,
				    AXISYPOS ,
				    AXISZPOS ,
				    &Grover_xaxmat);
    TNode* Grover_node4 = new TNode("","",&Grover_yaxis,
				    AXISXPOS ,
				    AXISYPOS ,
				    AXISZPOS ,
				    &Grover_yaxmat);
    TNode* Grover_node5 = new TNode("","",&Grover_zmrkr,
				    AXISXPOS ,
				    AXISYPOS ,
				    AXISZPOS +(ZAXISLENGTH /2.0),
				    &Grover_zaxmat);
    TNode* Grover_node6 = new TNode("","",&Grover_xmrkr,
				    AXISXPOS +(XAXISLENGTH /2.0),
				    AXISYPOS ,
				    AXISZPOS ,
				    &Grover_xaxmat);
    TNode* Grover_node7 = new TNode("","",&Grover_ymrkr,
				    AXISXPOS,
				    AXISYPOS+(YAXISLENGTH /2.0),
				    AXISZPOS,
				    &Grover_yaxmat);
    TNode* Grover_node8 = new TNode("","",&Grover_xltr_leftbar,
				    AXISXPOS + XAXISLENGTH/2.0 + 3.0*CONESIZE,
				    AXISYPOS,
				    AXISZPOS,
				    &Grover_45mat);
    TNode* Grover_node9 = new TNode("","",&Grover_xltr_rightbar,
				    AXISXPOS + XAXISLENGTH/2.0 + 3.0*CONESIZE,
				    AXISYPOS,
				    AXISZPOS,
				    &Grover_45mat);
    TNode* Grover_node10= new TNode("","",&Grover_yltr_shortbar,
				    AXISXPOS - (0.707*(YL_SHORTBAR_Y+YL_LONGBAR_Y)),
				    AXISYPOS + YAXISLENGTH/2.0 
				             + 3.0*CONESIZE 
				             + (0.707*(YL_SHORTBAR_Y+YL_LONGBAR_Y)),
				    AXISZPOS,
				    &Grover_45mat);
    TNode* Grover_node11= new TNode("","",&Grover_yltr_longbar,
				    AXISXPOS,
				    AXISYPOS + YAXISLENGTH/2.0 + 3.0*CONESIZE,
				    AXISZPOS,
				    &Grover_45mat);
    TNode* Grover_node12= new TNode("","",&Grover_zltr_topbar,
				    AXISXPOS,
				    AXISYPOS + ZL_TOPBAR_X/(2*0.707),
				    AXISZPOS + ZAXISLENGTH/2.0
				             + 3.0*CONESIZE,
				    &Grover_unit);
    TNode* Grover_node13= new TNode("","",&Grover_zltr_midbar,
				    AXISXPOS,
				    AXISYPOS,
				    AXISZPOS + ZAXISLENGTH/2.0
				             + 3.0*CONESIZE,
				    &Grover_45mat);
    TNode* Grover_node14= new TNode("","",&Grover_zltr_botbar,
				    AXISXPOS,
				    AXISYPOS - ZL_BOTBAR_X/(2*0.707),
				    AXISZPOS + ZAXISLENGTH/2.0
				             + 3.0*CONESIZE,
				    &Grover_unit);
  }


  
  // 6) Volumes for target wires
  int nodecounter = 0; // number of nodes
  char mname[40], sname[40], nname[40]; // matrix,shape,node names
  if (Wires != 0 && drawwires) {
    for (GroverWireList::const_iterator iwire = Wires->begin();
	 iwire != Wires->end(); ++iwire) {
      ++nodecounter;
      float  x  = iwire->X()*SCALE,      y = iwire->Y()*SCALE,      z = iwire->Z()*SCALE;
      float dx  = iwire->SizeX()*SCALE, dy = iwire->SizeY()*SCALE, dz = iwire->SizeZ()*SCALE;
      if (iwire->Number() == 2) {
	float swap = dy; dy=dz; dz=swap; // dirty hack for MC
      }
      sprintf(mname,"matrix%d",nodecounter);
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TBRIK *shape = new TBRIK(sname,sname,"void",dx /2.0,dy /2.0,dz /2.0);
      shape->SetLineColor(TARGETCOLOUR);
      TRotMatrix *matrix = new TRotMatrix(mname,mname,ut1,up1,ut2,up2,ut3,up3);
      TNode *node = new TNode(nname,nname,shape,
			      x ,y ,z ,matrix);        
    }
  }





  // 7) Volumes for primary vertices
  if (Primaries != 0 && drawprim) {
    for (GroverVertexList::const_iterator iprim = Primaries->begin();
	 iprim != Primaries->end(); ++iprim) {
      ++nodecounter;
      float x = iprim->X()*SCALE, y = iprim->Y()*SCALE, z = iprim->Z()*SCALE;
      float dx  = sqrt(fabs(iprim->Cov(0)))*SCALE;
      float dy  = sqrt(fabs(iprim->Cov(2)))*SCALE;
      float dz  = sqrt(fabs(iprim->Cov(5)))*SCALE;
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TBRIK *shape = new TBRIK(sname,sname,"void",
			       dx*nsig ,
			       dy*nsig ,
			       dz*nsig );
      shape->SetLineColor(PRIMARYCOLOUR);
      TNode *node = new TNode(nname,nname,shape,
			      x ,y ,z ,
			      &Grover_zaxmat);        
    }
  }





  // 8) Volumes for MC primary vertices
  if (MCPrimaries != 0 && drawmcprim) {
    for (GroverVertexList::const_iterator iprim = MCPrimaries->begin();
	 iprim != MCPrimaries->end(); ++iprim) {
      ++nodecounter;
      float x = iprim->X()*SCALE, y = iprim->Y()*SCALE, z = iprim->Z()*SCALE;
      float dx  = 0.005*SCALE; // perhaps a little sphere instead?
      float dy  = 0.005*SCALE;
      float dz  = 0.005*SCALE;
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TBRIK *shape = new TBRIK(sname,sname,"void",
			       dx ,
			       dy ,
			       dz );
      shape->SetLineColor(MCPRIMARYCOLOUR);
      TNode *node = new TNode(nname,nname,shape,
			      x ,y ,z ,
			      &Grover_zaxmat);        
    }
  }




  // 9) Volumes for Primary tracks
  if (Primaries != 0 && drawprimtracks) {
    for (GroverVertexList::const_iterator iprim = Primaries->begin();
         iprim != Primaries->end(); ++iprim) {
      for (GroverParticleList::const_iterator ipar = iprim->particles.begin();
           ipar != iprim->particles.end(); ++ipar) {
	tempgp = *ipar;
	float p  = p_tempgp->Momentum();
	float px = p*sin(atan(p_tempgp->BeginSlopeX()));
        float py = p*sin(atan(p_tempgp->BeginSlopeY())); 
	float pt = sqrt(px*px + py*py);
	if (p_tempgp->Momentum() < PCUT ||
	    pt < PTCUT ||
	    p_tempgp->Chi2() > CHI2CUT ||
	    p_tempgp->Nhit() < MINNH) continue; 
        ++nodecounter;
        p_tempgp->ResetParameters(); 
        p_tempgp->PropagateBeginTo(iprim->Z());
        float x = p_tempgp->BeginX()*SCALE, y = p_tempgp->BeginY()*SCALE, z = p_tempgp->BeginZ()*SCALE;
        float dx2  = p_tempgp->BeginCov(0);
        float dy2  = p_tempgp->BeginCov(2);
        float dz   = PRIMARYTRACKLENGTH;//fabs(ZAXISLENGTH - z);
        float d    = sqrt(fabs(dx2+dy2))*SCALE;
        sprintf(sname,"shape%d",nodecounter);
        sprintf(nname,"node%d",nodecounter);
        TTUBE *shape = new TTUBE(sname,sname,"void",
                                 0.0, d , dz /2.0);
        shape->SetLineColor(PRIMARYTRACKCOLOUR);
        shape->SetNumberOfDivisions(TRACKSMOOTHNESS);
        float tx = p_tempgp->BeginSlopeX();
        float ty = p_tempgp->BeginSlopeY();
	double theta[3]={0.0, 0.0, 0.0};
	double phi[3]={0.0, 0.0, 0.0};
	ConvertAngles(tx,ty,theta[0],phi[0],theta[1],phi[1],theta[2],phi[2]);
	TRotMatrix *matrix = 
	  new TRotMatrix(mname,mname,
			 theta[0], phi[0],
			 theta[1], phi[1],      
			 theta[2], phi[2]);     
	TNode *node = new TNode(nname,nname,shape,
				x ,y ,z ,
				matrix);        
      }
    }
  }



  // 10) Volumes for secondary vertices
  if (Secondaries != 0 && drawsec) {
    for (GroverVertexList::const_iterator isec = Secondaries->begin();
	 isec != Secondaries->end(); ++isec) {
      ++nodecounter;
      float x = isec->X()*SCALE, y = isec->Y()*SCALE, z = isec->Z()*SCALE;
      float dx  = sqrt(fabs(isec->Cov(0)))*SCALE;
      float dy  = sqrt(fabs(isec->Cov(2)))*SCALE;
      float dz  = sqrt(fabs(isec->Cov(5)))*SCALE;
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TBRIK *shape = new TBRIK(sname,sname,"void",
			       dx*nsig ,
			       dy*nsig ,
			       dz*nsig );
      shape->SetLineColor(SECONDARYCOLOUR);
      TNode *node = new TNode(nname,nname,shape,
			      x ,y ,z ,
			      &Grover_zaxmat); 
    }
  }



  // 11) Volumes for MC secondary vertices
  if (MCSecondaries != 0 && drawmcsec) {
    for (GroverVertexList::const_iterator isec = MCSecondaries->begin();
	 isec != MCSecondaries->end(); ++isec) {
      ++nodecounter;
      float x = isec->X()*SCALE, y = isec->Y()*SCALE, z = isec->Z()*SCALE;
      float dx  = 0.005*SCALE; // perhaps a little sphere instead?
      float dy  = 0.005*SCALE;
      float dz  = 0.005*SCALE;
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TBRIK *shape = new TBRIK(sname,sname,"void",
			       dx ,
			       dy ,
			       dz );
      shape->SetLineColor(MCSECONDARYCOLOUR);
      TNode *node = new TNode(nname,nname,shape,
			      x ,y ,z ,
			      &Grover_zaxmat);        
    }
  }




  // 12) Volumes for secondary tracks
  if (Secondaries != 0 && drawsectracks) {
    for (GroverVertexList::const_iterator isec = Secondaries->begin();
	 isec != Secondaries->end(); ++isec) {
      for (GroverParticleList::const_iterator ipar = isec->particles.begin();
	   ipar != isec->particles.end(); ++ipar) {
        float p  = ipar->Momentum();
        float px = p*sin(atan(ipar->BeginSlopeX()));
	float py = p*sin(atan(ipar->BeginSlopeY()));
        float pt = sqrt(px*px + py*py);
	if (ipar->Momentum() < PCUT ||
	    pt < PTCUT ||
	    ipar->Chi2() > CHI2CUT ||
	    ipar->Nhit() < MINNH) continue;
	++nodecounter;
        p_tempgp->ResetParameters(); 
        p_tempgp->PropagateBeginTo(isec->Z());
	float x = p_tempgp->BeginX()*SCALE, y = p_tempgp->BeginY()*SCALE, z = p_tempgp->BeginZ()*SCALE;
        float dx2  = p_tempgp->BeginCov(0);
        float dy2  = p_tempgp->BeginCov(2);
	float dz   = SECONDARYTRACKLENGTH; //fabs(ZAXISLENGTH - z);
	float d    = sqrt(fabs(dx2+dy2))*SCALE;
	sprintf(sname,"shape%d",nodecounter);
	sprintf(nname,"node%d",nodecounter);
	TTUBE *shape = new TTUBE(sname,sname,"void",
				 0.0, d , dz /2.0);
	shape->SetLineColor(SECONDARYTRACKCOLOUR);
	shape->SetNumberOfDivisions(TRACKSMOOTHNESS);
	float tx = p_tempgp->BeginSlopeX();
        float ty = p_tempgp->BeginSlopeY();
	double theta[3]={0.0, 0.0, 0.0};
	double phi[3]={0.0, 0.0, 0.0};
	ConvertAngles(tx,ty,theta[0],phi[0],theta[1],phi[1],theta[2],phi[2]); 
	TRotMatrix *matrix = 
	  new TRotMatrix(mname,mname,
			 theta[0], phi[0],
			 theta[1], phi[1],      
			 theta[2], phi[2]);     
	TNode *node = new TNode(nname,nname,shape,
				x ,y ,z ,
				matrix);        
      }
    }
  }




  // 13) Volumes for unused tracks
  if (Particles != 0 && drawunusedtracks) {
    for (GroverParticleList::const_iterator ipar = Particles->begin();
	 ipar != Particles->end(); ++ipar) {
      if (ipar->UsedInSecondary() != 0 ||
	  ipar->UsedInPrimary()   != 0   ) continue;
      tempgp = *ipar;
      float p  = p_tempgp->Momentum();
      float px = p*sin(atan(p_tempgp->BeginSlopeX()));
      float py = p*sin(atan(p_tempgp->BeginSlopeY()));
      float pt = sqrt(px*px + py*py);
      if (p_tempgp->Momentum() < PCUT ||
	  pt < PTCUT ||
	  p_tempgp->Chi2() > CHI2CUT ||
	  p_tempgp->Nhit() < MINNH) continue; 
      ++nodecounter;
      p_tempgp->PropagateBeginTo(TRACKCENTRE);
      float x = p_tempgp->BeginX()*SCALE, y = p_tempgp->BeginY()*SCALE, z = p_tempgp->BeginZ()*SCALE;
      float dx2  = p_tempgp->BeginCov(0);
      float dy2  = p_tempgp->BeginCov(2);
      float dz   = UNUSEDTRACKLENGTH;//fabs(ZAXISLENGTH - z);
      float d    = sqrt(fabs(dx2+dy2))*SCALE;
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TTUBE *shape = new TTUBE(sname,sname,"void",
			       0.0, d , dz /2.0);
      shape->SetLineColor(UNUSEDTRACKCOLOUR);
      shape->SetNumberOfDivisions(TRACKSMOOTHNESS);
      float tx = p_tempgp->BeginSlopeX();
      float ty = p_tempgp->BeginSlopeY();       
      double theta[3]={0.0, 0.0, 0.0};
      double phi[3]={0.0, 0.0, 0.0};
      ConvertAngles(tx,ty,theta[0],phi[0],theta[1],phi[1],theta[2],phi[2]);
      TRotMatrix *matrix = 
	new TRotMatrix(mname,mname,
		       theta[0], phi[0],
		       theta[1], phi[1],      
		       theta[2], phi[2]);     
      TNode *node = new TNode(nname,nname,shape,
			      x ,y ,z ,
			      matrix);        
    }
  }
  
  
  


  // 14) Volumes for special tracks
  if (Specials != 0 && drawspecialtracks) {
    for (GroverParticleList::const_iterator ipar = Specials->begin();
	 ipar != Specials->end(); ++ipar) {
      ++nodecounter;
      float x = ipar->BeginX()*SCALE, y = ipar->BeginY()*SCALE, z = ipar->BeginZ()*SCALE;
      float dx2  = ipar->BeginCov(0);
      float dy2  = ipar->BeginCov(2);
      float dz   = SPECIALTRACKLENGTH;//fabs(ZAXISLENGTH - z);
      float d    = sqrt(fabs(dx2+dy2))*SCALE;
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TTUBE *shape = new TTUBE(sname,sname,"void",
			       0.0, d , dz /2.0);
      shape->SetLineColor(SPECIALTRACKCOLOUR);
      shape->SetNumberOfDivisions(TRACKSMOOTHNESS);
      float tx = ipar->BeginSlopeX();
      float ty = ipar->BeginSlopeY();       
      double theta[3]={0.0, 0.0, 0.0};
      double phi[3]={0.0, 0.0, 0.0};
      ConvertAngles(tx,ty,theta[0],phi[0],theta[1],phi[1],theta[2],phi[2]);
      TRotMatrix *matrix = 
	new TRotMatrix(mname,mname,
		       theta[0], phi[0],
		       theta[1], phi[1],      
		       theta[2], phi[2]);     
      TNode *node = new TNode(nname,nname,shape,
			      x ,y ,z ,
			      matrix);        
    }
  }





  // 15) DRAW IT !!!
  MotherVolume.Draw();
  GroverCanvas->Update();
  if (verbose) cout << "*** GroverDraw3D: Mother volume has "
		    << nodecounter 
		    << " TNode daughters" << endl;
  if (cuts->TDX3D == 1) {
    cout << "*** GroverDraw3D: type 'm' in X3D window for menu"
	 << endl <<"                       'q' to quit and return to ARTE" 
	 << endl;
    GroverCanvas->x3d();
  }


  firstcall = false;

#endif  
#endif
  return SUCCESS;
}






////////////////////////////////////////////////////////////////////////



int ConvertAngles(const float& tx, const float& ty, 
		  double& theta1, double& phi1,
		  double& theta2, double& phi2,
		  double& theta3, double& phi3) 
{
  //
  // Converts tx,ty to the six Geant angles
  // 
  // S.Schaller 20/10/00
  //

  double a[3][3];
  double theta[3] = {0.0, 0.0, 0.0};
  double phi[3]   = {0.0, 0.0, 0.0};
  int i;         
  const double raddeg = 57.2957795131;        // 1 radian in deg

  double Theta = atan2(sqrt(tx*tx+ty*ty),1.0); // polar angle in radians
  double Phi   = atan2(ty,tx);                 // azimuthal angle in radians

  double ct = cos(Theta);
  double st = sin(Theta);
  double cp = cos(Phi);
  double sp = sin(Phi);
  
  // Rotation Matrix 
  a[0][0] = ct*cp;
  a[0][1] = ct*sp;
  a[0][2] = -1.0*st;
  
  a[1][0] = -1.0*sp;
  a[1][1] = cp;
  a[1][2] = 0;
  
  a[2][0] = st*cp;
  a[2][1] = st*sp;
  a[2][2] = ct; 

  for (i=0; i<3; i++) {  
    theta[i] = raddeg*atan2(sqrt(a[i][0]*a[i][0] +
    	                         a[i][1]*a[i][1]  ),
			         a[i][2]);
    phi[i]   = raddeg*atan2(a[i][1],
			    a[i][0]);  
  }

  theta1 = theta[0];
  theta2 = theta[1];
  theta3 = theta[2];
  phi1   = phi[0];
  phi2   = phi[1];
  phi3   = phi[2];

  return SUCCESS;
}


/////////////////////////////////////////////////////////////////////////


// A nice and easy way, since we can't have it in the kuip interface because
// Root won't run on the fourth level farm.
int GroverDraw3D(void)
{
  int err = SUCCESS;
  cout <<"*** GroverDraw3D: ggX sizes: par,prim,mcprim,sec,mcsec,wires,specials= "
       << ggParticles.size() << " "
       << ggPrimaries.size() << " "
       << ggMCPrimaries.size() << " "
       << ggSecondaries.size() << " "
       << ggMCSecondaries.size() << " "
       << ggWires.size() << " "
       << ggSpecialParticles.size() << " "
       << endl;
  err = GroverDraw3D(&ggParticles, &ggWires, &ggPrimaries, &ggSecondaries, 
		     &ggMCPrimaries, &ggMCSecondaries, &ggSpecialParticles, 
		     &ggCuts);
  return err;
}

