#include <algorithm>
#include <math.h>
#include "grover/DumpFormatter.hh"

#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"
#endif

#include "grover/grover.hh"

extern "C" {
#ifndef GROVER_WITHOUT_CERNLIB
  float prob_(const float&, const int&); // from cernlib
#endif
}

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                 GroverVertex member functions                      //
//                                                                    //
////////////////////////////////////////////////////////////////////////

//
//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
// 00-09-21  M.Sang   Introduced 'fit / not yet fit' flag for safety
// 00-09-26  M.Sang   Some dummy members for future (multi-constraint) fits
// 00-10-12  MS/DE    Added multiple-constraint fit Fit(void)
// 00-10-12  M.Sang   Increased array sizes and safety checks in both Fit(...)
// 00-11-16  M.Sang   SlopeX() and SlopeY() no longer inlined (safety checks)
// 00-11-20  M.Sang   improvement to GroverVertex(ArtePointer<MTRA>)
// 00-11-20  M.Sang   Bug fix in GroverVertex(ArtePointer<RVER>)
// 00-12-01  I.Kisel  Bug fix in chi^2 calculation for Fit(float m>0)
// 00-12-01  I.Kisel  Removed multiplication of V(p) by p^4 (done twice)
// 00-12-14  M.Sang   Improved GroverVertex(ArtePointer<RVER>) and updated
//                    initialisers & constructor bodies
// 00-12-18  D.Emeliyanov  Bug in GroverVertex::Fit cov. matrix fixed
// 01-01-12  M.Sang   Added operators ==, != and <<
// 01-01-18  M.Sang   GeoFit(vtx*) => vtx.GeoFit() (reverses change
//                    of May 16 2000)
// 01-01-25  M.Sang   Improved behaviour for MC tracks & vertices
// 01-01-25  M.Sang   Improved safety of GroverVertex::Remove(GroverParticle*)
// 01-01-25  M.Sang   Improved operator==
// 01-01-31  M.Sang   Correct GroverVertex::Prob(void).
// 01-02-15  M.Sang   Fit(...) now checks for user intervention before calling
//                    ResetParticles();
// 01-02-26  M.Sang   Fixed bug in GroverVertex::SlopeX & SlopeY reported
//                    by Jing-Ge Shiu
// 01-02-28  M.Sang   Chi2 of fits NO LONGER NORMALISED!
// 01-02-28  M.Sang   Chi2DistTo(X) output will now be very different where 
//                    X is a GroverVertex, a GroverWire, or a space point.
// 01-02-28  M.Sang   Fit() calls ResetParticles() ONLY if none of the 
//                    parameters of the tracks in vtx.particles have been 
//                    altered (this includes the momentum, cov. matrix, etc.)
// 01-02-28  M.Sang   Maximum number of tracks in a secondary is 100
// 01-03-13  M.Sang   Added preprocessor directives
// 01-04-09  Michael  In GeoFit set gparticle values direct instead with the
//                    GroverParticle maember functions. 
//                    To quere tampered bug. 
// 01-04-17 Michael   bug fix in GroverVertex::operator==
// 01-05-28 Michael   bug fix in ndim checking
// 01-07-05 Michael   added Jing-GE Shiu's pointing constrained vertex fit
//                    Fitp for testing. 
//                    See: ~physjg/public/pgrover/pgrover.ps.gz
// 01-07-06 Michael   ROVERMAXTRACKS in GeoFit is 200 as in GroverPrimaries
// 01-09-08 DE        Add GROVER_WITHOUT_ARTE envelope for Fitp.
// 
// 02-04-08 DE        Two routines FastFit and RobustFit are included
// 02-05-21 DE        FastFit and RobustFit timings are optimized
//                    NewGeoFit is added.
//
//   ==> SEE ALSO grover_inlines.hh !!!!!!!
//
//

float emptycov[45] = {1e3,
			0,1e3,
			0,  0,1e3,
			0,  0,  0,1e3,
			0,  0,  0,  0,1e3,
			0,  0,  0,  0,  0,1e3,
			0,  0,  0,  0,  0,  0,1e3,
			0,  0,  0,  0,  0,  0,  0,1e3,
			0,  0,  0,  0,  0,  0,  0,  0,1e3};



GroverVertex::GroverVertex()
  : index(0), x(0.), y(0.), z(0.), ax(0.), ay(0.), p(0.),
    m(0.), dm(0.), inlocal(false), inglobal(false),
    wprim(0.), wsec(0.), wnr(0),
    chi2(0.0), fitted(false), mom_estimated(false), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{// default constructor
  for (int k=0;k<COVDIM;++k) cov[k]=emptycov[k];
  for (int k=0;k<CONSTDIM;++k) {
    for (int l=0;l<MAXTRKS;++l) constr_map[k][l]=0;
    constr_mass[k] = 0.0;
#ifndef GROVER_WITHOUT_ARTE
    constr_topps[k]= 0;
#endif
  }
}



GroverVertex::~GroverVertex()
{// default destructor
}



GroverVertex::GroverVertex(const GroverParticle* gp)
  : index(0), x(0.), y(0.), z(0.), ax(0.), ay(0.), p(0.),
    m(0.), dm(0.), 
    inlocal(gp->InLocal()), inglobal(!inlocal),
    wprim(0.), wsec(0.), wnr(0),
    chi2(0.0), fitted(false), mom_estimated(false), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{// very simple constructor for a single particle (eg tag lepton seed)
  for (int k=0;k<COVDIM;++k) cov[k]=emptycov[k];
  for (int k=0;k<CONSTDIM;++k) {
    for (int l=0;l<MAXTRKS;++l) constr_map[k][l]=0;
    constr_mass[k] = 0.0;
#ifndef GROVER_WITHOUT_ARTE
    constr_topps[k]= 0;
#endif
  }
  if (gp != 0) particles.push_back(*gp);
}




GroverVertex::GroverVertex(const GroverParticle* gp1, const GroverParticle* gp2)
  : index(0), x(0.), y(0.), z(0.), ax(0.), ay(0.), p(0.),
    m(0.), dm(0.), 
    inlocal(gp1->InLocal()), inglobal(!inlocal),
    wprim(0.), wsec(0.), wnr(0),
    chi2(0.0), fitted(false), mom_estimated(false), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{ // construct from 2 particles
  for (int k=0;k<COVDIM;++k) cov[k]=emptycov[k];
  for (int k=0;k<CONSTDIM;++k) {
    for (int l=0;l<MAXTRKS;++l) constr_map[k][l]=0;
#ifndef GROVER_WITHOUT_ARTE
    constr_topps[k]= 0;
#endif
    constr_mass[k] = 0.0;
  }
  if (gp1 != 0) particles.push_back(*gp1);
  if (gp2 != 0) particles.push_back(*gp2);
}


GroverVertex::GroverVertex(const GroverParticle* gp1, const GroverParticle* gp2,
			   const GroverParticle* gp3)
  : index(0), x(0.), y(0.), z(0.), ax(0.), ay(0.), p(0.),
    m(0.), dm(0.), 
    inlocal(gp1->InLocal()), inglobal(!inlocal),
    wprim(0.), wsec(0.), wnr(0),
    chi2(0.0), fitted(false), mom_estimated(false), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{ // construct from 3 particles
  for (int k=0;k<COVDIM;++k) cov[k]=emptycov[k];
  for (int k=0;k<CONSTDIM;++k) {
    for (int l=0;l<MAXTRKS;++l) {
      constr_map[k][l]=0;
    }
    constr_mass[k] = 0.0;
#ifndef GROVER_WITHOUT_ARTE
    constr_topps[k]= 0;
#endif
  }
  if (gp1 != 0) particles.push_back(*gp1);
  if (gp2 != 0) particles.push_back(*gp2);
  if (gp3 != 0) particles.push_back(*gp3);
}


GroverVertex::GroverVertex(const GroverParticle* gp1, const GroverParticle* gp2,
			   const GroverParticle* gp3, const GroverParticle* gp4)
  : index(0), x(0.), y(0.), z(0.), ax(0.), ay(0.), p(0.),
    m(0.), dm(0.), 
    inlocal(gp1->InLocal()), inglobal(!inlocal),
    wprim(0.), wsec(0.), wnr(0),
    chi2(0.0), fitted(false), mom_estimated(false), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{ // construct from 4 particles
  for (int k=0;k<COVDIM;++k) cov[k]=emptycov[k];
  for (int k=0;k<CONSTDIM;++k) {
    for (int l=0;l<MAXTRKS;++l) constr_map[k][l]=0;
    constr_mass[k] = 0.0;
#ifndef GROVER_WITHOUT_ARTE
    constr_topps[k]= 0;
#endif
  }
  if (gp1 != 0) particles.push_back(*gp1);
  if (gp2 != 0) particles.push_back(*gp2);
  if (gp3 != 0) particles.push_back(*gp3);
  if (gp4 != 0) particles.push_back(*gp4);
}


GroverVertex::GroverVertex(const GroverParticle* gp1, const GroverParticle* gp2,
			   const GroverParticle* gp3, const GroverParticle* gp4,
			   const GroverParticle* gp5)
  : index(0), x(0.), y(0.), z(0.), ax(0.), ay(0.), p(0.),
    m(0.), dm(0.), 
    inlocal(gp1->InLocal()), inglobal(!inlocal),
    wprim(0.), wsec(0.), wnr(0),
    chi2(0.0), fitted(false), mom_estimated(false), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{ // construct from 5 particles
  for (int k=0;k<COVDIM;++k) cov[k]=emptycov[k];
  for (int k=0;k<CONSTDIM;++k) {
    for (int l=0;l<MAXTRKS;++l) constr_map[k][l]=0;
    constr_mass[k] = 0.0;
#ifndef GROVER_WITHOUT_ARTE
    constr_topps[k]= 0;
#endif
  }
  if (gp1 != 0) particles.push_back(*gp1);
  if (gp2 != 0) particles.push_back(*gp2);
  if (gp3 != 0) particles.push_back(*gp3);
  if (gp4 != 0) particles.push_back(*gp4);
  if (gp5 != 0) particles.push_back(*gp5);
}


GroverVertex::GroverVertex(const GroverParticle* gp1, const GroverParticle* gp2,
			   const GroverParticle* gp3, const GroverParticle* gp4,
			   const GroverParticle* gp5, const GroverParticle* gp6)
  : index(0), x(0.), y(0.), z(0.), ax(0.), ay(0.), p(0.),
    m(0.), dm(0.), 
    inlocal(gp1->InLocal()), inglobal(!inlocal),
    wprim(0.), wsec(0.), wnr(0),
    chi2(0.0), fitted(false), mom_estimated(false), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{ // construct from 6 particles
  for (int k=0;k<COVDIM;++k) cov[k]=emptycov[k];
  for (int k=0;k<CONSTDIM;++k) {
    for (int l=0;l<MAXTRKS;++l) constr_map[k][l]=0;
    constr_mass[k] = 0.0;
#ifndef GROVER_WITHOUT_ARTE
    constr_topps[k]= 0;
#endif
  }
  if (gp1 != 0) particles.push_back(*gp1);
  if (gp2 != 0) particles.push_back(*gp2);
  if (gp3 != 0) particles.push_back(*gp3);
  if (gp4 != 0) particles.push_back(*gp4);
  if (gp5 != 0) particles.push_back(*gp5);
  if (gp6 != 0) particles.push_back(*gp6);
}


GroverVertex::GroverVertex(const GroverParticleList* gpl)
  : index(0), x(0.), y(0.), z(0.), ax(0.), ay(0.), p(0.),
    m(0.), dm(0.), 
    inlocal(false), inglobal(false),
    wprim(0.), wsec(0.), wnr(0),
    chi2(0.0), fitted(false), mom_estimated(false), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{ // construct from a list of particles - keep it simple for now...
  // Coordinate system choice determined by last particle on list!
  for (GroverParticleList::const_iterator ipar = gpl->begin();
       ipar != gpl->end(); ++ipar){
    particles.push_back(*ipar);
  }
  if (!gpl->empty()) {
    inglobal = gpl->back().InGlobal();
    inlocal  = !inglobal;
  }
  for (int k=0;k<COVDIM;++k) cov[k]=emptycov[k];
  for (int k=0;k<CONSTDIM;++k) {
    for (int l=0;l<MAXTRKS;++l) constr_map[k][l]=0;
    constr_mass[k] = 0.0;
#ifndef GROVER_WITHOUT_ARTE
    constr_topps[k]= 0;
#endif
  }
}



GroverVertex::GroverVertex(const float xx, const float yy, const float zz)
  : index(0), x(xx), y(yy), z(zz), ax(0.), ay(0.), p(0.),
    m(0.), dm(0.), inlocal(true), inglobal(false),
    wprim(0.), wsec(0.), wnr(0),
    chi2(0.0), fitted(false), mom_estimated(false), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{// construct at a space point in local coordinates
  for (int k=0;k<COVDIM;++k) cov[k]=emptycov[k];
  for (int k=0;k<CONSTDIM;++k) {
    for (int l=0;l<MAXTRKS;++l) constr_map[k][l]=0;
    constr_mass[k] = 0.0;
#ifndef GROVER_WITHOUT_ARTE
    constr_topps[k]= 0;
#endif
  }
}


#ifndef GROVER_WITHOUT_ARTE
GroverVertex::GroverVertex(const ArtePointer<RVER> rver)
  : index(0),  x(rver->xv), y(rver->yv), z(rver->zv),
    ax(0.0), ay(0.0), p(0.0),
    m(rver->mass), dm(rver->merr), 
    inlocal(false), inglobal(true),
    wprim(0.), wsec(0.), wnr(rver->wire),
    chi2(0.0), fitted(true), mom_estimated(rver->btra!=0), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{ // construct from a entry in RVER
  for (int k=0;k<6; ++k) cov[k]= rver->cv[k];
  for (int k=6;k<COVDIM;++k) cov[k]= 0.0;
  cov[9]=cov[14]=cov[20]=cov[27]=cov[35]=10000.0;
  for (ArteRelationIterator<RREL> irrel = rver->rrel.begin();
       irrel != rver->rrel.end(); ++irrel) {
    particles.push_back(GroverParticle((*irrel)->rtra));
  }
  if (rver->btra != 0) { // associated parent 'track'
    chi2 = rver->btra->chi;
    ax = rver->btra->txf;   ay = rver->btra->tyf;
    p = fabs(rver->btra->pf);
    p = (p>0.0) ? (1.0/p) : 0.0;
    cov[6] = rver->btra->cvf[3];
    cov[7] = rver->btra->cvf[4];
    cov[8] = 0.0;
    cov[9] = rver->btra->cvf[5];
    cov[10]= rver->btra->cvf[6];
    cov[11]= rver->btra->cvf[7];
    cov[12]= 0.0;
    cov[13]= rver->btra->cvf[8];
    cov[14]= rver->btra->cvf[9];
    cov[15]= 0.0; // ! check !
    cov[16]= 0.0; // ! check !
    cov[17]= 0.0;
    cov[18]= 0.0; // ! check !
    cov[19]= 0.0; // ! check !
    cov[20]= (rver->btra->cvf[14])*p*p*p*p; // (1/p -> p)
  }
  for (int k=0;k<CONSTDIM;++k) { // empty constraint map
    for (int l=0;l<MAXTRKS;++l) constr_map[k][l]=0;
    constr_mass[k] = 0.0;
    constr_topps[k]= 0;
  }
}
#endif


#ifndef GROVER_WITHOUT_ARTE
GroverVertex::GroverVertex(const ArtePointer<MVER> mver)
  : index(0), ax(0.), ay(0.), p(0.),
    dm(0.), inlocal(false), inglobal(true),
    wprim(0.), wsec(0.),
    chi2(0.0), fitted(false), mom_estimated(false), n_constraints(0),
    bx(0.0), by(0.0), decay(0.0)
{ // construct from a entry in MVER
  // *** NOT FINISHED!!! ***
  for (int k=0;k<COVDIM;++k) cov[k]=0.0;
  x=mver->x;   y=mver->y;   z=mver->z;
  m   = mver->btra->mass;
  wnr = mver->mcme->wire;
  for (ArteRelationIterator<MTRA> imtra = mver->etra.begin();
       imtra != mver->etra.end(); ++imtra) {
    // check if long lived here...?
    GroverParticle temp(*imtra);
    particles.push_back(temp);
  }
  if (mver->btra != 0){
    p = sqrt(mver->btra->px*mver->btra->px +
	     mver->btra->py*mver->btra->py +
	     mver->btra->pz*mver->btra->pz  );
  }
  else p = 0.0;
  for (int k=0;k<CONSTDIM;++k) {
    for (int l=0;l<MAXTRKS;++l) constr_map[k][l]=0;
    constr_mass[k] = 0.0;
    constr_topps[k]= 0;
  }
}
#endif


/*
  GroverVertex::GroverVertex(int vtindex, GroverParticleList* gpl = 0)
  : index(-1), inlocal(true), inglobal(false), wnr(-1)
  { // Construct from a VT vertex with index vtindex (= 0 to n-1)
  // and a pointer to the corresponding GPL with which VT was 
  // initialised. The track parameters are updated. 
  // BE CAREFUL with the indices!
  if (vtpntr_.vtset[vtindex] == 1) { 
    // vertex not 'switched off' in VT
    int tracksonwire[9] = {0};
    for (int k=0;k<COVDIM;++k) cov[k]=emptycov[k];
    int    nTracks = 0;
    float  tcov[21], vcov[6];
    for (int k=0;k<6 ;++k) vcov[k]=emptycov[k];
    for (int k=0;k<COVDIM;++k) cov[k] = tcov[k] = emptycov[k];
    VtTrackParameters vMomentum;
    VtTCovMatrix      vMomCov;
    int    flag = 0;
    int tracknumber = 0, currentwire = -1;
    // Get vertex coordinates and their covariances
    x = vtkalm_.xvs[vtindex].fX;
    y = vtkalm_.xvs[vtindex].fY;
    z = vtkalm_.xvs[vtindex].fZ;
    tcov[0] = vcov[0] = vtkalm_.cs[vtindex].a[0][0];
    tcov[1] = vcov[1] = vtkalm_.cs[vtindex].a[0][1];
    tcov[2] = vcov[2] = vtkalm_.cs[vtindex].a[1][1];
    tcov[3] = vcov[3] = vtkalm_.cs[vtindex].a[0][2];
    tcov[4] = vcov[4] = vtkalm_.cs[vtindex].a[1][2];
    tcov[5] = vcov[5] = vtkalm_.cs[vtindex].a[2][2];
    // Fill 'particles' with this vertex's tracks
    nTracks = vtpntr_.ntr[vtindex];
    for (int j=0; j<nTracks; ++j) {
      GroverParticle track;
      // get index of original track
      tracknumber = (vtpntr_.ktr[j][vtindex]) - 1; // C++/F77 index nightmare!
      track.SetIndex(tracknumber);
      // track parameters (constrained to start at vertex)
      track.SetBeginSlopeX(vtkalm_.qvs[j][vtindex].ax);
      track.SetOBeginSlopeX(vtkalm_.qvs[j][vtindex].ax);
      track.SetBeginSlopeY(vtkalm_.qvs[j][vtindex].ay);
      track.SetOBeginSlopeY(vtkalm_.qvs[j][vtindex].ay);
      track.SetMomentum(vtkalm_.qvs[j][vtindex].p);
      track.SetBeginX(x);  track.SetBeginY(y);  track.SetBeginZ(z);
      track.SetOBeginX(x); track.SetOBeginY(y); track.SetOBeginZ(z);
      track.SetEndSlopeX(FAILURE); // undefined for a constrained track
      track.SetOEndSlopeX(FAILURE);
      track.SetEndSlopeY(FAILURE);
      track.SetOEndSlopeY(FAILURE);
      track.SetEndX(FAILURE);  track.SetEndY(FAILURE);  track.SetEndZ(FAILURE);
      track.SetOEndX(FAILURE); track.SetOEndY(FAILURE); track.SetOEndZ(FAILURE);
      // Covariance matrix (pos/slope covs are zero at the moment!)
      track.SetBeginCov(9,  vtkalm_.ds[j][vtindex].a[0][0]);
      track.SetOBeginCov(9, vtkalm_.ds[j][vtindex].a[0][0]);
      track.SetBeginCov(13, vtkalm_.ds[j][vtindex].a[0][1]);
      track.SetOBeginCov(13,vtkalm_.ds[j][vtindex].a[0][1]);
      track.SetBeginCov(14, vtkalm_.ds[j][vtindex].a[1][1]);
      track.SetOBeginCov(14,vtkalm_.ds[j][vtindex].a[1][1]);
      for (int k=0; k<6; ++k)  
	{track.SetBeginCov(k, tcov[k]); track.SetOBeginCov(k, tcov[k]);}
      for (int k=6; k<9; ++k)  
	{track.SetBeginCov(k, 0.0); track.SetOBeginCov(k, 0.0);}
      for (int k=10;k<13;++k)  
	{track.SetBeginCov(k, 0.0); track.SetOBeginCov(k, 0.0);}
      currentwire = vttrwi_.iwitr[tracknumber];
      if (currentwire>=0&&currentwire<=8) tracksonwire[currentwire]++;
      // Other stuff direct from original track in main common block VTTRBF
      track.SetMass(vttrbf_.xmass[tracknumber]);

      if (gpl != 0) { 
	// called with a non-null pointer to a GPL
	// pointer to parent object (also resets Type)
	switch ((*gpl)[tracknumber-1].Type()) {
	case 1: // RTRA
	  {track.SetParent((*gpl)[tracknumber-1].ParentRTRA()); break;}
	case 2: // RSEG
	  {track.SetParent((*gpl)[tracknumber-1].ParentRSEG()); break;}
	case 3: // RCCL
	  {track.SetParent((*gpl)[tracknumber-1].ParentRCCL()); break;}
	case 4: // MTRA
	  {track.SetParent((*gpl)[tracknumber-1].ParentMTRA()); break;}
	case 50:// GroverParticle
	  {track.SetParent((*gpl)[tracknumber-1].ParentGP()); break;}
	case 51:// GroverParticleList
	  {track.SetParent((*gpl)[tracknumber-1].ParentGPL()); break;}
	case 52:// GroverVertex
	  {track.SetParent((*gpl)[tracknumber-1].ParentGV()); break;}
	default:// do nothing
	  {break;}
	  break; // for added safety
	}
	track.SetNhit((*gpl)[tracknumber-1].Nhit());
	track.SetCharge((*gpl)[tracknumber-1].Charge());
	track.SetLocal((*gpl)[tracknumber-1].InLocal());
	track.SetGlobal((*gpl)[tracknumber-1].InGlobal());
      }
      else {
	track.SetNhit(-1);
	track.SetCharge(FAILURE);
      }
      track.SetPrimaryWeight(0.0);
      track.SetSecondaryWeight(0.0);
      track.SetOEndCov(emptycov);
      track.SetEndCov(emptycov);
      particles.push_back(track);
    }

    // Associate vertex to wire if >50% of the vertex
    // tracks are on the same wire
    int tracksinvertex = 0;
    for (int k=0; k<9; ++k) tracksinvertex += tracksonwire[k];
    for (int k=0; k<9; ++k) {
      // integer division rounds down
      if (tracksonwire[k] > tracksinvertex/2) wnr = k;
    }
    for (int k=0; k<6; ++k) cov[k] = vcov[k];
    m = 0.0;
    dm = 0.0;
    wprim = 0.0;
    wsec = 0.0;
    chi2 = vtkalm_.chi2[nTracks-1][vtindex]; // C++/F77 indexing nightmare!
    flag = -1;
    vtsum_(vtindex, vMomentum, vMomCov, flag); // call by reference!
    if (flag == 0) {
      p  = vMomentum.moment;
      ax = vMomentum.ax;
      ay = vMomentum.ay;
      // and the covariances? could be tricky. The values in vMomCov
      // are for the covariances at z=0.0, not z = z_vtx so some algebra
      // needed here. not yet. my brain is tired. leave at zero.
      for (int k=6; k<15; ++k)  cov[k] = 0.0;
    }
    else {
      p  = 0.0;
      ax = FAILURE;
      ay = FAILURE;
    }
  }
  else {
    // THIS VERTEX IS BAD IN VT! THE GROVERVERTEX WILL CONTAIN COMPLETE
    // FAILURE!!!
    cout << "*** GroverVertex::GroverVertex(const int& vtindex, const GroverParticleList* gpl) "
	 << endl << "    -> CONSTRUCTOR FAILED - bad argument(s)" << endl;
  }
}
*/


int GroverVertex::Add(const GroverParticle* gp)
{ // add a particle to the internal list. 
  // Please ensure that gp points to something sensible :-)
  // -> Type and parent pointers?
  if (gp == 0) return FAILURE;
  if (inlocal==inglobal) { // empty vertex
    inlocal = gp->InLocal();
    inglobal= !inlocal;
  }
  particles.push_back(*gp);
  return SUCCESS;
}



int GroverVertex::Add(const GroverParticleList* gpl)
{ // add all the particles in another list to the internal list
  // -> Type and parent pointers ???
  if (gpl->empty()) return FAILURE;
  else{
    if (inlocal==inglobal) { // empty vertex
      inlocal = gpl->back().InLocal();
      inglobal= !inlocal;
    }
    for (GroverParticleList::const_iterator ipar = gpl->begin();
	 ipar != gpl->end(); ++ipar) {
      particles.push_back(*ipar);
    }
    return SUCCESS;
  }
}



int GroverVertex::Remove(const int victim)
{ // Remove a particle from the list (numbered 0 to n-1)
  if (this->Size() <= victim || victim < 0) return FAILURE;
  else return (Remove(&particles[victim]));
}


int GroverVertex::Remove(GroverParticle* victim)
{ // Remove a particle from the list
  bool found = false;
  for (GroverParticleList::iterator ip = particles.begin();
       ip != particles.end(); ++ip) {
    if (ip == victim) found = true;
  }
  if (particles.empty() || victim == 0 || !found) return FAILURE;
  else {
    particles.erase(victim);
    ClearConstraints(); // wipe mass constraint array
    fitted = false;
    mom_estimated = false;
    return SUCCESS;
  }
}



int GroverVertex::Size(void) const
{
  if (!this->particles.empty()) 
    return static_cast<int>(this->particles.size()); 
  else return 0;
}




int GroverVertex::GlobalToLocal(void)
{ // Covariance unchanged (not a bad approximation, in fact).
  if (inglobal && !inlocal) {

#ifndef GROVER_WITHOUT_ARTE
    float oldpos[5], newpos[5];
    oldpos[0] = x; oldpos[1] = y; oldpos[2] = z; oldpos[3] = ax; oldpos[4] = ay;
    acGlobalToLocal(newpos, oldpos, 1);
    x = newpos[0]; y = newpos[1]; z = newpos[2]; ax = newpos[3]; ay = newpos[4];
#endif

    if (!particles.empty()) {
      for (GroverParticleList::iterator ipart = particles.begin();
	   ipart != particles.end(); ++ipart) {
	ipart->GlobalToLocal();
      }
    }
    inlocal = true; inglobal = false;
  }
  return SUCCESS;
}



int GroverVertex::LocalToGlobal(void)
{ // Covariance unchanged (not a bad approximation, in fact)
  if (inlocal && !inglobal) {

#ifndef GROVER_WITHOUT_ARTE
    float oldpos[5], newpos[5];
    oldpos[0] = x; oldpos[1] = y; oldpos[2] = z; oldpos[3] = ax; oldpos[4] = ay;
    acLocalToGlobal(oldpos, newpos, 1);
    x = newpos[0]; y = newpos[1]; z = newpos[2]; ax = newpos[3]; ay = newpos[4];
#endif

    if (!particles.empty()) {
      for (GroverParticleList::iterator ipart = particles.begin();
	   ipart != particles.end(); ++ipart) {
	ipart->LocalToGlobal();
      }
    }
    inlocal = false; inglobal = true;
  }
  return SUCCESS;
}



int GroverVertex::EstimateCov(const float scale = 1.0, const float zcov = 0.01) 
{
  // Estimate diagonal vertex covariance matrix using the distribution of the
  // particles in a plane. If z has not been set using SetZ() the vertex
  // will lie at z=0, which may not be what you wanted. This is a 2D method:
  // You must supply the error on z.
  //
  // D.Emeliyanov/I.Kisel
  //
  float cov_x = 0.0, cov_y = 0.0, cov_z = 0.0;
  float npar = static_cast<float>(particles.size());
  if (npar < 0.5) return FAILURE;
  for(GroverParticleList::iterator ipart = particles.begin();
      ipart != particles.end(); ++ipart) {
    ipart->PropagateBeginTo(z);
    cov_x += (ipart->BeginX()-x) * (ipart->BeginX()-x);
    cov_y += (ipart->BeginY()-y) * (ipart->BeginY()-y);
  }
  cov_x /= npar;
  cov_y /= npar;
  cov_z  = zcov;
  cov[0] = cov_x*scale;
  cov[2] = cov_y*scale;
  cov[5] = cov_z*scale;
  cov[9] = cov[14] = 10000;
  cov[1]=cov[3]=cov[4]=cov[6]=cov[7]=cov[8]=
    cov[10]=cov[11]=cov[12]=cov[13]=0.0;
  return SUCCESS;
}


int GroverVertex::UpdateDistances(void)
{ // update dist of each particle with the chi2 distance to the
  // vertex position (after a refit)
  float dist = 0.0;
  float sigma = 0.0, sigsquared=0.0;
  float dxsquared=0.0, dysquared=0.0;
  for (GroverParticleList::iterator ipar = particles.begin();
       ipar != particles.end(); ++ipar) {
    ipar->PropagateBeginTo(z);    
    sigsquared = ipar->BeginCov(0)+ipar->BeginCov(2);
    sigma = (sigsquared>0.0) ? sqrt(sigsquared) : 1e9;
    dxsquared = (ipar->BeginX()-x)*(ipar->BeginX()-x);
    dysquared = (ipar->BeginY()-y)*(ipar->BeginY()-y);
    dist = (sigma<1e9) ? sqrt(fabs(dxsquared+dysquared))/sigma : 1e9;
    ipar->SetDist(dist);
  }
  return SUCCESS;
}




float GroverVertex::DistanceTo(const GroverParticle* that) const 
{ // distance of closest approach of a particle to the vertex
  return that->DistanceTo(this);
}




float GroverVertex::Chi2DistTo(const GroverParticle* that) const 
{ // Chi2 distance of closest approach of a particle to the vertex
  return that->Chi2DistTo(this);
}




float GroverVertex::DistanceTo(const GroverVertex* that) const 
{ // spatial distance to another vertex
  double dx = this->X() - that->X();
  double dy = this->Y() - that->Y();
  double dz = this->Z() - that->Z();
  return static_cast<float>(sqrt(dx*dx + dy*dy + dz*dz));
}




float GroverVertex::Chi2DistTo(const GroverVertex* that) const  
{ // Chi2 distance to a vertex
  double Dx = this->X() - that->X();
  double Dy = this->Y() - that->Y();
  double Dz = this->Z() - that->Z();
  double sum = (Dx*Dx / (this->Cov(0)+that->Cov(0)) ) +
               (Dy*Dy / (this->Cov(2)+that->Cov(2)) ) +
               (Dz*Dz / (this->Cov(5)+that->Cov(5)) );
  return (sum >= 0.0) ? static_cast<float>(sqrt(sum)) : 1e9;
}




float GroverVertex::DistanceTo(const GroverWire* wire) const 
{
  float space=1e9;
  if (wire->Horizontal() == 1) {
    float dy = this->Y() - wire->Y();
    float dz = this->Z() - wire->Z();
    space = sqrt(dy*dy + dz*dz);
  }
  else if (wire->Vertical() == 1) {
    float dx = this->X() - wire->X();
    float dz = this->Z() - wire->Z();
    space = sqrt(dx*dx + dz*dz);
  }
  return space;
}




float GroverVertex::Chi2DistTo(const GroverWire* wire) const 
{ 
  double sum = 0.0;
  if (wire->Horizontal() == 1) {
    double dy  = this->Y() - wire->Y();
    double dz  = this->Z() - wire->Z();
    double sy2 = this->Cov(2);
    double sz2 = this->Cov(5);
    sum = dy*dy/sy2 + dz*dz/sz2;
  }
  else if (wire->Vertical() == 1) {
    double dx  = this->X() - wire->X();
    double dz  = this->Z() - wire->Z();
    double sx2 = this->Cov(0);
    double sz2 = this->Cov(5);
    sum = dx*dx/sx2 + dz*dz/sz2;
  }
  return (sum >= 0.0) ? static_cast<float>(sqrt(sum)) : 1e9;
}





float GroverVertex::DistanceTo(const float xx, const float yy, const float zz) const 
{
  // 24.07.00 Vika Kekelidze
  double dx = xx - this->X();
  double dy = yy - this->Y();
  double dz = zz - this->Z();
  return static_cast<float>(sqrt(dx*dx+dy*dy+dz*dz));
}




float GroverVertex::Chi2DistTo(const float xx, const float yy, const float zz) const 
{
  double Dx = xx - this->X();
  double Dy = yy - this->Y();
  double Dz = zz - this->Z();
  double sum = Dx*Dx/this->Cov(0) +
               Dy*Dy/this->Cov(2) +
               Dz*Dz/this->Cov(5);
  return (sum >= 0.0) ? static_cast<float>(sqrt(sum)) : 1e9;
}




int GroverVertex::ResetParticles(void)
{
  int cumerr = SUCCESS;
  for (GroverParticleList::iterator ipar = particles.begin();
       ipar != particles.end(); ++ipar)
    cumerr += ipar->ResetParameters();
  fitted = false;
  mom_estimated = false;
  return cumerr;
}




int GroverVertex::ClearConstraints(void)
{ // wipe constraint array
  n_constraints = 0;
  for (int i=0; i<CONSTDIM; ++i) {
    for (int j=0; j<MAXTRKS; ++j) constr_map[i][j]=0;
    constr_mass[i] = 0.0;
#ifndef GROVER_WITHOUT_ARTE
    constr_topps[i]= 0;
#endif
  }
  return SUCCESS;
}



#ifndef GROVER_WITHOUT_ARTE
int GroverVertex::AddConstraint(const ArtePointer<TOPP> hyp, 
				const int p1,    const int p2, 
				const int p3=-1, const int p4=-1, 
				const int p5=-1, const int p6=-1, 
				const int p7=-1, const int p8=-1, 
				const int p9=-1, const int p10=-1)
{ // Add a new mass constraint and flag the particles taking part
  if (hyp==0) return FAILURE;
  float mass = hyp->mass;
  if (mass<=0.0) return FAILURE;
  int err = AddConstraint(mass,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10);
  if (err==0 && n_constraints>0 && n_constraints<=CONSTDIM)
    constr_topps[n_constraints-1] = hyp;
  return err;
}
#endif



#ifndef GROVER_WITHOUT_ARTE
int GroverVertex::AddConstraint(const ArtePointer<TOPP> hyp,
				const GroverParticle* p1,   const GroverParticle* p2,
				const GroverParticle* p3=0, const GroverParticle* p4=0, 
				const GroverParticle* p5=0, const GroverParticle* p6=0, 
				const GroverParticle* p7=0, const GroverParticle* p8=0, 
				const GroverParticle* p9=0, const GroverParticle* p10=0)
{ // Add a new mass constraint and flag the particles taking part
  if (hyp==0) return FAILURE;
  float mass = hyp->mass;
  if (mass<=0.0) return FAILURE;
  int err = AddConstraint(mass,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10);
  if (err==0 && n_constraints>0 && n_constraints<=CONSTDIM)
    constr_topps[n_constraints-1] = hyp;
  return err;
}
#endif



int GroverVertex::AddConstraint(const float mass, 
				const int p1,    const int p2, 
				const int p3=-1, const int p4=-1, 
				const int p5=-1, const int p6=-1, 
				const int p7=-1, const int p8=-1, 
				const int p9=-1, const int p10=-1)
{ // Add a new mass constraint and flag the particles taking part
  int ntr = particles.size();
  if (
    mass<=0.0 || ntr<2 || ntr>MAXTRKS || p1==p2 || n_constraints >= CONSTDIM
      ) return FAILURE;
  if (p1>=0  &&  p1<ntr) constr_map[n_constraints][p1] = 1;
  if (p2>=0  &&  p2<ntr) constr_map[n_constraints][p2] = 1;
  if (p3>=0  &&  p3<ntr) constr_map[n_constraints][p3] = 1;
  if (p4>=0  &&  p4<ntr) constr_map[n_constraints][p4] = 1;
  if (p5>=0  &&  p5<ntr) constr_map[n_constraints][p5] = 1;
  if (p6>=0  &&  p6<ntr) constr_map[n_constraints][p6] = 1;
  if (p7>=0  &&  p7<ntr) constr_map[n_constraints][p7] = 1;
  if (p8>=0  &&  p8<ntr) constr_map[n_constraints][p8] = 1;
  if (p9>=0  &&  p9<ntr) constr_map[n_constraints][p9] = 1;
  if (p10>=0 && p10<ntr) constr_map[n_constraints][p10]= 1;
  constr_mass[n_constraints] = mass;
  ++n_constraints;
  return SUCCESS;
}


int GroverVertex::AddConstraint(const float mass, 
				const GroverParticle* p1,   const GroverParticle* p2, 
				const GroverParticle* p3=0, const GroverParticle* p4=0, 
				const GroverParticle* p5=0, const GroverParticle* p6=0, 
				const GroverParticle* p7=0, const GroverParticle* p8=0, 
				const GroverParticle* p9=0, const GroverParticle* p10=0)
{ // Add a new mass constraint and flag the particles taking part
  // Safe with invalid arguments since they are never de-referenced. :-)
  if (mass<=0.0 || this->Size()<2 || 
      p1==0 || p2==0 || p1==p2 ||
      n_constraints >= CONSTDIM) return FAILURE;
  int ctr = 0;
  for (GroverParticleList::iterator ipar = particles.begin();
       ipar != particles.end(); ++ipar) {
    if (ipar == p1 || ipar == p2 || ipar == p3 || ipar == p4 || ipar == p5 ||
	ipar == p6 || ipar == p7 || ipar == p8 || ipar == p9 || ipar == p10
	) constr_map[n_constraints][ctr]=1;
    ++ctr;
  }
  constr_mass[n_constraints] = mass;
  ++n_constraints;
  return SUCCESS;
}




void GroverVertex::dump_cons(void) const
{ // just for testing MS 26/09/00
  cout << "======================= Constraint Map ================================" 
       <<endl;
  cout << "Number of constraints = " << n_constraints<<endl;
  for (int i=0; i<CONSTDIM; ++i) {
    cout <<"constraint "<<i<<": ";
    for (int j=0; j<MAXTRKS; ++j) cout << constr_map[i][j] <<" ";
    cout << "mass="<<constr_mass[i]
#ifndef GROVER_WITHOUT_ARTE
	 <<", <TOPP> ="<<constr_topps[i]
#endif
	 <<endl;
  }
  cout << "=======================================================================" 
       <<endl;
}



int GroverVertex::Ndof(void) const
{
  if (!fitted) return FAILURE;
  int ndof = 2*this->Size() - 3 + this->Nconstraints();
  if (ndof<1) return FAILURE;
  else return ndof;
}



int GroverVertex::Fit(void)
{ 
  //
  // This is the main secondary fitting function. All
  // other fitting routines (GroverConstrainedFit(vtx),
  // ver.Fit(mass) etc.) just set the constraint map
  // correctly and then call this function.
  //
  // D.Emeliyanov 03/10/00
  // 
  static bool verbose = false;//(groverCuts.Verbose == 1);
  const int ROVERMAXTRACKS = 50;
  if (this->Size() >= ROVERMAXTRACKS || this->Size() < 2) return FAILURE;

  bool MC = false;
  bool tampered = false;
  for (GroverParticleList::const_iterator ip = this->particles.begin();
       ip != this->particles.end(); ++ip) {
    if (ip->Type() == 4) {MC = true; break;}
    if (ip->Tampered()) tampered = true;
  }
  if (MC) return FAILURE; // can't fit a vertex with > 0 perfect tracks!

  // Reset all tracks UNLESS the user has been fiddling with them
  if (!tampered) this->ResetParticles();

  if (!trashcan.empty()) trashcan.clear();

  typedef struct
  {
    double Zcov[5][5];
    double Pcov,mass,p,ax,ay;
    double dzeta[5];
    int flag;
    float dchi2;
    int constr[CONSTDIM]; // --> CONSTRAINTS: set to 1 if used in constraint i,
                          // 0 otherwise. CONSTDIM set in class definition.
  } TRACK_TYPE, *pTRACK_TYPE;
 
  TRACK_TYPE atr[ROVERMAXTRACKS];
  
  //  --> CONSTRAINTS
  // constraint masses are available in 
  // constr_mass[0], constr_mass[1], ... constr_mass[n_contraints-1]
  const int   MaxIter=5;
  const float sigma=0.00001;
  
  double Gk[3*ROVERMAXTRACKS+3][3*ROVERMAXTRACKS+3];
  double Rk[3*ROVERMAXTRACKS+3];
  double Kk[3*ROVERMAXTRACKS+3][5];
  double Hm[3*ROVERMAXTRACKS+3];
  double Ak[5][3*ROVERMAXTRACKS+3];
  double L[5][5], L_1[5][5];
  double dchi2t,Mest,Merr;
  
  /////////////////
  float zref=0.0;// <- note!
  /////////////////
  
  double zeta[5];
  double Px,Py,Pz,nx,ny,nz,P,E,epsilon;
  int i,j,k,l,ndim,jjj,jcons;
  
  GroverParticleList::iterator trackincluster[ROVERMAXTRACKS] ={0};
  
  i=1;
  for (GroverParticleList::iterator it1 = particles.begin(); 
       it1 != particles.end(); ++it1) {
    trackincluster[i] = it1;
    it1->PropagateBeginTo(zref);
    ++i;
  }
  
  int ntrincluster = i-1;
  ndim=3+3*ntrincluster;
  
  for (i=1; i<=ntrincluster; ++i) {
    GroverParticleList::iterator it = trackincluster[i];
    for(int v=0;v<5;++v) for(int vv=0;vv<5;++vv) atr[i].Zcov[v][vv]=0.0;
    atr[i].Zcov[0][0] =  it->BeginCov(0);
    atr[i].Zcov[1][0] = atr[i].Zcov[0][1] =  it->BeginCov(1);
    atr[i].Zcov[2][0] = atr[i].Zcov[0][2] =  it->BeginCov(3);
    atr[i].Zcov[3][0] = atr[i].Zcov[0][3] =  it->BeginCov(6);
    atr[i].Zcov[2][1] = atr[i].Zcov[1][2] =  it->BeginCov(4);
    atr[i].Zcov[3][1] = atr[i].Zcov[1][3] =  it->BeginCov(7);
    atr[i].Zcov[1][1] = it->BeginCov(2);
    atr[i].Zcov[2][2] = it->BeginCov(5);
    atr[i].Zcov[2][3] = atr[i].Zcov[3][2] =  it->BeginCov(8);
    atr[i].Zcov[3][3] = it->BeginCov(9);
    atr[i].Zcov[4][4] = it->BeginCov(14);
    atr[i].mass       = it->Mass();
    atr[i].dzeta[0]   = it->BeginX();
    atr[i].dzeta[1]   = it->BeginY();
    atr[i].dzeta[2]   = it->BeginSlopeX();
    atr[i].dzeta[3]   = it->BeginSlopeY();
    atr[i].dzeta[4]   = it->Momentum();
    atr[i].ax         = it->BeginSlopeX();
    atr[i].ay         = it->BeginSlopeY();
    atr[i].p          = it->Momentum();
    atr[i].flag       = 1;
    //  --> CONSTRAINTS
    for (int j=0; j<CONSTDIM; ++j) 
      atr[i].constr[j] = constr_map[j][i-1];
    Rk[3*i]=atr[i].ax;   Rk[3*i+1]=atr[i].ay;   Rk[3*i+2]=atr[i].p;
  }
  
  Rk[0]= this->X();   Rk[1]=this->Y();  Rk[2]=this->Z();
  
  for (jjj=1;jjj<=MaxIter;++jjj) {//MaxIter
    for(i=0;i<ndim;++i)  
      for(j=0;j<ndim;++j)  
	Gk[i][j]=0.0;    
    Gk[0][0]=100.0;//cov[0];  
    Gk[1][1]=100.0;//cov[5];
    Gk[2][2]=100.0;//cov[9];
    
    for(i=1;i<=ntrincluster;++i) {
      Gk[3*i][3*i]    = 1.0; 
      Gk[3*i+1][3*i+1]= 1.0; 
      Gk[3*i+2][3*i+2]= 100.0;
    }
    
    for(i=1;i<=ntrincluster;++i) {
      L[0][0] = Gk[0][0] + Rk[3*i]*Rk[3*i]*Gk[2][2] + Rk[2]*Rk[2]*Gk[3*i][3*i]
	      - 2.0*(Rk[3*i]*Gk[0][2] + 
		     Rk[2]*Gk[0][3*i] - 
		     Rk[3*i]*Rk[2]*Gk[2][3*i]);
      L[1][1] = Gk[1][1] + Rk[3*i+1]*Rk[3*i+1]*Gk[2][2]
               +Rk[2]*Rk[2]*Gk[3*i+1][3*i+1]
	       -2.0*(Rk[3*i+1]*Gk[1][2] +
		     Rk[2]*Gk[1][3*i+1] - 
		     Rk[2]*Rk[3*i+1]*Gk[2][3*i+1]);
      L[0][1] = Gk[0][1] + Rk[3*i]*Rk[3*i+1]*Gk[2][2]
	       +Rk[2]*Rk[2]*Gk[3*i][3*i+1]
	       -Rk[2]*(Gk[1][3*i] + Gk[0][3*i+1])
	       +Rk[2]*(Rk[3*i]*Gk[2][3*i+1]+Rk[3*i+1]*Gk[2][3*i])
	       -Rk[3*i]*Gk[1][2] 
               -Rk[3*i+1]*Gk[0][2];
      L[1][0] = L[0][1];
      L[2][2] = Gk[3*i][3*i];
      L[3][3] = Gk[3*i+1][3*i+1];
      L[2][3] = Gk[3*i][3*i+1];
      L[3][2] = L[2][3];
      L[0][2] = Gk[0][3*i]   -Rk[3*i]*Gk[2][3*i]    -Rk[2]*Gk[3*i][3*i];
      L[2][0] = L[0][2];
      L[0][3] = Gk[0][3*i+1] -Rk[3*i]*Gk[2][3*i+1]  -Rk[2]*Gk[3*i][3*i+1];
      L[3][0] = L[0][3];
      L[1][2] = Gk[1][3*i]   -Rk[3*i+1]*Gk[2][3*i]  -Rk[2]*Gk[3*i][3*i+1];
      L[2][1] = L[1][2];
      L[1][3] = Gk[1][3*i+1] -Rk[3*i+1]*Gk[2][3*i+1]-Rk[2]*Gk[3*i+1][3*i+1];
      L[3][1] = L[1][3];
      L[0][4] = Gk[0][3*i+2] -Rk[3*i]*Gk[2][3*i+2]  -Rk[2]*Gk[3*i][3*i+2];
      L[4][0] = L[0][4];
      L[1][4] = Gk[1][3*i+2] -Rk[3*i+1]*Gk[2][3*i+2]-Rk[2]*Gk[3*i+1][3*i+2];
      L[4][1] = L[1][4];
      L[2][4] = Gk[3*i][3*i+2];
      L[4][2] = L[2][4];
      L[3][4] = Gk[3*i+1][3*i+2];
      L[4][3] = L[3][4];
      L[4][4] = Gk[3*i+2][3*i+2];
      
      for(k=0;k<5;++k)  
	for(l=0;l<5;++l) {
	  L[k][l]   = L[k][l] + atr[i].Zcov[k][l];
	  L_1[k][l] = L[k][l];
	}
      
      five_dim_inv(L_1);
	
      for(j=0;j<ndim;++j) {
	Ak[0][j] = Gk[j][0] - Rk[3*i]*Gk[j][2]   - Rk[2]*Gk[j][3*i];
	Ak[1][j] = Gk[1][j] - Rk[3*i+1]*Gk[j][2] - Rk[2]*Gk[j][3*i+1];
	Ak[2][j] = Gk[j][3*i];
	Ak[3][j] = Gk[j][3*i+1];
	Ak[4][j] = Gk[j][3*i+2];
	for(int v=0;v<5;++v) {
	  Kk[j][v] = 0.0;
	  for(int vv=0;vv<5;++vv) Kk[j][v] += Ak[vv][j]*L_1[vv][v];
	}
      }
      zeta[0] = atr[i].dzeta[0] - Rk[0] + Rk[3*i]*Rk[2];
      zeta[1] = atr[i].dzeta[1] - Rk[1] + Rk[3*i+1]*Rk[2];
      zeta[2] = atr[i].dzeta[2] - Rk[3*i];
      zeta[3] = atr[i].dzeta[3] - Rk[3*i+1];
      zeta[4] = atr[i].dzeta[4] - Rk[3*i+2];
      for(j=0;j<ndim;++j) 
	for(int v=0;v<5;++v) 
	  Rk[j] += Kk[j][v]*zeta[v];
      for(k=0;k<ndim;++k) 
	for(l=0;l<ndim;++l) 
	  for(int v=0;v<5;v++)
	    Gk[k][l] -= Kk[k][v]*Ak[v][l];
    }
    
    if (n_constraints) {
      for (jcons=0;jcons<n_constraints; ++jcons) {
	for (i=0;i<ndim;++i) Hm[i]=0.0;
	Px=0.0;Py=0.0;   Pz=0.0;   E=0.0;
	for(i=1;i<=ntrincluster;++i) 
	  if(atr[i].constr[jcons]) {
	    nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
	    nx = Rk[3*i]*nz;
	    ny = Rk[3*i+1]*nz;
	    Px += Rk[3*i+2]*nx;
	    Py += Rk[3*i+2]*ny;
	    Pz += Rk[3*i+2]*nz;
	    E  += sqrt(Rk[3*i+2]*Rk[3*i+2]+atr[i].mass*atr[i].mass);
	  }    
	P = sqrt(Px*Px+Py*Py+Pz*Pz);
	for (i=1;i<=ntrincluster;++i) 
	  if (atr[i].constr[jcons]) {
	    nz = 1.0 / sqrt(1.0 + Rk[3*i]*Rk[3*i] + Rk[3*i+1]*Rk[3*i+1]);
	    nx = Rk[3*i]*nz;
	    ny = Rk[3*i+1]*nz;
	    Hm[3*i]  = -2.0*Rk[3*i+2]*nz*((1.0-nx*nx)*Px-nx*ny*Py-nx*nz*Pz);
	    Hm[3*i+1]= -2.0*Rk[3*i+2]*nz*(-nx*ny*Px+(1.0-ny*ny)*Py-ny*nz*Pz);
	    Hm[3*i+2]=  2.0*Rk[3*i+2]
	                * E/sqrt(Rk[3*i+2]*Rk[3*i+2] + atr[i].mass*atr[i].mass)
	                - 2*(nx*Px + ny*Py + nz*Pz);
	  }
	epsilon = -E*E + P*P + constr_mass[jcons]*constr_mass[jcons];
	if (verbose) cout<<"epsilon = "<<epsilon<<endl;
	for (i=0;i<ndim;++i) {
	  Ak[0][i] = 0.0;
	  for (j=0;j<ndim;++j) Ak[0][i] += Hm[j]*Gk[j][i];
	}
	L[0][0] = 0.0;
	for (i=0;i<ndim;++i) L[0][0] += Ak[0][i]*Hm[i];
	L[0][0] += sigma*sigma;
	L_1[0][0] = 1.0/L[0][0];
	for (i=0;i<ndim;++i) Kk[i][0] = Ak[0][i]*L_1[0][0];
	for (i=0;i<ndim;++i) Rk[i]   += Kk[i][0]*epsilon;
	for (i=0;i<ndim;++i)
	  for (j=0;j<ndim;++j) Gk[i][j] -= Ak[0][i]*Kk[j][0];
      }
    }
  }

  if(n_constraints == 0) {
    for (i=0;i<ndim;++i) Hm[i]=0.0;
    Px = 0.0;   Py = 0.0;   Pz = 0.0;   E = 0.0;
    for (i=1;i<=ntrincluster;++i) {
      nz =  1.0 / sqrt(1.0 + Rk[3*i]*Rk[3*i] + Rk[3*i+1]*Rk[3*i+1]);
      nx =  Rk[3*i]*nz;
      ny =  Rk[3*i+1]*nz;
      Px += Rk[3*i+2]*nx;
      Py += Rk[3*i+2]*ny;
      Pz += Rk[3*i+2]*nz;
      E  += sqrt(Rk[3*i+2]*Rk[3*i+2] + atr[i].mass*atr[i].mass);
    }
    P = sqrt(Px*Px + Py*Py + Pz*Pz);
    for (i=1;i<=ntrincluster;++i) {
      nz = 1.0 / sqrt(1.0 + Rk[3*i]*Rk[3*i] + Rk[3*i+1]*Rk[3*i+1]);
      nx = Rk[3*i]*nz;
      ny = Rk[3*i+1]*nz;
      Hm[3*i]  = -2.0*Rk[3*i+2]*nz*((1.0-nx*nx)*Px
				    -nx*ny*Py
				    -nx*nz*Pz      );
      Hm[3*i+1]= -2.0*Rk[3*i+2]*nz*(-nx*ny*Px
				    +(1.0-ny*ny)*Py
				    -ny*nz*Pz      );
      Hm[3*i+2]=   2.0*Rk[3*i+2]
	           *E/sqrt(Rk[3*i+2]*Rk[3*i+2] + atr[i].mass*atr[i].mass)
	         - 2*(nx*Px + ny*Py + nz*Pz);
    }
    for (i=0;i<ndim;++i) {
      Ak[0][i] = 0.0;
      for (j=0;j<ndim;++j) Ak[0][i] += Hm[j]*Gk[j][i];
    }
    L[0][0] = 0.0;
    for (i=0;i<ndim;++i) L[0][0] += Ak[0][i]*Hm[i];
    L_1[0][0] = 1.0/L[0][0];
    Mest = sqrt(E*E-P*P);
    Merr = 0.5 / sqrt(L_1[0][0]*(E*E-P*P));
    if (verbose) cout<<"Fitted mass = "<<Mest<<" Error = "<<Merr<<endl;
    this->SetMass(Mest);
    this->SetMassError(Merr);
  }
  else {
    this->SetMass(constr_mass[0]);
    this->SetMassError(sigma);
  }
  
  
  for (i=1;i<=ntrincluster;++i) {
    L[0][0]= Gk[0][0] 
            +Rk[3*i]*Rk[3*i]*Gk[2][2] 
            +Rk[2]*Rk[2]*Gk[3*i][3*i]
            -2.0*( Rk[3*i]*Gk[0][2] 
		  +Rk[2]*Gk[0][3*i] 
		  -Rk[3*i]*Rk[2]*Gk[2][3*i] );
    L[1][1]= Gk[1][1]
            +Rk[3*i+1]*Rk[3*i+1]*Gk[2][2]
            +Rk[2]*Rk[2]*Gk[3*i+1][3*i+1]
            -2.0*( Rk[3*i+1]*Gk[1][2]
                  +Rk[2]*Gk[1][3*i+1]
		  -Rk[2]*Rk[3*i+1]*Gk[2][3*i+1] );
    L[0][1]= Gk[0][1]
            +Rk[3*i]*Rk[3*i+1]*Gk[2][2]
            +Rk[2]*Rk[2]*Gk[3*i][3*i+1]
            -Rk[2]*( Gk[1][3*i] 
		    +Gk[0][3*i+1] )
            +Rk[2]*( Rk[3*i]*Gk[2][3*i+1] 
		    +Rk[3*i+1]*Gk[2][3*i] )
            -Rk[3*i]*Gk[1][2] 
            -Rk[3*i+1]*Gk[0][2];
    L[2][2]= Gk[3*i][3*i];
    L[3][3]= Gk[3*i+1][3*i+1];
    L[2][3]= Gk[3*i][3*i+1];
    L[0][2]= Gk[0][3*i]
            -Rk[3*i]*Gk[2][3*i]
            -Rk[2]*Gk[3*i][3*i];
    L[0][3]= Gk[0][3*i+1]
            -Rk[3*i]*Gk[2][3*i+1]
            -Rk[2]*Gk[3*i][3*i+1];
    L[1][2]= Gk[1][3*i]
            -Rk[3*i+1]*Gk[2][3*i]
            -Rk[2]*Gk[3*i][3*i+1];
    L[1][3]= Gk[1][3*i+1]
            -Rk[3*i+1]*Gk[2][3*i+1]
            -Rk[2]*Gk[3*i+1][3*i+1];
    L[0][4]= Gk[0][3*i+2]
            -Rk[3*i]*Gk[2][3*i+2]
            -Rk[2]*Gk[3*i][3*i+2];
    L[1][4]= Gk[1][3*i+2]
            -Rk[3*i+1]*Gk[2][3*i+2]
            -Rk[2]*Gk[3*i+1][3*i+2];
    L[2][4]= Gk[3*i][3*i+2];
    L[3][4]= Gk[3*i+1][3*i+2];
    L[4][4]= Gk[3*i+2][3*i+2];    
    L[1][0] = L[0][1];
    L[3][2] = L[2][3];
    L[2][0] = L[0][2];
    L[3][0] = L[0][3];
    L[2][1] = L[1][2];
    L[3][1] = L[1][3];  
    L[4][0] = L[0][4];
    L[4][1] = L[1][4];
    L[4][2] = L[2][4];
    L[4][3] = L[3][4];
    for (k=0;k<5;++k)  
      for (l=0;l<5;++l)  
	L_1[k][l] = L[k][l]+atr[i].Zcov[k][l];
    five_dim_inv(L_1);
    zeta[0] = atr[i].dzeta[0] - Rk[0] + Rk[3*i]*Rk[2];
    zeta[1] = atr[i].dzeta[1] - Rk[1] + Rk[3*i+1]*Rk[2];
    zeta[2] = atr[i].dzeta[2] - Rk[3*i];
    zeta[3] = atr[i].dzeta[3] - Rk[3*i+1];
    zeta[4] = atr[i].dzeta[4] - Rk[3*i+2];
    atr[i].dchi2 = 0.0;
    for(int v=0;v<5;++v) 
      for(int vv=0;vv<5;++vv) 
	atr[i].dchi2 += zeta[v]*zeta[vv]*L_1[v][vv];
  }

  if (verbose) cout<<"ntr = "<<ntrincluster<<endl;  
  this->SetX(Rk[0]);   this->SetY(Rk[1]);   this->SetZ(Rk[2]);
  this->SetCov(0, Gk[0][0]);
  this->SetCov(1, Gk[0][1]);
  this->SetCov(2, Gk[1][1]);
  this->SetCov(3, Gk[0][2]);     
  this->SetCov(4, Gk[1][2]);
  this->SetCov(5, Gk[2][2]);
  
  if (verbose) 
    for(i=1;i<=ntrincluster;++i) 
      cout<<"CONSTR.FIT: Chi2 contrib. from track "<<i<<" = " <<atr[i].dchi2<<endl;
  
  dchi2t = 0.0;
  for (i=1;i<=ntrincluster;++i) dchi2t += atr[i].dchi2;
  
  //dchi2t/=(n_constraints+2.0*ntrincluster-3.0); // no normalisation now!
  
  if (verbose) cout<<"Total chi2 of the fit = "<<dchi2t<<endl;
  this->SetChi2(dchi2t);
  
  // fill updated info for each track
  for (i=1; i<=ntrincluster; ++i) {
    GroverParticleList::iterator it = trackincluster[i];
    atr[i].ax = Rk[3*i];
    atr[i].ay = Rk[3*i+1];
    atr[i].p  = Rk[3*i+2];
    it->PropagateBeginTo(this->Z());
    it->xb=this->X(); 
    it->yb=this->Y();
    it->axb=atr[i].ax;
    it->ayb=atr[i].ay;
    it->p=atr[i].p;
    it->SetChi2(atr[i].dchi2);
  }
  
  gcov.clear();  
  for(int i=0; i<ndim; ++i) {
    for(int j=0; j<=i; ++j) {
      gcov.push_back(Gk[i][j]);
    }
  }

  fitted = true;
  return SUCCESS;
}

int GroverVertex::NewGeoFit(void)
{ 
  //
  // This is the main primary fitting function for the new
  // primary vertex finder.
  //
  // D.Emeliyanov 05/18/02
  // 

  static bool verbose = false;//(groverCuts.Verbose == 1);
  const int ROVERMAXTRACKS = 80;
  if (this->Size() >= ROVERMAXTRACKS || this->Size() < 2) return FAILURE;

  bool MC = false;
  bool tampered = false;
  for (GroverParticleList::const_iterator ip = this->particles.begin();
       ip != this->particles.end(); ++ip) {
    if (ip->Type() == 4) {MC = true; break;}
    if (ip->Tampered()) tampered = true;
  }
  if (MC) return FAILURE; // can't fit a vertex with > 0 perfect tracks!

  // Reset all tracks UNLESS the user has been fiddling with them
  if (!tampered) this->ResetParticles();

  if (!trashcan.empty()) trashcan.clear();

  typedef struct
  {
    double Zcov[4][4];
    double ax,ay;
    double dzeta[4];
    int flag;
    float dchi2;
  } TRACK_TYPE, *pTRACK_TYPE;
 
  TRACK_TYPE atr[ROVERMAXTRACKS];
  
  const int   MaxIter=3;
  
  double Gk[2*ROVERMAXTRACKS+3][2*ROVERMAXTRACKS+3];
  double Rk[2*ROVERMAXTRACKS+3];
  double Kk[2*ROVERMAXTRACKS+3][4];
  double Ak[4][3*ROVERMAXTRACKS+3];
  double L[4][4], L_1[4][4];
  double dchi2t;
  
  /////////////////
  float zref=0.0;// <- note!
  /////////////////
  
  double zeta[4];
  int i,j,k,l,ndim,jjj;
  
  GroverParticleList::iterator trackincluster[ROVERMAXTRACKS] ={0};
  
  i=1;
  for (GroverParticleList::iterator it1 = particles.begin(); 
       it1 != particles.end(); ++it1) {
    trackincluster[i] = it1;
    it1->PropagateBeginTo(zref);
    ++i;
  }
  
  int ntrincluster = i-1;
  ndim=3+2*ntrincluster;
  
  for (i=1; i<=ntrincluster; ++i) {
    GroverParticleList::iterator it = trackincluster[i];
    for(int v=0;v<4;++v) for(int vv=0;vv<4;++vv) atr[i].Zcov[v][vv]=0.0;
    atr[i].Zcov[0][0] =  it->BeginCov(0);
    atr[i].Zcov[1][0] = atr[i].Zcov[0][1] =  it->BeginCov(1);
    atr[i].Zcov[2][0] = atr[i].Zcov[0][2] =  it->BeginCov(3);
    atr[i].Zcov[3][0] = atr[i].Zcov[0][3] =  it->BeginCov(6);
    atr[i].Zcov[2][1] = atr[i].Zcov[1][2] =  it->BeginCov(4);
    atr[i].Zcov[3][1] = atr[i].Zcov[1][3] =  it->BeginCov(7);
    atr[i].Zcov[1][1] = it->BeginCov(2);
    atr[i].Zcov[2][2] = it->BeginCov(5);
    atr[i].Zcov[2][3] = atr[i].Zcov[3][2] =  it->BeginCov(8);
    atr[i].Zcov[3][3] = it->BeginCov(9);
    atr[i].dzeta[0]   = it->BeginX();
    atr[i].dzeta[1]   = it->BeginY();
    atr[i].dzeta[2]   = it->BeginSlopeX();
    atr[i].dzeta[3]   = it->BeginSlopeY();
    atr[i].ax         = it->BeginSlopeX();
    atr[i].ay         = it->BeginSlopeY();
    atr[i].flag       = 1;
    Rk[1+2*i]=atr[i].ax;   Rk[2+2*i]=atr[i].ay;
  }
  
  Rk[0]= this->X();   Rk[1]=this->Y();  Rk[2]=this->Z();
  
  for (jjj=1;jjj<=MaxIter;++jjj) {//MaxIter
    for(i=0;i<ndim;++i)  
      for(j=0;j<ndim;++j)  
	Gk[i][j]=0.0;    
    Gk[0][0]=100.0;//cov[0];  
    Gk[1][1]=100.0;//cov[5];
    Gk[2][2]=100.0;//cov[9];
    
    for(i=1;i<=ntrincluster;++i) {
      Gk[1+2*i][1+2*i]    = 1.0; 
      Gk[2+2*i][2+2*i]= 1.0; 
    }
    
    for(i=1;i<=ntrincluster;++i) {
      L[0][0]=Gk[0][0]+Rk[1+2*i]*Rk[1+2*i]*Gk[2][2]+Rk[2]*Rk[2]*
	Gk[1+2*i][1+2*i]-2.0*(Rk[1+2*i]*Gk[0][2]+Rk[2]*Gk[0][1+2*i]- 
			      Rk[1+2*i]*Rk[2]*Gk[2][1+2*i]);
      L[1][1]=Gk[1][1]+Rk[2+2*i]*Rk[2+2*i]*Gk[2][2]+Rk[2]*Rk[2]*
	Gk[2+2*i][2+2*i]-2.0*(Rk[2+2*i]*Gk[1][2]+Rk[2]*Gk[1][2+2*i]- 
			      Rk[2]*Rk[2+2*i]*Gk[2][2+2*i]);
      L[0][1]=Gk[0][1]+Rk[1+2*i]*Rk[2+2*i]*Gk[2][2]+Rk[2]*Rk[2]*
	Gk[1+2*i][2+2*i]-Rk[2]*(Gk[1][1+2*i]+Gk[0][2+2*i])
	+Rk[2]*(Rk[1+2*i]*Gk[2][2+2*i]+Rk[2+2*i]*Gk[2][1+2*i])
	-Rk[1+2*i]*Gk[1][2]-Rk[2+2*i]*Gk[0][2];
      L[1][0]=L[0][1];
      L[2][2]=Gk[1+2*i][1+2*i];
      L[3][3]=Gk[2+2*i][2+2*i];
      L[2][3]=Gk[1+2*i][2+2*i];
      L[3][2]=L[2][3];
      L[0][2]=Gk[0][1+2*i]-Rk[1+2*i]*Gk[2][1+2*i]-Rk[2]*Gk[1+2*i][1+2*i];
      L[2][0]=L[0][2];
      L[0][3]=Gk[0][2+2*i]-Rk[1+2*i]*Gk[2][2+2*i]-Rk[2]*Gk[1+2*i][2+2*i];
      L[3][0]=L[0][3];
      L[1][2]=Gk[1][1+2*i]-Rk[2+2*i]*Gk[2][1+2*i]-Rk[2]*Gk[1+2*i][2+2*i];
      L[2][1]=L[1][2];
      L[1][3]=Gk[1][2+2*i]-Rk[2+2*i]*Gk[2][2+2*i]-Rk[2]*Gk[2+2*i][2+2*i];
      L[3][1]=L[1][3];
      for(k=0;k<4;++k) for(l=0;l<4;++l) 
	{
	  L[k][l]=L[k][l] + atr[i].Zcov[k][l];L_1[k][l]=L[k][l];
	}
      four_dim_inv(L_1);	
      for(j=0;j<ndim;++j) {
	Ak[0][j] = Gk[j][0] - Rk[1+2*i]*Gk[j][2]   - Rk[2]*Gk[j][1+2*i];
	Ak[1][j] = Gk[1][j] - Rk[2+2*i]*Gk[j][2] - Rk[2]*Gk[j][2+2*i];
	Ak[2][j] = Gk[j][1+2*i];
	Ak[3][j] = Gk[j][2+2*i];
	for(int v=0;v<4;++v) {
	  Kk[j][v] = 0.0;
	  for(int vv=0;vv<4;++vv) Kk[j][v] += Ak[vv][j]*L_1[vv][v];
	}
      }
      zeta[0] = atr[i].dzeta[0] - Rk[0] + Rk[1+2*i]*Rk[2];
      zeta[1] = atr[i].dzeta[1] - Rk[1] + Rk[2+2*i]*Rk[2];
      zeta[2] = atr[i].dzeta[2] - Rk[1+2*i];
      zeta[3] = atr[i].dzeta[3] - Rk[2+2*i];
      for(j=0;j<ndim;++j) 
	for(int v=0;v<4;++v) 
	  Rk[j] += Kk[j][v]*zeta[v];
      for(k=0;k<ndim;++k) 
	for(l=0;l<ndim;++l) 
	  for(int v=0;v<4;v++)
	    Gk[k][l] -= Kk[k][v]*Ak[v][l];
    }
  }
  this->SetMass(0.0);
  this->SetMassError(0.0);
  for (i=1;i<=ntrincluster;++i) {
    L[0][0]= Gk[0][0] 
            +Rk[1+2*i]*Rk[1+2*i]*Gk[2][2] 
            +Rk[2]*Rk[2]*Gk[1+2*i][1+2*i]
            -2.0*( Rk[1+2*i]*Gk[0][2] 
		  +Rk[2]*Gk[0][1+2*i] 
		  -Rk[1+2*i]*Rk[2]*Gk[2][1+2*i] );
    L[1][1]= Gk[1][1]
            +Rk[2+2*i]*Rk[2+2*i]*Gk[2][2]
            +Rk[2]*Rk[2]*Gk[2+2*i][2+2*i]
            -2.0*( Rk[2+2*i]*Gk[1][2]
                  +Rk[2]*Gk[1][2+2*i]
		  -Rk[2]*Rk[2+2*i]*Gk[2][2+2*i] );
    L[0][1]= Gk[0][1]
            +Rk[1+2*i]*Rk[2+2*i]*Gk[2][2]
            +Rk[2]*Rk[2]*Gk[1+2*i][2+2*i]
            -Rk[2]*( Gk[1][1+2*i] 
		    +Gk[0][2+2*i] )
            +Rk[2]*( Rk[1+2*i]*Gk[2][2+2*i] 
		    +Rk[2+2*i]*Gk[2][1+2*i] )
            -Rk[1+2*i]*Gk[1][2] 
            -Rk[2+2*i]*Gk[0][2];
    L[2][2]= Gk[1+2*i][1+2*i];
    L[3][3]= Gk[2+2*i][2+2*i];
    L[2][3]= Gk[1+2*i][2+2*i];
    L[0][2]= Gk[0][1+2*i]
            -Rk[1+2*i]*Gk[2][1+2*i]
            -Rk[2]*Gk[1+2*i][1+2*i];
    L[0][3]= Gk[0][2+2*i]
            -Rk[1+2*i]*Gk[2][2+2*i]
            -Rk[2]*Gk[1+2*i][2+2*i];
    L[1][2]= Gk[1][1+2*i]
            -Rk[2+2*i]*Gk[2][1+2*i]
            -Rk[2]*Gk[1+2*i][2+2*i];
    L[1][3]= Gk[1][2+2*i]
            -Rk[2+2*i]*Gk[2][2+2*i]
            -Rk[2]*Gk[2+2*i][2+2*i];
    L[1][0] = L[0][1];
    L[3][2] = L[2][3];
    L[2][0] = L[0][2];
    L[3][0] = L[0][3];
    L[2][1] = L[1][2];
    L[3][1] = L[1][3];  
    for (k=0;k<4;++k)  
      for (l=0;l<4;++l)  
	L_1[k][l] = L[k][l]+atr[i].Zcov[k][l];
    four_dim_inv(L_1);
    zeta[0] = atr[i].dzeta[0] - Rk[0] + Rk[1+2*i]*Rk[2];
    zeta[1] = atr[i].dzeta[1] - Rk[1] + Rk[2+2*i]*Rk[2];
    zeta[2] = atr[i].dzeta[2] - Rk[1+2*i];
    zeta[3] = atr[i].dzeta[3] - Rk[2+2*i];
    atr[i].dchi2 = 0.0;
    for(int v=0;v<4;++v) 
      for(int vv=0;vv<4;++vv) 
	atr[i].dchi2 += zeta[v]*zeta[vv]*L_1[v][vv];
  }

  if (verbose) cout<<"ntr = "<<ntrincluster<<endl;  
  this->SetX(Rk[0]);   this->SetY(Rk[1]);   this->SetZ(Rk[2]);
  this->SetCov(0, Gk[0][0]);
  this->SetCov(1, Gk[0][1]);
  this->SetCov(2, Gk[1][1]);
  this->SetCov(3, Gk[0][2]);     
  this->SetCov(4, Gk[1][2]);
  this->SetCov(5, Gk[2][2]);
  
  if (verbose) 
    for(i=1;i<=ntrincluster;++i) 
      cout<<"GEOFIT: Chi2 contrib. from track "<<i<<" = " <<atr[i].dchi2<<endl;
  
  dchi2t = 0.0;
  for (i=1;i<=ntrincluster;++i) dchi2t += atr[i].dchi2;
  if (verbose) cout<<"Total chi2 of the fit = "<<dchi2t<<endl;
  this->SetChi2(dchi2t);
  
  // fill updated info for each track
  for (i=1; i<=ntrincluster; ++i) {
    GroverParticleList::iterator it = trackincluster[i];
    atr[i].ax = Rk[1+2*i];
    atr[i].ay = Rk[2+2*i];
    it->PropagateBeginTo(this->Z());
    it->xb=this->X(); 
    it->yb=this->Y();
    it->axb=atr[i].ax;
    it->ayb=atr[i].ay;
    it->SetChi2(atr[i].dchi2);
  }
  fitted = true;
  return SUCCESS;
}

int GroverVertex::FastFit(void)
{ 
  //
  // This is the fitting function for 2-prong vertex,
  // employed in GroverPrimaries for seed generation.
  //
  // D.Emeliyanov 01/04/02
  // 

  static bool verbose = false;//(groverCuts.Verbose == 1);
  const int ROVERMAXTRACKS = 3;
  if (this->Size() >= ROVERMAXTRACKS || this->Size() < 2) return FAILURE;

  this->ResetParticles();
  typedef struct
  {
    double Zcov[4][4];
    double dzeta[4];
    float dchi2;
  } TRACK_TYPE, *pTRACK_TYPE;
 
  TRACK_TYPE atr[ROVERMAXTRACKS];
  
  const int   MaxIter=3;
  
  double Gk[3][3];
  double Rk[3];
  double Kk[3][2];
  double Ak[3][2];
  double L[2][2], L_1[2][2], Vk[2][2];
  double dchi2t;
  
  /////////////////
  float zref=0.0;// <- note!
  /////////////////
  
  double zeta[2],zv0,detr;
  int i,j,k,l,jjj;
  
  GroverParticleList::iterator trackincluster[ROVERMAXTRACKS] ={0};
  
  i=1;
  for (GroverParticleList::iterator it1 = particles.begin(); 
       it1 != particles.end(); ++it1) {
    trackincluster[i] = it1;
    it1->PropagateBeginTo(zref);
    ++i;
  }  
  int ntrincluster = i-1;
  
  for (i=1; i<=ntrincluster; ++i) {
    GroverParticleList::iterator it = trackincluster[i];
    for(int v=0;v<4;++v) for(int vv=0;vv<4;++vv) atr[i].Zcov[v][vv]=0.0;
    atr[i].Zcov[0][0] =  it->BeginCov(0);
    atr[i].Zcov[1][0] = atr[i].Zcov[0][1] =  it->BeginCov(1);
    atr[i].Zcov[2][0] = atr[i].Zcov[0][2] =  it->BeginCov(3);
    atr[i].Zcov[3][0] = atr[i].Zcov[0][3] =  it->BeginCov(6);
    atr[i].Zcov[2][1] = atr[i].Zcov[1][2] =  it->BeginCov(4);
    atr[i].Zcov[3][1] = atr[i].Zcov[1][3] =  it->BeginCov(7);
    atr[i].Zcov[1][1] = it->BeginCov(2);
    atr[i].Zcov[2][2] = it->BeginCov(5);
    atr[i].Zcov[2][3] = atr[i].Zcov[3][2] =  it->BeginCov(8);
    atr[i].Zcov[3][3] = it->BeginCov(9);
    atr[i].dzeta[0]   = it->BeginX();
    atr[i].dzeta[1]   = it->BeginY();
    atr[i].dzeta[2]   = it->BeginSlopeX();
    atr[i].dzeta[3]   = it->BeginSlopeY();
  }
  Rk[0]=0.0;   Rk[1]=0.0;  Rk[2]=0.0;
  for(jjj=1;jjj<=MaxIter;++jjj) 
    {
      for(i=0;i<3;++i) for(j=0;j<3;++j) Gk[i][j]=0.0;    
      Gk[0][0]=100.0;  
      Gk[1][1]=100.0;
      Gk[2][2]=100.0;zv0=Rk[2];
      for(i=1;i<=ntrincluster;++i) 
	{
	  Vk[0][0]=atr[i].Zcov[0][0]+2.0*zv0*atr[i].Zcov[0][2]+
	    zv0*zv0*atr[i].Zcov[2][2];
	  Vk[0][1]=atr[i].Zcov[0][1]+zv0*(atr[i].Zcov[1][2]+atr[i].Zcov[0][3])
	    +zv0*zv0*atr[i].Zcov[2][3];
	  Vk[1][0]=Vk[0][1];
	  Vk[1][1]=atr[i].Zcov[1][1]+2.0*zv0*atr[i].Zcov[1][3]+
	    zv0*zv0*atr[i].Zcov[3][3];
	  L[0][0]=Gk[0][0]-2.0*atr[i].dzeta[2]*Gk[0][2]+atr[i].dzeta[2]*
	    atr[i].dzeta[2]*Gk[2][2];
	  L[0][1]=Gk[0][1]-atr[i].dzeta[2]*Gk[1][2]-atr[i].dzeta[3]*Gk[0][2]+
	    atr[i].dzeta[2]*atr[i].dzeta[3]*Gk[2][2];
	  L[1][0]=L[0][1];
	  L[1][1]=Gk[1][1]-2.0*atr[i].dzeta[3]*Gk[1][2]+atr[i].dzeta[3]*
	    atr[i].dzeta[3]*Gk[2][2];      
	  for(k=0;k<2;++k) for(l=0;l<2;++l) L[k][l]+=Vk[k][l];
	  detr=L[0][0]*L[1][1]-L[0][1]*L[1][0];
	  L_1[0][0]=L[1][1]/detr;
	  L_1[1][1]=L[0][0]/detr;
	  L_1[0][1]=-L[0][1]/detr;
	  L_1[1][0]=L_1[0][1];

	  Ak[0][0]=Gk[0][0]-atr[i].dzeta[2]*Gk[0][2];
	  Ak[0][1]=Gk[0][1]-atr[i].dzeta[3]*Gk[0][2];
	  Ak[1][0]=Gk[0][1]-atr[i].dzeta[2]*Gk[1][2];
	  Ak[1][1]=Gk[1][1]-atr[i].dzeta[3]*Gk[1][2];
	  Ak[2][0]=Gk[0][2]-atr[i].dzeta[2]*Gk[2][2];
	  Ak[2][1]=Gk[1][2]-atr[i].dzeta[3]*Gk[2][2];
	  
	  for(j=0;j<3;j++)
	    for(int v=0;v<2;v++)
	      {
		Kk[j][v]=0.0;
		for(int vv=0;vv<2;++vv) Kk[j][v]+=Ak[j][vv]*L_1[vv][v];
	      }
	  zeta[0]=atr[i].dzeta[0]-Rk[0]+atr[i].dzeta[2]*Rk[2];
	  zeta[1]=atr[i].dzeta[1]-Rk[1]+atr[i].dzeta[3]*Rk[2];
	  for(j=0;j<3;++j) 
	    for(int v=0;v<2;++v) 
	      Rk[j]+=Kk[j][v]*zeta[v];
	  for(k=0;k<3;++k) 
	    for(l=0;l<3;++l) 
	      for(int v=0;v<2;v++)
		Gk[k][l]-=Kk[k][v]*Ak[l][v];
	  /*
	  for(j=0;j<3;j++) 
	    {
	      for(l=0;l<3;l++)
		printf("%f ",Gk[j][l]);
	      printf("\n");
	    }
	  cout<<" xv="<<Rk[0]<<" yv="<<Rk[1]<<" zv="<<Rk[2]<<endl;
	  */	
	}
    }
  zv0=Rk[2];
  for(i=1;i<=ntrincluster;++i) 
    {
      Vk[0][0]=atr[i].Zcov[0][0]+2.0*zv0*atr[i].Zcov[0][2]+
	zv0*zv0*atr[i].Zcov[2][2];
      Vk[0][1]=atr[i].Zcov[0][1]+zv0*(atr[i].Zcov[1][2]+atr[i].Zcov[0][3])
	+zv0*zv0*atr[i].Zcov[2][3];
      Vk[1][0]=Vk[0][1];
      Vk[1][1]=atr[i].Zcov[1][1]+2.0*zv0*atr[i].Zcov[1][3]+
	zv0*zv0*atr[i].Zcov[3][3];
      L[0][0]=Gk[0][0]-2.0*atr[i].dzeta[2]*Gk[0][2]+atr[i].dzeta[2]*
	atr[i].dzeta[2]*Gk[2][2];
      L[0][1]=Gk[0][1]-atr[i].dzeta[2]*Gk[1][2]-atr[i].dzeta[3]*Gk[0][2]+
	atr[i].dzeta[2]*atr[i].dzeta[3]*Gk[2][2];
      L[1][0]=L[0][1];
      L[1][1]=Gk[1][1]-2.0*atr[i].dzeta[3]*Gk[1][2]+atr[i].dzeta[3]*
	atr[i].dzeta[3]*Gk[2][2];      
      for(k=0;k<2;++k) for(l=0;l<2;++l) L[k][l]=L[k][l]+Vk[k][l];
      detr=L[0][0]*L[1][1]-L[0][1]*L[1][0];
      L_1[0][0]=L[1][1]/detr;
      L_1[1][1]=L[0][0]/detr;
      L_1[0][1]=-L[0][1]/detr;
      L_1[1][0]=L_1[0][1];
      atr[i].dchi2 = 0.0;
      zeta[0]=atr[i].dzeta[0]-Rk[0]+atr[i].dzeta[2]*Rk[2];
      zeta[1]=atr[i].dzeta[1]-Rk[1]+atr[i].dzeta[3]*Rk[2];
      for(int v=0;v<2;++v) 
	for(int vv=0;vv<2;++vv) 
	  atr[i].dchi2+=zeta[v]*zeta[vv]*L_1[v][vv];
    }
  if (verbose) cout<<"ntr = "<<ntrincluster<<endl;  
  this->SetX(Rk[0]);   this->SetY(Rk[1]);   this->SetZ(Rk[2]);
  this->SetCov(0, Gk[0][0]);
  this->SetCov(1, Gk[0][1]);
  this->SetCov(2, Gk[1][1]);
  this->SetCov(3, Gk[0][2]);     
  this->SetCov(4, Gk[1][2]);
  this->SetCov(5, Gk[2][2]);
  if (verbose) 
    for(i=1;i<=ntrincluster;++i) 
      cout<<"FAST FIT: Chi2 contrib. from track "<<i<<" = " <<
	atr[i].dchi2<<endl;
  dchi2t = 0.0;
  for (i=1;i<=ntrincluster;++i) dchi2t += atr[i].dchi2;
  if (verbose) cout<<"Total chi2 of the fit = "<<dchi2t<<endl;
  this->SetChi2(dchi2t);
  fitted = true;
  return SUCCESS;
}

int GroverVertex::RobustFit()
{ 
  //
  // This is the robust fitting routine for primary vertex 
  // reconstruction which should be called from GroverPrimaries.
  //
  // D.Emeliyanov 23/03/02
  // 

  const double ValRegion=12.0;
  const double Alpha1=0.01;
  const double Noise_Density=2.0;
  const double MinWeight=0.7;

  static bool verbose = false;//(groverCuts.Verbose == 1);
 
  const int ROVERMAXTRACKS = 80;//up to 80 tracks !
  if (this->Size() >= ROVERMAXTRACKS || this->Size() < 2) return FAILURE;


  this->ResetParticles();
  if (!trashcan.empty()) trashcan.clear();

  typedef struct
  {
    double Zcov[4][4];
    double dzeta[4];
    int flag,ID;
    float dchi2,beta1,beta0;
  } TRACK_TYPE, *pTRACK_TYPE;
 
  TRACK_TYPE atr[ROVERMAXTRACKS];
  
  const int   MaxIter=5;
  
  double Gk[3][3];
  double Gn[3][3];
  double Ga[3][3];
  double Rk[3];
  double Kk[3][2];
  double Ak[3][2];
  double L[2][2],L_1[2][2],gk[5][5],M[5][5],M_1[5][5];
  double Hk[2][5],Vk[2][2];
  double lambda,detr,zv0;
  
  /////////////////
  float zref=0.0;// <- note!
  /////////////////
  
  double zeta[2];
  int i,j,k,l,jjj;
  
  GroverParticleList::iterator trackincluster[ROVERMAXTRACKS] ={0};
  
  i=1;
  for(GroverParticleList::iterator it1=particles.begin(); 
      it1!=particles.end();++it1) 
    {
      trackincluster[i] = it1;
      it1->PropagateBeginTo(zref);
      i++;
    }
  int ntrincluster = i-1;
  for (i=1; i<=ntrincluster; ++i) {
    GroverParticleList::iterator it = trackincluster[i];
    for(int v=0;v<4;++v) for(int vv=0;vv<4;++vv) atr[i].Zcov[v][vv]=0.0;
    atr[i].Zcov[0][0] =  it->BeginCov(0);
    atr[i].Zcov[1][0] = atr[i].Zcov[0][1] =  it->BeginCov(1);
    atr[i].Zcov[2][0] = atr[i].Zcov[0][2] =  it->BeginCov(3);
    atr[i].Zcov[3][0] = atr[i].Zcov[0][3] =  it->BeginCov(6);
    atr[i].Zcov[2][1] = atr[i].Zcov[1][2] =  it->BeginCov(4);
    atr[i].Zcov[3][1] = atr[i].Zcov[1][3] =  it->BeginCov(7);
    atr[i].Zcov[1][1] = it->BeginCov(2);
    atr[i].Zcov[2][2] = it->BeginCov(5);
    atr[i].Zcov[2][3] = atr[i].Zcov[3][2] =  it->BeginCov(8);
    atr[i].Zcov[3][3] = it->BeginCov(9);
    atr[i].dzeta[0]   = it->BeginX();
    atr[i].dzeta[1]   = it->BeginY();
    atr[i].dzeta[2]   = it->BeginSlopeX();
    atr[i].dzeta[3]   = it->BeginSlopeY();
    atr[i].ID=it->Index();
  }
  Rk[0]= this->X();   Rk[1]=this->Y();  Rk[2]=this->Z();
  k=0;
  for(int v=0;v<3;v++)
    for(int vv=0;vv<=v;vv++)
      {
	Gk[v][vv]=Gk[vv][v]=this->Cov(k);
	k++;
      }
  for(jjj=1;jjj<=MaxIter;++jjj) 
    {
      for(int v=0;v<2;v++) 
	for(int vv=0;vv<5;vv++) Hk[v][vv]=0.0;
      Hk[0][0]=Hk[1][1]=1.0;Hk[0][2]=Rk[2];Hk[1][2]=Rk[2];
      for(i=1;i<=ntrincluster;++i) 
	{
	  Hk[0][4]=atr[i].dzeta[2];Hk[1][4]=atr[i].dzeta[3];
	  for(int v=0;v<5;v++)
	    for(int vv=0;vv<5;vv++) gk[v][vv]=0.0;
	  for(int v=0;v<3;v++)
	    for(int vv=0;vv<3;vv++) gk[v][vv]=atr[i].Zcov[v][vv];
	  gk[4][4]=Gk[2][2];
	  for(int v=0;v<5;v++) 
	    for(int vv=0;vv<5;vv++) 
	      {
		M[v][vv]=0.0;M_1[v][vv]=0.0;
	      }
	  for(int v=0;v<2;v++)
	    for(int vv=0;vv<5;vv++) 
	      {
		for(int m=0;m<5;m++) M_1[v][vv]+=Hk[v][m]*gk[m][vv];
	      }
	  for(int v=0;v<2;v++)
	    for(int vv=0;vv<2;vv++)
	      {
		for(int m=0;m<5;m++) M[v][vv]+=M_1[v][m]*Hk[vv][m];
	      }
	  M[0][0]+=Gk[0][0];M[1][1]+=Gk[1][1];
	  detr=M[0][0]*M[1][1]-M[0][1]*M[1][0];
	  M_1[0][0]=M[1][1]/detr;M_1[0][1]=-M[0][1]/detr;M_1[1][0]=M_1[0][1];
	  M_1[1][1]=M[0][0]/detr;
	  zeta[0]=atr[i].dzeta[0]+atr[i].dzeta[2]*Rk[2]-Rk[0];
	  zeta[1]=atr[i].dzeta[1]+atr[i].dzeta[3]*Rk[2]-Rk[1];
	  lambda=0.0;
	  for(int v=0;v<2;v++) 
	    for(int vv=0;vv<2;vv++) 
	      lambda+=zeta[v]*M_1[v][vv]*zeta[vv];
	  if(lambda>ValRegion)
	    {
	      atr[i].flag=0;atr[i].beta0=1.0;atr[i].beta1=0.0;
	    }
	  else
	    {
	      atr[i].flag=1;
	      atr[i].beta0=1.0/(1.0+exp(-0.5*lambda)/
				(2*3.14159*sqrt(detr)*Noise_Density*Alpha1));
	      atr[i].beta1=1.0-atr[i].beta0;
	      if(atr[i].beta1<MinWeight)
		{
		  atr[i].flag=0;atr[i].beta0=1.0;atr[i].beta1=0.0;
		}
	    }
	}
      for(i=0;i<3;++i) for(j=0;j<3;++j) Gk[i][j]=0.0;    
      Gk[0][0]=100.0;  
      Gk[1][1]=100.0;
      Gk[2][2]=100.0;zv0=Rk[2];
      for(i=1;i<=ntrincluster;++i) 
	{
	  if(atr[i].flag==0) continue;
	  Vk[0][0]=atr[i].Zcov[0][0]+2.0*zv0*atr[i].Zcov[0][2]+
	    zv0*zv0*atr[i].Zcov[2][2];
	  Vk[0][1]=atr[i].Zcov[0][1]+zv0*(atr[i].Zcov[1][2]+atr[i].Zcov[0][3])
	    +zv0*zv0*atr[i].Zcov[2][3];
	  Vk[1][0]=Vk[0][1];
	  Vk[1][1]=atr[i].Zcov[1][1]+2.0*zv0*atr[i].Zcov[1][3]+
	    zv0*zv0*atr[i].Zcov[3][3];
	  L[0][0]=Gk[0][0]-2.0*atr[i].dzeta[2]*Gk[0][2]+atr[i].dzeta[2]*
	    atr[i].dzeta[2]*Gk[2][2];
	  L[0][1]=Gk[0][1]-atr[i].dzeta[2]*Gk[1][2]-atr[i].dzeta[3]*Gk[0][2]+
	    atr[i].dzeta[2]*atr[i].dzeta[3]*Gk[2][2];
	  L[1][0]=L[0][1];
	  L[1][1]=Gk[1][1]-2.0*atr[i].dzeta[3]*Gk[1][2]+atr[i].dzeta[3]*
	    atr[i].dzeta[3]*Gk[2][2];      
	  for(k=0;k<2;++k) for(l=0;l<2;++l) L[k][l]+=Vk[k][l];
	  detr=L[0][0]*L[1][1]-L[0][1]*L[1][0];
	  L_1[0][0]=L[1][1]/detr;
	  L_1[1][1]=L[0][0]/detr;
	  L_1[0][1]=-L[0][1]/detr;
	  L_1[1][0]=L_1[0][1];

	  Ak[0][0]=Gk[0][0]-atr[i].dzeta[2]*Gk[0][2];
	  Ak[0][1]=Gk[0][1]-atr[i].dzeta[3]*Gk[0][2];
	  Ak[1][0]=Gk[0][1]-atr[i].dzeta[2]*Gk[1][2];
	  Ak[1][1]=Gk[1][1]-atr[i].dzeta[3]*Gk[1][2];
	  Ak[2][0]=Gk[0][2]-atr[i].dzeta[2]*Gk[2][2];
	  Ak[2][1]=Gk[1][2]-atr[i].dzeta[3]*Gk[2][2];
	  
	  for(j=0;j<3;j++)
	    for(int v=0;v<2;v++)
	      {
		Kk[j][v]=0.0;
		for(int vv=0;vv<2;++vv) Kk[j][v]+=Ak[j][vv]*L_1[vv][v];
	      }
	  zeta[0]=atr[i].dzeta[0]-Rk[0]+atr[i].dzeta[2]*Rk[2];
	  zeta[1]=atr[i].dzeta[1]-Rk[1]+atr[i].dzeta[3]*Rk[2];
	  for(j=0;j<3;++j) 
	    for(int v=0;v<2;++v) 
	      Rk[j]+=atr[i].beta1*Kk[j][v]*zeta[v];
	  for(k=0;k<3;++k) for(l=0;l<3;++l) 
	      {
		Gn[k][l]=Gk[k][l];
		for(int v=0;v<2;v++)
		  Gn[k][l]-=Kk[k][v]*Ak[l][v];
	      }
	  for(k=0;k<2;k++) for(l=0;l<2;l++) 
	L[k][l]=zeta[k]*zeta[l]*atr[i].beta0*atr[i].beta1;
	  for(k=0;k<3;k++) for(l=0;l<2;l++)
	    {
	      Ak[k][l]=0.0;
	      for(int v=0;v<2;v++) Ak[k][l]+=Kk[k][v]*L[v][l];
	    }
	  for(k=0;k<3;k++) for(l=0;l<3;l++)
	    {
	      Ga[k][l]=0.0;
	      for(int v=0;v<2;v++) Ga[k][l]+=Ak[k][v]*Kk[l][v];
	    }
	  for(k=0;k<3;k++) for(l=0;l<3;l++)
	    {
	      Gk[k][l]=atr[i].beta0*Gk[k][l]+(1.0-atr[i].beta0)*Gn[k][l]+
		Ga[k][l];
	    }
	  /*
	    for(j=0;j<3;j++) 
	    {
	    for(l=0;l<3;l++)
	    printf("%f ",Gk[j][l]);
	    printf("\n");
	    }
	    cout<<" xv="<<Rk[0]<<" yv="<<Rk[1]<<" zv="<<Rk[2]<<endl;
	  */	
	}
    }
  this->SetMass(0.0);
  this->SetMassError(100.0);
  if (verbose) cout<<"ntr = "<<ntrincluster<<endl;  
  this->SetX(Rk[0]);   this->SetY(Rk[1]);   this->SetZ(Rk[2]);
  this->SetCov(0, Gk[0][0]);
  this->SetCov(1, Gk[0][1]);
  this->SetCov(2, Gk[1][1]);
  this->SetCov(3, Gk[0][2]);     
  this->SetCov(4, Gk[1][2]);
  this->SetCov(5, Gk[2][2]);
  this->SetChi2(0.0);  
  i=1;
  for(GroverParticleList::iterator ipart=particles.begin();
       ipart!=particles.end();++ipart)
    {
      if((atr[i].flag == 0)||(atr[i].beta1<MinWeight))
	{
	  ipart->SetUsedInPrimary(-1);
	  trashcan.push_back(*ipart);
	  particles.erase(ipart); --ipart;
	}
      ++i;
    }
  fitted = true;
  return SUCCESS;
}


#ifndef GROVER_WITHOUT_ARTE
int GroverVertex::Fit(const ArtePointer<TOPP> topp)
{
  this->ClearConstraints();
  if (topp!=0)
    this->AddConstraint(topp,0,1,2,3,4,5,6,7,8,9);
  return this->Fit();
}
#endif




int GroverVertex::Fit(const float Mver)
{
  this->ClearConstraints();
  if (Mver > 0.0)
    this->AddConstraint(Mver,0,1,2,3,4,5,6,7,8,9);
  return this->Fit();
}




int GroverVertex::GeoFit(void)
{ //
  // Fit primary vertex using purely geometric method of 
  // Rover 1.05 
  //
  // I.Kisel/D.Emeliyanov
  //
  const int ROVERMAXTRACKS = 200;
  if (this->Size() > ROVERMAXTRACKS || this->Size() < 2) return FAILURE;

  bool MC = false;
  bool tampered = false;
  for (GroverParticleList::const_iterator ip = this->particles.begin();
       ip != this->particles.end(); ++ip) {
    if (ip->Type() == 4) {MC = true; break;}
    if (ip->Tampered()) tampered = true;
  }
  if (MC) return FAILURE; // can't fit a vertex with > 0 perfect tracks!

  // Reset all tracks UNLESS the user has been fiddling with them
  if (!tampered) this->ResetParticles();

  if (!trashcan.empty()) trashcan.clear();

  typedef struct
  {
    float cov[4][4];
    float Tcov[2][2];
    float Tpar[2];
    float dzeta[4];
    float zf;
    float ChiInc;
    int flag;
  } TRACK_TYPE;

  TRACK_TYPE atr[ROVERMAXTRACKS];

  int n_track;
  const float dchi2  = 20.0;
  const float Pfalse =  0.1;
  const int   MaxIter=  3;

  float Vcov[3][3];
  float Rk[5];
  float Vpar[3];
  float dG[5][5],KH[5][5],Gk[5][5];
  float Hk[4][5];
  float Kk[5][4];
  double L[4][4];
  float Fk[5][4];
  float lambda,ax,ay,beta1,dz,locchi2;
  float res[4],ext[4];
  int i,j,itr,k,nValidTrack;

  Vpar[0] = x;   Vpar[1] = y;   Vpar[2] = z;

  // Fill cov matrices and track parameters
  n_track = 0;
  for (GroverParticleList::iterator ipart = particles.begin();
       ipart != particles.end(); ++ipart) {
    if (n_track > ROVERMAXTRACKS) break; // hard limit
    ipart->PropagateBeginTo(Vpar[2]);
    atr[n_track].cov[0][0] = ipart->BeginCov(0);
    atr[n_track].cov[1][1] = ipart->BeginCov(2);
    atr[n_track].cov[2][2] = ipart->BeginCov(5);
    atr[n_track].cov[3][3] = ipart->BeginCov(9);
    atr[n_track].cov[1][0] = atr[n_track].cov[0][1] = ipart->BeginCov(1);
    atr[n_track].cov[2][0] = atr[n_track].cov[0][2] = ipart->BeginCov(3);
    atr[n_track].cov[2][1] = atr[n_track].cov[1][2] = ipart->BeginCov(4);
    atr[n_track].cov[3][0] = atr[n_track].cov[0][3] = ipart->BeginCov(6);
    atr[n_track].cov[3][1] = atr[n_track].cov[1][3] = ipart->BeginCov(7);
    atr[n_track].cov[3][2] = atr[n_track].cov[2][3] = ipart->BeginCov(8);
    atr[n_track].dzeta[0]  = ipart->BeginX();
    atr[n_track].dzeta[1]  = ipart->BeginY();
    atr[n_track].dzeta[2]  = ipart->BeginSlopeX();
    atr[n_track].dzeta[3]  = ipart->BeginSlopeY();
    atr[n_track].zf        = Vpar[2];
    atr[n_track].Tpar[0]   = atr[n_track].dzeta[2];
    atr[n_track].Tpar[1]   = atr[n_track].dzeta[3];
    for(i=0;i<2;++i) for(j=0;j<2;++j)  
      atr[n_track].Tcov[i][j] = atr[n_track].cov[i+2][j+2];
    atr[n_track].flag=1;
    ++n_track;
  }

  // Iteratively fit vertex - number of iterations fixed at 3
  for(int iteration=1; iteration<=MaxIter; ++iteration) { 
    // (re)initialize the vertex covariance
    for(i=0;i<3;++i) for(j=0;j<3;++j) Vcov[i][j]=0.0;
    Vcov[0][0] = cov[0];  
    Vcov[1][1] = cov[2];  
    Vcov[2][2] = cov[5];
    
    // Part one: vertex position estimation 
    for(itr=0; itr<n_track; ++itr){
      // Construct covariance from vertex and track parts 
      for(i=0;i<5;++i)  for(j=0;j<5;++j)  Gk[i][j] = 0.0;
      for(i=0;i<3;++i)  for(j=0;j<3;++j)  Gk[i][j] = Vcov[i][j];
      for(i=3;i<5;++i)  for(j=3;j<5;++j)  Gk[i][j] = atr[itr].Tcov[i-3][j-3];
      // Construct state vector from vertex and track parts
      for (i=0;i<3;++i) Rk[i]=Vpar[i];
      Rk[3] = atr[itr].Tpar[0];
      Rk[4] = atr[itr].Tpar[1];
      // Linearized measurement matrix 
      dz = atr[itr].zf - Rk[2];
      ax = Rk[3];
      ay = Rk[4];
      for(i=0;i<4;++i)  for(j=0;j<5;++j)  Hk[i][j]=0.0;
      Hk[0][0] = Hk[1][1] = Hk[2][3] = Hk[3][4] = 1.0;
      Hk[0][2] = -ax; Hk[1][2] = -ay;
      Hk[0][3] = Hk[1][4] = dz;
      // Check if track is in the validation region
      for(i=0;i<4;++i)  for(j=0;j<4;++j)  L[i][j]=0.0;
      L[0][0] = Gk[0][0] + ax*ax*Gk[2][2] + dz*dz*Gk[3][3] +
	2.0*(dz*Gk[0][3] - ax*Gk[0][2]    - ax*dz*Gk[2][3]);
      L[1][1] = Gk[1][1] + ay*ay*Gk[2][2] + dz*dz*Gk[4][4] +
	2.0*(dz*Gk[1][4] - ay*Gk[1][2]    - ay*dz*Gk[2][4]);
      L[0][1] = Gk[0][1] + ax*ay*Gk[2][2] + dz*dz*Gk[3][4] +
	dz*(Gk[0][4]+Gk[1][3]) - ay*(Gk[0][2]+dz*Gk[2][3]) -
	ax*(Gk[1][2]+dz*Gk[2][4]);
      L[1][0] = L[0][1];
      L[2][2] = Gk[3][3];
      L[0][2] = Gk[0][3] - ax*Gk[2][3] + dz*Gk[3][3];
      L[2][0] = L[0][2];
      L[0][3] = Gk[0][4] - ax*Gk[2][4] + dz*Gk[3][4];
      L[3][0] = L[0][3];
      L[1][2] = Gk[1][3] - ay*Gk[2][3] + dz*Gk[3][4];
      L[2][1] = L[1][2];
      L[1][3] = Gk[1][4] - ay*Gk[2][4] + dz*Gk[4][4];
      L[3][1] = L[1][3];
      L[3][3] = Gk[4][4];
      L[2][3] = L[3][2] = Gk[3][4];
      for(i=0;i<4;++i) for (j=0;j<4;++j) L[i][j]+=atr[itr].cov[i][j];
      four_dim_inv(L); //  <= Invert L
      ext[0] = Rk[0] + dz*Rk[3];
      ext[1] = Rk[1] + dz*Rk[4];
      ext[2] = Rk[3];
      ext[3] = Rk[4];
      for(i=0;i<4;++i) res[i] = atr[itr].dzeta[i] - ext[i];
      lambda=0.0;
      for(i=0;i<4;++i)  for(j=0;j<4;++j)  lambda += res[i]*res[j]*L[i][j];
      if (lambda>dchi2) {
	atr[itr].flag=0;
	continue;
      }
      else if (atr[itr].flag){
	// Probability of correct classification
	beta1=1.0-Pfalse/(Pfalse+exp(-lambda));
	// Kalman gain
	Fk[0][0] = Gk[0][0] - ax*Gk[0][2] + dz*Gk[0][3];
	Fk[0][1] = Gk[0][1] - ay*Gk[0][2] + dz*Gk[0][4];
	Fk[0][2] = Gk[0][3];
	Fk[0][3] = Gk[0][4];
	Fk[1][0] = Gk[0][1] - ax*Gk[1][2] + dz*Gk[1][3];
	Fk[1][1] = Gk[1][1] - ay*Gk[1][2] + dz*Gk[1][4];
	Fk[1][2] = Gk[1][3];
	Fk[1][3] = Gk[1][4];
	Fk[2][0] = Gk[0][2] - ax*Gk[2][2] + dz*Gk[2][3];
	Fk[2][1] = Gk[1][2] - ay*Gk[2][2] + dz*Gk[2][4];
	Fk[2][2] = Gk[2][3];
	Fk[2][3] = Gk[2][4];
	Fk[3][0] = Gk[0][3] - ax*Gk[2][3] + dz*Gk[3][3];
	Fk[3][1] = Gk[1][3] - ay*Gk[2][3] + dz*Gk[3][4];
	Fk[3][2] = Gk[3][3];
	Fk[3][3] = Gk[3][4];
	Fk[4][0] = Gk[0][4] - ax*Gk[2][4] + dz*Gk[3][4];
	Fk[4][1] = Gk[1][4] - ay*Gk[2][4] + dz*Gk[4][4];
	Fk[4][2] = Gk[3][4];
	Fk[4][3] = Gk[4][4];
	for(i=0;i<5;++i) for(j=0;j<4;++j){
	  Kk[i][j]=0.0;
	  for(k=0;k<4;++k)  Kk[i][j] += Fk[i][k]*L[k][j];
	}
	// New estimation of the vertex position 
	for(i=0;i<5;++i)  for(j=0;j<4;++j)  Rk[i] += beta1*Kk[i][j]*res[j];
	// New covariance matrix
	for(i=0;i<5;++i) for(j=0;j<5;++j){
	  KH[i][j] = 0.0;
	  for(k=0;k<4;++k) KH[i][j] += Kk[i][k]*Hk[k][j];
	}
	for(i=0;i<5;++i)  for(j=0;j<5;++j){
	  dG[i][j]=0.0;
	  for(k=0;k<5;++k) dG[i][j] += KH[i][k]*Gk[k][j];
	}
	for(i=0;i<5;++i)  for(j=0;j<5;++j)  Gk[i][j] -= beta1*dG[i][j];
	// Extract vertex part
	for(i=0;i<3;++i) Vpar[i] = Rk[i];
	for(i=0;i<3;++i)  for (j=0;j<3;++j)  Vcov[i][j] = Gk[i][j];
      }
    }
    
    // Part two: smoothing track parameters using estimated vertex position
    for(itr=0;itr<n_track;++itr) if(atr[itr].flag){
      // Construct covariance from vertex and track parts
      for(i=0;i<5;++i)  for(j=0;j<5;++j)  Gk[i][j] = 0.0;
      for(i=0;i<3;++i)  for(j=0;j<3;++j)  Gk[i][j] = Vcov[i][j];
      for(i=3;i<5;++i)  for(j=3;j<5;++j)  Gk[i][j] = atr[itr].Tcov[i-3][j-3];
      // Construct state vector from vertex and track parts
      for (i=0;i<3;++i) Rk[i]=Vpar[i];
      Rk[3] = atr[itr].Tpar[0];
      Rk[4] = atr[itr].Tpar[1];
      // Linearized measurement matrix
      dz = atr[itr].zf - Rk[2];
      ax = Rk[3];
      ay = Rk[4];
      for(i=0;i<4;++i)  for(j=0;j<5;++j)  Hk[i][j]=0.0;
      Hk[0][0] = Hk[1][1] = Hk[2][3] = Hk[3][4] = 1.0;
      Hk[0][2] = -ax;
      Hk[1][2] = -ay;
      Hk[0][3] = Hk[1][4] = dz;
      // Check if track is still in the validation region
      for(i=0;i<4;++i)  for(j=0;j<4;++j)  L[i][j] = 0.0;
      L[0][0] = Gk[0][0] + ax*ax*Gk[2][2] + dz*dz*Gk[3][3] +
	2.0*(dz*Gk[0][3] - ax*Gk[0][2]    - ax*dz*Gk[2][3]);
      L[1][1] = Gk[1][1] + ay*ay*Gk[2][2] + dz*dz*Gk[4][4] +
	2.0*(dz*Gk[1][4] - ay*Gk[1][2]    - ay*dz*Gk[2][4]);
      L[0][1] = Gk[0][1] + ax*ay*Gk[2][2] + dz*dz*Gk[3][4] + 
	dz*(Gk[0][4]+Gk[1][3]) - ay*(Gk[0][2]+dz*Gk[2][3]) - 
	ax*(Gk[1][2]+dz*Gk[2][4]);
      L[1][0] = L[0][1];
      L[2][2] = Gk[3][3];
      L[0][2] = Gk[0][3] - ax*Gk[2][3] + dz*Gk[3][3];
      L[2][0] = L[0][2];
      L[0][3] = Gk[0][4] - ax*Gk[2][4] + dz*Gk[3][4];
      L[3][0] = L[0][3];
      L[1][2] = Gk[1][3] - ay*Gk[2][3] + dz*Gk[3][4];
      L[2][1] = L[1][2];
      L[1][3] = Gk[1][4] - ay*Gk[2][4] + dz*Gk[4][4];
      L[3][1] = L[1][3];
      L[3][3] = Gk[4][4];
      L[2][3] = L[3][2] = Gk[3][4];
      for(i=0;i<4;++i)  for (j=0;j<4;++j)  L[i][j]+=atr[itr].cov[i][j];
      four_dim_inv(L); // <= Invert L
      ext[0] = Rk[0] + dz*Rk[3];
      ext[1] = Rk[1] + dz*Rk[4];
      ext[2] = Rk[3];
      ext[3] = Rk[4];
      for(i=0;i<4;++i) res[i] = atr[itr].dzeta[i] - ext[i];
      lambda=0.0;
      for(i=0;i<4;++i)  for(j=0;j<4;++j)  lambda += res[i]*res[j]*L[i][j];
      if(lambda>dchi2) {
	// discard track
	atr[itr].flag=0;
	continue;
      }
      else{
	if(iteration == MaxIter) atr[itr].ChiInc=lambda;
	// Probability of correct classification
	beta1 = 1.0 - Pfalse/(Pfalse+exp(-lambda));
	// Kalman gain
	Fk[0][0] = Gk[0][0] - ax*Gk[0][2] + dz*Gk[0][3];
	Fk[0][1] = Gk[0][1] - ay*Gk[0][2] + dz*Gk[0][4];
	Fk[0][2] = Gk[0][3];
	Fk[0][3] = Gk[0][4];
	Fk[1][0] = Gk[0][1] - ax*Gk[1][2] + dz*Gk[1][3];
	Fk[1][1] = Gk[1][1] - ay*Gk[1][2] + dz*Gk[1][4];
	Fk[1][2] = Gk[1][3];
	Fk[1][3] = Gk[1][4];
	Fk[2][0] = Gk[0][2] - ax*Gk[2][2] + dz*Gk[2][3];
	Fk[2][1] = Gk[1][2] - ay*Gk[2][2] + dz*Gk[2][4];
	Fk[2][2] = Gk[2][3];
	Fk[2][3] = Gk[2][4];
	Fk[3][0] = Gk[0][3] - ax*Gk[2][3] + dz*Gk[3][3];
	Fk[3][1] = Gk[1][3] - ay*Gk[2][3] + dz*Gk[3][4];
	Fk[3][2] = Gk[3][3];
	Fk[3][3] = Gk[3][4];
	Fk[4][0] = Gk[0][4] - ax*Gk[2][4] + dz*Gk[3][4];
	Fk[4][1] = Gk[1][4] - ay*Gk[2][4] + dz*Gk[4][4];
	Fk[4][2] = Gk[3][4];
	Fk[4][3] = Gk[4][4];
	for(i=0;i<5;++i)  for(j=0;j<4;++j){
	  Kk[i][j] = 0.0;
	  for(k=0;k<4;++k) Kk[i][j] += Fk[i][k]*L[k][j];
	}
	// New estimation	
	for(i=0;i<5;++i)  for(j=0;j<4;++j)  Rk[i] += beta1*Kk[i][j]*res[j];
	// New covariance matrix
	for(i=0;i<5;++i)  for(j=0;j<5;++j) {
	  KH[i][j] = 0.0;
	  for(k=0;k<4;++k) KH[i][j] += Kk[i][k]*Hk[k][j];
	}
	for(i=0;i<5;++i)  for(j=0;j<5;++j) {
	  dG[i][j] = 0.0;
	  for(k=0;k<5;++k) dG[i][j] += KH[i][k]*Gk[k][j];
	}
	for(i=0;i<5;++i)  for(j=0;j<5;++j)  Gk[i][j] -= beta1*dG[i][j];
	// Extract track and vertex parts
	atr[itr].Tpar[0] = Rk[3];
	atr[itr].Tpar[1] = Rk[4];
      }
    }
  } // finished iterations

  locchi2 = 0.0;
  nValidTrack = 0;
  for(int itr=0; itr<n_track; itr++) 
    if (atr[itr].flag != 0) {
      locchi2 += atr[itr].ChiInc;
      ++nValidTrack;
    }

  // no normalisation now!
  //locchi2 /= (4.0*nValidTrack - 3.0 - 2.0*nValidTrack); 

  // Write vertex parameters and update lists
  chi2   = locchi2;
  x      = Vpar[0];
  y      = Vpar[1];
  z      = Vpar[2];
  cov[0] = Vcov[0][0];     
  cov[1] = Vcov[0][1];
  cov[2] = Vcov[1][1];
  cov[3] = Vcov[0][2];     
  cov[4] = Vcov[1][2];
  cov[5] = Vcov[2][2];
  itr = 0;
  for (GroverParticleList::iterator ipart = particles.begin();
       ipart != particles.end(); ++ipart) {
    // if it's been excluded from vertex, mark as unused and move to trash
    int idx = 1;
    if (atr[itr].flag == 0) {
      ipart->SetUsedInPrimary(0);
      trashcan.push_back(*ipart);
      particles.erase(ipart); --ipart;
    }
    else {
      if (this->Index() > 0) idx = this->Index();
      ipart->SetUsedInPrimary(idx);
    }
    ++itr;
  }

  fitted = true;
  return SUCCESS;
}



float GroverVertex::Momentum(void)
{
  if (mom_estimated) return p;
  else return this->EstimateMomentum();
}




float GroverVertex::SlopeX(void)
{
  if (mom_estimated) return ax;
  else {
    if (this->EstimateMomentum() > 0.0) return ax;
    else return FAILURE;
  }
}




float GroverVertex::SlopeY(void)
{
  if (mom_estimated) return ay;
  else {
    if (this->EstimateMomentum() > 0.0) return ay;
    else return FAILURE;
  }
}




float GroverVertex::EstimateMomentum(void)
{
  //
  // Estimate momentum vector (and covariances) of 
  // vertex mother particle.
  //
  // 09.07.2000 V.Kekelidze
  //

  // Quick checks
  if (mom_estimated) return p;
  int err = 0;
  if (!fitted) err = this->Fit(-1.0);
  if (err != 0) return FAILURE;

  float fact=0.0, xn=0.0, yn=0.0, zn=0.0;

  const int ROVERMAXTRACKS = 100;
  double Gk[3*ROVERMAXTRACKS+3][3*ROVERMAXTRACKS+3];

  int ntrincluster = this->Size();
  int ndim = 3*ntrincluster+3;

  //  cout<<"Sum_p: Input cov matrix:"<<endl;
  int n=0;
  int gcsize = static_cast<int>(gcov.size());
  if (gcsize < 1) return FAILURE;
  if (ndim*(ndim+1)/2 > gcsize) {
    fitted = false; mom_estimated = false;
    return FAILURE;
  }

  for(int i=0; i<ndim; ++i){
    for(int j=0; j<=i; ++j){
      Gk[i][j] = Gk[j][i] = gcov[n];
      ++n;
    }
  }

  // ****** sum momenta *********

  for (GroverParticleList::iterator i_tr = this->particles.begin();
       i_tr != this->particles.end(); ++i_tr){
    
    if (i_tr->Momentum() <= 0.0) return FAILURE;

    float ax = i_tr->BeginSlopeX();
    float ay = i_tr->BeginSlopeY();
    float vp = i_tr->Momentum();
    fact = 1.0 / sqrt(1.0 + ax*ax + ay*ay);
    xn += ax*fact*vp;
    yn += ay*fact*vp;
    zn += fact*vp;
  }

  float ax_new = xn/zn;
  float ay_new = yn/zn;

  float x_new = this->X() - this->Z()*ax_new;
  float y_new = this->Y() - this->Z()*ay_new; 
  float p_new = sqrt(xn*xn +yn*yn + zn*zn);

  // ****** get Jacobian *********
  
  float Jac[5][3*ROVERMAXTRACKS+3];

  for(int i=0; i<5; ++i)
    for(int j=0; j<3; ++j)
      Jac[i][j] = 0.0;

  Jac[2][0]=  1.0;
  Jac[2][2]= -ax_new;
  Jac[3][1]=  1.0;
  Jac[3][2]= -ay_new;

  for(int pp=1; pp<=ntrincluster; ++pp){
    if(this->particles[pp-1].Momentum()<=0)return FAILURE;

    float ax = this->particles[pp-1].BeginSlopeX();
    float ay = this->particles[pp-1].BeginSlopeY();
    float znn = 1.0 / sqrt(1 + ax*ax + ay*ay); 
    float xnn = ax*znn; 
    float ynn = ay*znn;

    fact = this->particles[pp-1].Momentum()*znn/zn;
    
    Jac[0][3*pp]   = fact*(1.0 - xnn*xnn + ax_new*xnn*znn); 
    Jac[0][3*pp+1] = fact*ynn*(ax_new*znn - xnn); 
    
    Jac[1][3*pp]   = fact*xnn*(ay_new*znn - ynn);         
    Jac[1][3*pp+1] = fact*(1.0-ynn*ynn + ay_new*ynn*znn);
    
    Jac[2][3*pp]   = -this->Z() * Jac[0][3*pp];   
    Jac[2][3*pp+1] = -this->Z() * Jac[0][3*pp+1];
    
    Jac[3][3*pp]   = -this->Z() * Jac[1][3*pp]; 
    Jac[3][3*pp+1] = -this->Z() * Jac[1][3*pp+1];

    fact = this->particles[pp-1].Momentum()*znn; 
    float ps_x = xn/p_new;
    float ps_y = yn/p_new;
    float ps = zn/p_new;
    Jac[4][3*pp]   = fact*(ps_x*(1 - xnn*xnn) - ps_y*xnn*ynn - ps*xnn*znn);
    Jac[4][3*pp+1] = fact*(ps_y*(1 - ynn*ynn) - ps_x*xnn*ynn - ps*xnn*znn);

    fact = 1.0/zn;
    Jac[0][3*pp+2] = fact*(xnn - ax_new*znn);
    Jac[1][3*pp+2] = fact*(ynn - ay_new*znn);
    Jac[2][3*pp+2] = -this->Z()*Jac[0][3*pp+2];
    Jac[3][3*pp+2] = -this->Z()*Jac[1][3*pp+2];
    Jac[4][3*pp+2] = ps_x*xnn + ps_y*ynn + ps*znn;
  }
     
  float vCov[5][5];
  for(int i=0; i<5; ++i)
    for(int j=0; j<5; ++j){
      vCov[i][j]=0;
      for(int k=0; k<ndim; ++k){
	float sum=0;
	for(int l=0; l<ndim; ++l)
	  sum+=Gk[k][l]*Jac[j][l];
	vCov[i][j]+=Jac[i][k]*sum;
      }
      vCov[j][i]=vCov[i][j];
    }

  //set vertex  parameters

  this->SetSlopeX(ax_new);
  this->SetSlopeY(ay_new);
  this->SetBx(x_new);
  this->SetBy(y_new);
  this->SetMomentum(p_new);

  // for cov ~ {x, y, z, ax, ay, p, bx, by}
  
  this->SetCov(9, vCov[0][0]);//ax-ax
  this->SetCov(13,vCov[0][1]);//ax-ay
  this->SetCov(14,vCov[1][1]);//ay-ay

  this->SetCov(18,vCov[0][4]);//ax-p
  this->SetCov(19,vCov[1][4]);//ay-p

  this->SetCov(27,vCov[2][2]);//bx-bx
  this->SetCov(34,vCov[2][3]);//bx-by
  this->SetCov(35,vCov[3][3]);//by-by

  this->SetCov(26,vCov[2][4]);//bx-p
  this->SetCov(33,vCov[3][4]);//by-p

  this->SetCov(24,vCov[0][2]);//ax-bx
  this->SetCov(25,vCov[1][2]);//ay-bx
  this->SetCov(31,vCov[0][3]);//ax-by
  this->SetCov(32,vCov[1][3]);//ay-by

  this->SetCov(20,vCov[4][4]);//p-p

  mom_estimated = true;
  return p_new;
}



float GroverVertex::Prob(void) const
{ // At the moment this just uses the CERNLIB 
  // PROB(REAL,INTEGER) function. The F77 code for this
  // routine is pasted at the end of this file, in case 
  // CERNLIB disappears (!). Alternatively one could 
  // implement the algorithm in C++. See Numerical Recipes
  // for the gamma function.
  const float error = FAILURE;
  float c2 = this->Chi2();
  if (c2 <= 0.0) return error;
  int nd = this->Ndof();
  if (nd < 1) return error;
  float prob = -1.0;
#ifndef GROVER_WITHOUT_CERNLIB
  prob = prob_(c2, nd);
#endif
  return (prob>=0.0 && prob<=1.0)  ?  prob : error;
}




void GroverVertex::Dump(void) const { // Dump members to stdout
  DumpFormatter Nice;
  Nice.DumpClassName("GroverVertex");
  Nice.DumpComment("Part of the Grover vertexing package");
  Nice.DumpComment("Please send questions about this class "
		    "to the Grover authors");
  Nice.DumpHeader();
  Nice.DumpPrint("index",&index);
  Nice.DumpPrint("x",&x);
  Nice.DumpPrint("y",&y);
  Nice.DumpPrint("z",&z);
  Nice.DumpPrint("ax",&ax);
  Nice.DumpPrint("ay",&ay);
  Nice.DumpPrint("p",&p);
  Nice.DumpPrint("m",&m);
  Nice.DumpPrint("dm",&dm);
  Nice.DumpPrint("inlocal",&inlocal);
  Nice.DumpPrint("inglobal",&inglobal);
  Nice.DumpPrint("cov",cov,COVDIM);
  Nice.DumpPrint("wprim",&wprim);
  Nice.DumpPrint("wsec",&wsec);
  Nice.DumpPrint("wnr",&wnr);
  Nice.DumpPrint("refntrincluster",&refntrincluster);
  Nice.DumpPrint("chi2",&chi2);
  Nice.DumpPrint("fitted",&fitted);
  Nice.DumpPrint("mom_estimated",&mom_estimated);
  Nice.DumpPrint("n_constraints",&n_constraints);
  Nice.DumpPrint("constr_mass",&constr_mass,CONSTDIM);
  Nice.DumpPrint("bx",&bx);
  Nice.DumpPrint("by",&by);
  Nice.DumpPrint("decay",&decay);
  Nice.DumpFooter();  
  this->dump_cons();
  cout << "*** Number of particles in particle list = " 
       << particles.size() << endl;
  cout << "*** Number of particles in trashcan      = "
       << trashcan.size()  << endl;
  cout << 
    "************************************************************************"
       << endl;
}

#ifndef GROVER_WITHOUT_ARTE

int GroverVertex::Fitp(float x0, float y0, float z0, float cv0[], RVER &newver)
{ 
  // TEST VERSION
  // 
  // Author: D.Emeliyanov / Robert Harr / Jing-GE Shiu  
  //
  // Description:
  // Fitp uses the constrain that the reconstructed particle (giving the
  // vertex through it's decay) points back to the vertex given as argument.
  // The function fits the vertex it is applied on and gives
  // back the changed parameters of the vertex given to it as argument.
  // See: ~physjg/public/pgrover/pgrover.ps.gz


  // physjg 03/27/01 
  // check the input cv0[]
  if ( (cv0[0]<0)||(cv0[2]<0)||(cv0[5]<0) ) {
    cout << "GROVER::Fitp : Negative covariant martrix element(s) at diagonal position."
         << cv0[0] << "/" << cv0[2] << "/" << cv0[5] << endl;
    return -1;
  } 

  // D.Emeliyanov 03/10/00
  // 
  static bool verbose = false;//(groverCuts.Verbose == 1);
  //static bool verbose = true;
  const int ROVERMAXTRACKS = 50;
  if (this->Size() > ROVERMAXTRACKS ||
      this->Size() < 2                ) return FAILURE;

  bool MC = false;
  bool tampered = false;
  for (GroverParticleList::const_iterator ip = this->particles.begin();
       ip != this->particles.end(); ++ip) {
    if (ip->Type() == 4) {MC = true; break;}
    if (ip->Tampered()) tampered = true;
  }
  if (MC) return FAILURE; // can't fit a vertex with > 0 perfect tracks!

  // Reset all tracks UNLESS the user has been fiddling with them
  if (!tampered) this->ResetParticles();
  //this->ResetParticles();

  if (!trashcan.empty()) trashcan.clear();

  typedef struct{
    double Zcov[5][5];
    double Pcov,mass,p,ax,ay;
    double dzeta[5];
    int flag;
    float dchi2;
    int constr[CONSTDIM]; // --> CONSTRAINTS: set to 1 if used in constraint i,
                          // 0 otherwise. CONSTDIM set in class definition.
  } TRACK_TYPE, *pTRACK_TYPE;
  
  TRACK_TYPE atr[ROVERMAXTRACKS];
  
  //  --> CONSTRAINTS
  // constraint masses are available in 
  // constr_mass[0], constr_mass[1], ... constr_mass[n_contraints-1]

  //  const int   MaxIter=5;
  const int   MaxIter=10;   // this is for the converge study
  const float sigma=0.00001;
  
  // PHYSJG : add 3x3 more for Gk and 3 more for Rk,Kk, Hm, and Ak
  //          in order to include primary information.
  double Gk[3*ROVERMAXTRACKS+3+3][3*ROVERMAXTRACKS+3+3];
  double Rk[3*ROVERMAXTRACKS+3+3];
  double Kk[3*ROVERMAXTRACKS+3+3][5];
  double Hm[3*ROVERMAXTRACKS+3+3];
  double Ak[5][3*ROVERMAXTRACKS+3+3];
  double L[5][5], L_1[5][5], L3_1[3][3];
  double dchi2t,Mest,Merr;
    
  /////////////////
  float zref=0.0;// <- note!
  /////////////////
  
  double zeta[5],zetapvx[3];
  double Px,Py,Pz,nx,ny,nz,P,E,epsilon; 
  double epsilonx, epsilony, epsilonz;
  int i,j,k,l,ndim,ndim2,jjj,jcons;

  // PHYSJG : add variables for pointing constrain
  double x2,y2,z2,xpri,ypri,zpri,pvxchi2;
  const float sigmap = 0.0001;
  
  GroverParticleList::iterator trackincluster[ROVERMAXTRACKS] ={0};
  
  i=1;
  for (GroverParticleList::iterator it1 = particles.begin(); 
       it1 != particles.end(); ++it1) {
    trackincluster[i] = it1;
    it1->PropagateBeginTo(zref);
    i++;
  }
  int ntrincluster = i-1;
  ndim=3+3*ntrincluster;
  ndim2=ndim+3;         // PHYSJG : I got to add 3 more for primary
    
  for (i=1; i<=ntrincluster; i++) {
    GroverParticleList::iterator it = trackincluster[i];
    for(int v=0;v<5;v++) for(int vv=0;vv<5;vv++) atr[i].Zcov[v][vv]=0.0;
    atr[i].Zcov[0][0] =  it->BeginCov(0);
    atr[i].Zcov[1][0] = atr[i].Zcov[0][1] =  it->BeginCov(1);
    atr[i].Zcov[2][0] = atr[i].Zcov[0][2] =  it->BeginCov(3);
    atr[i].Zcov[3][0] = atr[i].Zcov[0][3] =  it->BeginCov(6);
    atr[i].Zcov[2][1] = atr[i].Zcov[1][2] =  it->BeginCov(4);
    atr[i].Zcov[3][1] = atr[i].Zcov[1][3] =  it->BeginCov(7);
    atr[i].Zcov[1][1] = it->BeginCov(2);
    atr[i].Zcov[2][2] = it->BeginCov(5);
    atr[i].Zcov[2][3] = atr[i].Zcov[3][2] =  it->BeginCov(8);
    atr[i].Zcov[3][3] = it->BeginCov(9);
    atr[i].Zcov[4][4] = it->BeginCov(14);
    atr[i].mass       = it->Mass();
    atr[i].dzeta[0]   = it->BeginX();
    atr[i].dzeta[1]   = it->BeginY();
    atr[i].dzeta[2]   = it->BeginSlopeX();
    atr[i].dzeta[3]   = it->BeginSlopeY();
    atr[i].dzeta[4]   = it->Momentum();
    atr[i].ax         = it->BeginSlopeX();
    atr[i].ay         = it->BeginSlopeY();
    atr[i].p          = it->Momentum();
    atr[i].flag       = 1;

    //  --> CONSTRAINTS
    for (int j=0; j<CONSTDIM; j++) 
      atr[i].constr[j] = constr_map[j][i-1];

    Rk[3*i]=atr[i].ax;   Rk[3*i+1]=atr[i].ay;   Rk[3*i+2]=atr[i].p;
  }

  Rk[0]= this->X();   Rk[1]=this->Y();  Rk[2]=this->Z();
  // for primary position
  Rk[3*ntrincluster+3] = x0;        
  Rk[3*ntrincluster+4] = y0;        
  Rk[3*ntrincluster+5] = z0;        

    /****************************************/  
  for (jjj=1;jjj<=MaxIter;jjj++) {//MaxIter
    
    for(i=0;i<ndim2;i++)  for(j=0;j<ndim2;j++)  Gk[i][j]=0.0;    
    if (jjj==1) {
      Gk[0][0]=100.0;//cov[0];  
      Gk[1][1]=100.0;//cov[5];
      Gk[2][2]=100.0;//cov[9];
    } else {
      Gk[0][0]=100.0;//cov[0];  
      Gk[1][1]=100.0;//cov[5];
      Gk[2][2]=100.0;//cov[9];
    }    

    for(i=1;i<=ntrincluster;i++) {
      Gk[3*i][3*i]    = 1.0; 
      Gk[3*i+1][3*i+1]= 1.0; 
      Gk[3*i+2][3*i+2]= 100.0;
    }
    
    for(i=1;i<=ntrincluster;i++) {
      L[0][0] = Gk[0][0] + Rk[3*i]*Rk[3*i]*Gk[2][2] + Rk[2]*Rk[2]*Gk[3*i][3*i]
	      - 2.0*(Rk[3*i]*Gk[0][2] + 
		     Rk[2]*Gk[0][3*i] - 
		     Rk[3*i]*Rk[2]*Gk[2][3*i]);
      L[1][1] = Gk[1][1] + Rk[3*i+1]*Rk[3*i+1]*Gk[2][2]
               +Rk[2]*Rk[2]*Gk[3*i+1][3*i+1]
	       -2.0*(Rk[3*i+1]*Gk[1][2] +
		     Rk[2]*Gk[1][3*i+1] - 
		     Rk[2]*Rk[3*i+1]*Gk[2][3*i+1]);
      L[0][1] = Gk[0][1] + Rk[3*i]*Rk[3*i+1]*Gk[2][2]
	       +Rk[2]*Rk[2]*Gk[3*i][3*i+1]
	       -Rk[2]*(Gk[1][3*i] + Gk[0][3*i+1])
	       +Rk[2]*(Rk[3*i]*Gk[2][3*i+1]+Rk[3*i+1]*Gk[2][3*i])
	       -Rk[3*i]*Gk[1][2] 
               -Rk[3*i+1]*Gk[0][2];
      L[1][0] = L[0][1];
      L[2][2] = Gk[3*i][3*i];
      L[3][3] = Gk[3*i+1][3*i+1];
      L[2][3] = Gk[3*i][3*i+1];
      L[3][2] = L[2][3];
      L[0][2] = Gk[0][3*i]   -Rk[3*i]*Gk[2][3*i]    -Rk[2]*Gk[3*i][3*i];
      L[2][0] = L[0][2];
      L[0][3] = Gk[0][3*i+1] -Rk[3*i]*Gk[2][3*i+1]  -Rk[2]*Gk[3*i][3*i+1];
      L[3][0] = L[0][3];
      L[1][2] = Gk[1][3*i]   -Rk[3*i+1]*Gk[2][3*i]  -Rk[2]*Gk[3*i][3*i+1];
      L[2][1] = L[1][2];
      L[1][3] = Gk[1][3*i+1] -Rk[3*i+1]*Gk[2][3*i+1]-Rk[2]*Gk[3*i+1][3*i+1];
      L[3][1] = L[1][3];
      L[0][4] = Gk[0][3*i+2] -Rk[3*i]*Gk[2][3*i+2]  -Rk[2]*Gk[3*i][3*i+2];
      L[4][0] = L[0][4];
      L[1][4] = Gk[1][3*i+2] -Rk[3*i+1]*Gk[2][3*i+2]-Rk[2]*Gk[3*i+1][3*i+2];
      L[4][1] = L[1][4];
      L[2][4] = Gk[3*i][3*i+2];
      L[4][2] = L[2][4];
      L[3][4] = Gk[3*i+1][3*i+2];
      L[4][3] = L[3][4];
      L[4][4] = Gk[3*i+2][3*i+2];
        
      for(k=0;k<5;++k)  
	for(l=0;l<5;++l) {
	  L[k][l]   = L[k][l] + atr[i].Zcov[k][l];
	  L_1[k][l] = L[k][l];
	}

      five_dim_inv(L_1);

      for(j=0;j<ndim;++j) {
	Ak[0][j] = Gk[j][0] - Rk[3*i]*Gk[j][2]   - Rk[2]*Gk[j][3*i];
	Ak[1][j] = Gk[1][j] - Rk[3*i+1]*Gk[j][2] - Rk[2]*Gk[j][3*i+1];
	Ak[2][j] = Gk[j][3*i];
	Ak[3][j] = Gk[j][3*i+1];
	Ak[4][j] = Gk[j][3*i+2];
	for(int v=0;v<5;++v) {
	  Kk[j][v] = 0.0;
	  for(int vv=0;vv<5;++vv) Kk[j][v] += Ak[vv][j]*L_1[vv][v];
	}
      }
      zeta[0] = atr[i].dzeta[0] - Rk[0] + Rk[3*i]*Rk[2];
      zeta[1] = atr[i].dzeta[1] - Rk[1] + Rk[3*i+1]*Rk[2];
      zeta[2] = atr[i].dzeta[2] - Rk[3*i];
      zeta[3] = atr[i].dzeta[3] - Rk[3*i+1];
      zeta[4] = atr[i].dzeta[4] - Rk[3*i+2];
      for(j=0;j<ndim;++j) 
	for(int v=0;v<5;++v) 
	  Rk[j] += Kk[j][v]*zeta[v];
      for(k=0;k<ndim;++k) 
	for(l=0;l<ndim;++l) 
	  for(int v=0;v<5;v++)
	    Gk[k][l] -= Kk[k][v]*Ak[v][l];
    }

    // mass constraints : 
    if(n_constraints)
      for(jcons=0;jcons<n_constraints;jcons++)
        {
          for(i=0;i<ndim;i++) Hm[i]=0.0;
          Px=0.0;Py=0.0;   Pz=0.0;   E=0.0;
          for(i=1;i<=ntrincluster;i++) if(atr[i].constr[jcons]) 
            {
              nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
              nx = Rk[3*i]*nz;
              ny = Rk[3*i+1]*nz;
              Px += Rk[3*i+2]*nx;
              Py += Rk[3*i+2]*ny;
              Pz += Rk[3*i+2]*nz;
              E  += sqrt(Rk[3*i+2]*Rk[3*i+2]+atr[i].mass*atr[i].mass);
            }    
          P = sqrt(Px*Px+Py*Py+Pz*Pz);
          for(i=1;i<=ntrincluster;i++) if(atr[i].constr[jcons]) 
            {
              nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
              nx = Rk[3*i]*nz;
              ny = Rk[3*i+1]*nz;
              Hm[3*i]  = -2.0*Rk[3*i+2]*nz*((1.0-nx*nx)*Px-nx*ny*Py-nx*nz*Pz);
              Hm[3*i+1]= -2.0*Rk[3*i+2]*nz*(-nx*ny*Px+(1.0-ny*ny)*Py-ny*nz*Pz);
              Hm[3*i+2]=  2.0*Rk[3*i+2]*E/sqrt(Rk[3*i+2]*Rk[3*i+2]+atr[i].mass*atr[i].mass)-
                2*(nx*Px+ny*Py+nz*Pz);
            }
          epsilon = -E*E + P*P + constr_mass[jcons]*constr_mass[jcons];
          if (verbose) cout<<"epsilon = "<<epsilon<<endl;
          for(i=0;i<ndim;i++) {
            Ak[0][i] = 0.0;
            for(j=0;j<ndim;j++) Ak[0][i] += Hm[j]*Gk[j][i];
          }
          L[0][0] = 0.0;
          for(i=0;i<ndim;i++) L[0][0] += Ak[0][i]*Hm[i];
          L[0][0] += sigma*sigma;
          L_1[0][0] = 1.0/L[0][0];
          for(i=0;i<ndim;i++) Kk[i][0] = Ak[0][i]*L_1[0][0];
          for(i=0;i<ndim;i++) Rk[i]   += Kk[i][0]*epsilon;
          for(i=0;i<ndim;i++)
            for(j=0;j<ndim;j++) Gk[i][j] -= Ak[0][i]*Kk[j][0];
        }


    //Now, try my pointing constrain.  In x first. 
    // start from here, I am working with ndim2=ndim+3 dimension
    // Zeroth step, to fill the primary covariance matrix information
      
    Rk[3*ntrincluster+3] = x0;        
    Rk[3*ntrincluster+4] = y0;        
    Rk[3*ntrincluster+5] = z0;        
    Gk[ndim][ndim] = cv0[0];
    Gk[ndim+1][ndim+1] = cv0[2];
    Gk[ndim+2][ndim+2] = cv0[5];
    Gk[ndim][ndim+1] = (Gk[ndim+1][ndim] = cv0[1]);
    Gk[ndim][ndim+2] = (Gk[ndim+2][ndim] = cv0[3]);
    Gk[ndim+1][ndim+2] = (Gk[ndim+2][ndim+1] = cv0[4]);

    // first, in x.   
          for(i=0;i<ndim2;i++) Hm[i]=0.0;
          Px=0.0; Py=0.0; Pz=0.0; 
          x2 = Rk[0]; y2 = Rk[1]; z2 = Rk[2];          
          xpri = Rk[3*ntrincluster+3];
          ypri = Rk[3*ntrincluster+4];
          zpri = Rk[3*ntrincluster+5];

          for(i=1;i<=ntrincluster;i++) 
            {
              nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
              nx = Rk[3*i]*nz;
              ny = Rk[3*i+1]*nz;
              Px += Rk[3*i+2]*nx;
              Py += Rk[3*i+2]*ny;
              Pz += Rk[3*i+2]*nz;
            }    
//          epsilonx = Pz*(y2-ypri) - Py*(z2-zpri);
//          epsilony = Px*(z2-zpri) - Pz*(x2-xpri);
//          epsilonz = Py*(x2-xpri) - Px*(y2-ypri);

          for(i=1;i<=ntrincluster;i++) 
            {
              nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
              nx = Rk[3*i]*nz;
              ny = Rk[3*i+1]*nz;
              Hm[3*i]   = Rk[3*i+2]*nz*(nx*nz*(y2-ypri)-nx*ny*(z2-zpri));
              Hm[3*i+1] = Rk[3*i+2]*nz*(ny*nz*(y2-ypri)+(1-ny*ny)*(z2-zpri));
              Hm[3*i+2] = ny*(z2-zpri) - nz*(y2-ypri);
            }
          Hm[0] = Hm[3*ntrincluster+3] = 0 ;
          Hm[1] = -(Hm[3*ntrincluster+4] = Pz);
          Hm[2] = -(Hm[3*ntrincluster+5] = -Py);

          epsilonx = epsilon = Pz*(y2-ypri) - Py*(z2-zpri);
//          errx[jjj-1]=epsilon;
          if (verbose) cout<<"x epsilon = "<<epsilon<<endl;

          for(i=0;i<ndim2;i++) {
            Ak[0][i] = 0.0;
            for(j=0;j<ndim2;j++) Ak[0][i] += Hm[j]*Gk[j][i];
          }
          L[0][0] = 0.0;
          for(i=0;i<ndim2;i++) L[0][0] += Ak[0][i]*Hm[i];
          L[0][0] += sigmap*sigmap;
          L_1[0][0] = 1.0/L[0][0];
          for(i=0;i<ndim2;i++) Kk[i][0] = Ak[0][i]*L_1[0][0];
          for(i=0;i<ndim2;i++) Rk[i]   += Kk[i][0]*epsilon;
          for(i=0;i<ndim2;i++)
            for(j=0;j<ndim2;j++) Gk[i][j] -= Ak[0][i]*Kk[j][0];

    // second, in y.   
          for(i=0;i<ndim2;i++) Hm[i]=0.0;
          Px=0.0;Py=0.0;   Pz=0.0;   E=0.0;
          xpri = Rk[3*ntrincluster+3];
          ypri = Rk[3*ntrincluster+4];
          zpri = Rk[3*ntrincluster+5];
          x2 = Rk[0]; y2 = Rk[1]; z2 = Rk[2];          

          for(i=1;i<=ntrincluster;i++) 
            {
              nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
              nx = Rk[3*i]*nz;
              ny = Rk[3*i+1]*nz;
              Px += Rk[3*i+2]*nx;
              Py += Rk[3*i+2]*ny;
              Pz += Rk[3*i+2]*nz;
            }    

          for(i=1;i<=ntrincluster;i++) 
            {
              nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
              nx = Rk[3*i]*nz;
              ny = Rk[3*i+1]*nz;
              Hm[3*i]   = -Rk[3*i+2]*nz*(nx*nz*(x2-xpri)+(1-nx*nx)*(z2-zpri));
              Hm[3*i+1] = -Rk[3*i+2]*nz*(ny*nz*(x2-xpri)-nx*ny*(z2-zpri));
              Hm[3*i+2] = nz*(x2-xpri) - nx*(z2-zpri);
            }
          Hm[0] = -(Hm[3*ntrincluster+3] = -Pz);
          Hm[1] = Hm[3*ntrincluster+4] = 0;
          Hm[2] = -(Hm[3*ntrincluster+5] = Px);

          epsilonx = Pz*(y2-ypri) - Py*(z2-zpri);
          epsilony = epsilon = Px*(z2-zpri) - Pz*(x2-xpri);
//          erry[jjj-1]=epsilon;
          if (verbose) cout<<"y epsilon = "<<epsilon<<endl;

          for(i=0;i<ndim2;i++) {
            Ak[0][i] = 0.0;
            for(j=0;j<ndim2;j++) Ak[0][i] += Hm[j]*Gk[j][i];
          }
          L[0][0] = 0.0;
          for(i=0;i<ndim2;i++) L[0][0] += Ak[0][i]*Hm[i];
          L[0][0] += sigmap*sigmap;
          L_1[0][0] = 1.0/L[0][0];
          for(i=0;i<ndim2;i++) Kk[i][0] = Ak[0][i]*L_1[0][0];
          for(i=0;i<ndim2;i++) Rk[i]   += Kk[i][0]*epsilon;
          for(i=0;i<ndim2;i++)
            for(j=0;j<ndim2;j++) Gk[i][j] -= Ak[0][i]*Kk[j][0];

    // finally, in z
          for(i=0;i<ndim2;i++) Hm[i]=0.0;
          Px=0.0;Py=0.0;   Pz=0.0;   E=0.0;
          x2 = Rk[0]; y2 = Rk[1]; z2 = Rk[2];          
          xpri = Rk[3*ntrincluster+3];
          ypri = Rk[3*ntrincluster+4];
          zpri = Rk[3*ntrincluster+5];

          for(i=1;i<=ntrincluster;i++) 
            {
              nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
              nx = Rk[3*i]*nz;
              ny = Rk[3*i+1]*nz;
              Px += Rk[3*i+2]*nx;
              Py += Rk[3*i+2]*ny;
              Pz += Rk[3*i+2]*nz;
            }    

          for(i=1;i<=ntrincluster;i++) 
            {
              nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
              nx = Rk[3*i]*nz;
              ny = Rk[3*i+1]*nz;
              Hm[3*i]   = Rk[3*i+2]*nz*(nx*ny*(x2-xpri)+(1-nx*nx)*(y2-ypri));
              Hm[3*i+1] = -Rk[3*i+2]*nz*(nx*ny*(y2-ypri)+(1-ny*ny)*(x2-xpri));
              Hm[3*i+2] = nx*(y2-ypri) - ny*(x2-xpri);
            }
          Hm[0] = -(Hm[3*ntrincluster+3] = Py);
          Hm[1] = -(Hm[3*ntrincluster+4] = -Px);
          Hm[2] = Hm[3*ntrincluster+5] = 0;

          epsilonz = epsilon = Py*(x2-xpri) - Px*(y2-ypri);
//          errz[jjj-1]=epsilon;
          if (verbose) cout<<"z epsilon = "<<epsilon<<endl;

          for(i=0;i<ndim2;i++) {
            Ak[0][i] = 0.0;
            for(j=0;j<ndim2;j++) Ak[0][i] += Hm[j]*Gk[j][i];
          }
          L[0][0] = 0.0;
          for(i=0;i<ndim2;i++) L[0][0] += Ak[0][i]*Hm[i];
          L[0][0] += sigmap*sigmap;
          L_1[0][0] = 1.0/L[0][0];
          for(i=0;i<ndim2;i++) Kk[i][0] = Ak[0][i]*L_1[0][0];
          for(i=0;i<ndim2;i++) Rk[i]   += Kk[i][0]*epsilon;
          for(i=0;i<ndim2;i++)
            for(j=0;j<ndim2;j++) Gk[i][j] -= Ak[0][i]*Kk[j][0];

  }

      /*
  if(n_constraints==0)
    {
      for(i=0;i<ndim;i++) Hm[i]=0.0;
      Px=0.0;   Py=0.0;   Pz=0.0;   E=0.0;
      for(i=1;i<=ntrincluster;i++) 
        {
          nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
          nx = Rk[3*i]*nz;
          ny = Rk[3*i+1]*nz;
          Px += Rk[3*i+2]*nx;
          Py += Rk[3*i+2]*ny;
          Pz += Rk[3*i+2]*nz;
          E  += sqrt(Rk[3*i+2]*Rk[3*i+2]+atr[i].mass*atr[i].mass);
        }
      P = sqrt(Px*Px+Py*Py+Pz*Pz);
      for(i=1;i<=ntrincluster;i++) 
        {
          nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
          nx = Rk[3*i]*nz;
          ny = Rk[3*i+1]*nz;
          Hm[3*i]  = -2.0*Rk[3*i+2]*nz*((1.0-nx*nx)*Px-nx*ny*Py-nx*nz*Pz);
          Hm[3*i+1]= -2.0*Rk[3*i+2]*nz*(-nx*ny*Px+(1.0-ny*ny)*Py-ny*nz*Pz);
          Hm[3*i+2]=  2.0*Rk[3*i+2]*E/sqrt(Rk[3*i+2]*Rk[3*i+2]+atr[i].mass*atr[i].mass)-
            2*(nx*Px+ny*Py+nz*Pz);
        }
      for(i=0;i<ndim;i++) {
        Ak[0][i] = 0.0;
        for(j=0;j<ndim;j++) Ak[0][i] += Hm[j]*Gk[j][i];
      }
      L[0][0] = 0.0;for(i=0;i<ndim;i++) L[0][0] += Ak[0][i]*Hm[i];
      L_1[0][0] = 1.0/L[0][0];
      Mest=sqrt(-P*P+E*E);  Merr=0.5/sqrt(L_1[0][0]*(-P*P+E*E));
      if (verbose) cout<<"Fitted mass = "<<Mest<<" Error = "<<Merr<<endl;
      this->SetMass(Mest);this->SetMassError(Merr);
    }
  else
    {
      this->SetMass(constr_mass[0]);this->SetMassError(sigma);
    }
      */

  // estimation of the mass and mass error (using the non-mass constraint formula)
      for(i=0;i<ndim2;i++) Hm[i]=0.0;
      Px=0.0;   Py=0.0;   Pz=0.0;   E=0.0;
      for(i=1;i<=ntrincluster;i++) 
        {
          nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
          nx = Rk[3*i]*nz;
          ny = Rk[3*i+1]*nz;
          Px += Rk[3*i+2]*nx;
          Py += Rk[3*i+2]*ny;
          Pz += Rk[3*i+2]*nz;
          E  += sqrt(Rk[3*i+2]*Rk[3*i+2]+atr[i].mass*atr[i].mass);
        }
      P = sqrt(Px*Px+Py*Py+Pz*Pz);
      for(i=1;i<=ntrincluster;i++) 
        {
          nz = 1.0/sqrt(1.0+Rk[3*i]*Rk[3*i]+Rk[3*i+1]*Rk[3*i+1]);
          nx = Rk[3*i]*nz;
          ny = Rk[3*i+1]*nz;
          Hm[3*i]  = -2.0*Rk[3*i+2]*nz*((1.0-nx*nx)*Px
                                       -nx*ny*Py
                                       -nx*nz*Pz);
          Hm[3*i+1]= -2.0*Rk[3*i+2]*nz*(-nx*ny*Px
                                        +(1.0-ny*ny)*Py
                                        -ny*nz*Pz);
          Hm[3*i+2]=  2.0*Rk[3*i+2]
                      *E/sqrt(Rk[3*i+2]*Rk[3*i+2]+atr[i].mass*atr[i].mass)
                     -2*(nx*Px+ny*Py+nz*Pz);
        }
      for(i=0;i<ndim2;i++) {
        Ak[0][i] = 0.0;
        for(j=0;j<ndim2;j++) Ak[0][i] += Hm[j]*Gk[j][i];
      }
      L[0][0] = 0.0;for(i=0;i<ndim2;i++) L[0][0] += Ak[0][i]*Hm[i];
      L_1[0][0] = 1.0/L[0][0];
      Mest=sqrt(-P*P+E*E);  Merr=0.5/sqrt(L_1[0][0]*(-P*P+E*E));
      if (verbose) cout<<"Fitted mass = "<<Mest<<" Error = "<<Merr<<endl;
      this->SetMass(Mest);this->SetMassError(Merr);


  // final procedure to estimate the chi2 contributed from tracks and primary.
  for (i=1;i<=ntrincluster;++i) {
    L[0][0]= Gk[0][0] 
            +Rk[3*i]*Rk[3*i]*Gk[2][2] 
            +Rk[2]*Rk[2]*Gk[3*i][3*i]
            -2.0*( Rk[3*i]*Gk[0][2] 
		  +Rk[2]*Gk[0][3*i] 
		  -Rk[3*i]*Rk[2]*Gk[2][3*i] );
    L[1][1]= Gk[1][1]
            +Rk[3*i+1]*Rk[3*i+1]*Gk[2][2]
            +Rk[2]*Rk[2]*Gk[3*i+1][3*i+1]
            -2.0*( Rk[3*i+1]*Gk[1][2]
                  +Rk[2]*Gk[1][3*i+1]
		  -Rk[2]*Rk[3*i+1]*Gk[2][3*i+1] );
    L[0][1]= Gk[0][1]
            +Rk[3*i]*Rk[3*i+1]*Gk[2][2]
            +Rk[2]*Rk[2]*Gk[3*i][3*i+1]
            -Rk[2]*( Gk[1][3*i] 
		    +Gk[0][3*i+1] )
            +Rk[2]*( Rk[3*i]*Gk[2][3*i+1] 
		    +Rk[3*i+1]*Gk[2][3*i] )
            -Rk[3*i]*Gk[1][2] 
            -Rk[3*i+1]*Gk[0][2];
    L[2][2]= Gk[3*i][3*i];
    L[3][3]= Gk[3*i+1][3*i+1];
    L[2][3]= Gk[3*i][3*i+1];
    L[0][2]= Gk[0][3*i]
            -Rk[3*i]*Gk[2][3*i]
            -Rk[2]*Gk[3*i][3*i];
    L[0][3]= Gk[0][3*i+1]
            -Rk[3*i]*Gk[2][3*i+1]
            -Rk[2]*Gk[3*i][3*i+1];
    L[1][2]= Gk[1][3*i]
            -Rk[3*i+1]*Gk[2][3*i]
            -Rk[2]*Gk[3*i][3*i+1];
    L[1][3]= Gk[1][3*i+1]
            -Rk[3*i+1]*Gk[2][3*i+1]
            -Rk[2]*Gk[3*i+1][3*i+1];
    L[0][4]= Gk[0][3*i+2]
            -Rk[3*i]*Gk[2][3*i+2]
            -Rk[2]*Gk[3*i][3*i+2];
    L[1][4]= Gk[1][3*i+2]
            -Rk[3*i+1]*Gk[2][3*i+2]
            -Rk[2]*Gk[3*i+1][3*i+2];
    L[2][4]= Gk[3*i][3*i+2];
    L[3][4]= Gk[3*i+1][3*i+2];
    L[4][4]= Gk[3*i+2][3*i+2];    
    L[1][0] = L[0][1];
    L[3][2] = L[2][3];
    L[2][0] = L[0][2];
    L[3][0] = L[0][3];
    L[2][1] = L[1][2];
    L[3][1] = L[1][3];  
    L[4][0] = L[0][4];
    L[4][1] = L[1][4];
    L[4][2] = L[2][4];
    L[4][3] = L[3][4];
    for (k=0;k<5;++k)  
      for (l=0;l<5;++l)  
	L_1[k][l] = L[k][l]+atr[i].Zcov[k][l];
    five_dim_inv(L_1);
    zeta[0] = atr[i].dzeta[0] - Rk[0] + Rk[3*i]*Rk[2];
    zeta[1] = atr[i].dzeta[1] - Rk[1] + Rk[3*i+1]*Rk[2];
    zeta[2] = atr[i].dzeta[2] - Rk[3*i];
    zeta[3] = atr[i].dzeta[3] - Rk[3*i+1];
    zeta[4] = atr[i].dzeta[4] - Rk[3*i+2];
    atr[i].dchi2 = 0.0;
    for(int v=0;v<5;++v) 
      for(int vv=0;vv<5;++vv) 
	atr[i].dchi2 += zeta[v]*zeta[vv]*L_1[v][vv];
  }


  L3_1[0][0] = cv0[0]+Gk[ndim2-3][ndim2-3];  
  L3_1[0][1] =  L3_1[1][0] = cv0[1]+Gk[ndim2-3][ndim2-2];  
  L3_1[1][1] = cv0[2]+Gk[ndim2-2][ndim2-2];  
  L3_1[0][2] = L3_1[2][0] = cv0[3]+Gk[ndim2-3][ndim2-1];  
  L3_1[1][2] = L3_1[2][1] = cv0[4]+Gk[ndim2-2][ndim2-1];  
  L3_1[2][2] = cv0[5]+Gk[ndim2-1][ndim2-1];  
  zetapvx[0] = x0 - Rk[3*ntrincluster+3]; 
  zetapvx[1] = y0 - Rk[3*ntrincluster+4]; 
  zetapvx[2] = z0 - Rk[3*ntrincluster+5]; 

  three_dim_inv(L3_1);

  pvxchi2 = 0;
  for (int v=0;v<3;v++) for(int vv=0;vv<3;vv++)
    pvxchi2+=zetapvx[v]*zetapvx[vv]*L3_1[v][vv];

  
  if (verbose) cout<<"ntr = "<<ntrincluster<<endl;  
  
  this->SetX(Rk[0]);   this->SetY(Rk[1]);   this->SetZ(Rk[2]);
  this->SetCov(0, Gk[0][0]);
  this->SetCov(1, Gk[0][1]);
  this->SetCov(2, Gk[1][1]);
  this->SetCov(3, Gk[0][2]);     
  this->SetCov(4, Gk[1][2]);
  this->SetCov(5, Gk[2][2]);
  
  if (verbose) 
  {
      cout<<"PCONSTR.FIT: Chi2 contrib. from primary  = " << pvxchi2<< endl;
    for(i=1;i<=ntrincluster;i++) 
      cout<<"PCONSTR.FIT: Chi2 contrib. from track "<<i<<" = " <<atr[i].dchi2<<endl;
  }


  //  dchi2t=0.0;
  dchi2t=pvxchi2;
  for(i=1;i<=ntrincluster;i++) dchi2t+=atr[i].dchi2;
   
  if (verbose) cout<<"Total chi2 of the fit = "<<dchi2t<<endl;
  this->SetChi2(dchi2t);
  // fill updated info for each track
  for(i=1; i<=ntrincluster; i++) {
    GroverParticleList::iterator it = trackincluster[i];
    atr[i].ax = Rk[3*i];
    atr[i].ay = Rk[3*i+1];
    atr[i].p  = Rk[3*i+2];
    it->PropagateBeginTo(this->Z());
    it->SetBeginX(this->X()); 
    it->SetBeginY(this->Y());
    it->SetBeginSlopeX(atr[i].ax);
    it->SetBeginSlopeY(atr[i].ay);
    it->SetMomentum(atr[i].p);
    it->SetChi2(atr[i].dchi2);
  }


  // and the new ver
    newver.xv = Rk[3*ntrincluster+3];
    newver.yv = Rk[3*ntrincluster+4];
    newver.zv = Rk[3*ntrincluster+5];
    newver.cv[0] = Gk[ndim][ndim];
    newver.cv[1] = Gk[ndim][ndim+1];
    newver.cv[2] = Gk[ndim+1][ndim+1];
    newver.cv[3] = Gk[ndim][ndim+2];
    newver.cv[4] = Gk[ndim+1][ndim+2];
    newver.cv[5] = Gk[ndim+2][ndim+2];
    newver.chi2  = pvxchi2;
  
  gcov.clear();  
  for(int i=0; i<ndim2; i++) {
    for(int j=0; j<=i; j++) {
      gcov.push_back(Gk[i][j]);
    }
  }

  fitted = true;
  return SUCCESS;
}
#endif

#ifndef GROVER_WITHOUT_ARTE
int GroverVertex::Fitp(ArtePointer<RVER> rver, RVER &newver)
{ 
  // TEST VERSION
  // 
  // Author: Jing-GE Shiu/D.Emeliyanov
  // see: 
  // int GroverVertex::Fitp(float x0, float y0, float z0, float cv0[], RVER &newver)

  float x = rver->xv;
  float y = rver->yv;
  float z = rver->zv;
  float cv[6]; 
  for (int i=0;i<6;i++) cv[i]=rver->cv[i];

  return this->Fitp(x,y,z,cv,newver);
}

#endif



//////////////////////////////////////////////////////////////////////
//
// Operators
//

bool operator==(const GroverVertex& a, const GroverVertex& b)
{
  if (a.Size() != b.Size()) return false;
  if (a.Size() == 0) {
    // special case for two empty vertices constructed from the
    // same RVER (without RREL->RTRA) and not changed since
    bool covsame = true;
    for (int i = 0; i<6; ++i) {
      if (a.Cov(i) != b.Cov(i)) covsame = false;
    }
    if (covsame              &&
	a.X()    != 0.0      && 
	a.Y()    != 0.0      && 
	a.Z()    != 0.0      &&
	a.Mass() != 0.0      && 
	a.Chi2() != 0.0      && 
	a.X()    == b.X()    && 
	a.Y()    == b.Y()    && 
	a.Z()    == b.Z()    &&
	a.Mass() == b.Mass() && 
	a.Chi2() == b.Chi2()   ) 
      return true;
    else return false; 
  }

  // all other cases
  bool flag = true;
  for (GroverParticleList::const_iterator ia = a.particles.begin();
       ia != a.particles.end(); ++ia) {
    bool found_this_one = false;
    for (GroverParticleList::const_iterator ib = b.particles.begin();
	 ib != b.particles.end(); ++ib) {
      if (*ia == *ib) {found_this_one = true; break;}
    }
    if (!found_this_one) {flag = false; break;}
  }
  return flag;
}


bool operator!=(const GroverVertex& a, const GroverVertex& b)
{
  return !(a==b);
}

ostream& operator<<(ostream& s, GroverVertex& v)
{ // non-const because of use of v.Momentum()
  s << "[";
  s << "idx=" << v.Index();
  s << " x,y,z={" << v.X() << "," << v.Y() << "," << v.Z() << "}";
  float dx2 = v.Cov(0), dy2 = v.Cov(2), dz2 = v.Cov(5);
  float dx = (dx2 >= 0.0) ? sqrt(dx2) : 0.0;
  float dy = (dy2 >= 0.0) ? sqrt(dy2) : 0.0;
  float dz = (dz2 >= 0.0) ? sqrt(dz2) : 0.0;
  s << " dx,dy,dz={" << dx << "," << dy << "," << dz << "}";
  s << " m,dm,p={"   << v.Mass() << "," << v.MassError() << "," << v.Momentum() << "}";
  s << " ntr=" << v.Size();
  if (v.Size() > 0) {
    s << "(idx=";
    GroverParticleList::const_iterator first = v.particles.begin();
    for (GroverParticleList::const_iterator it = v.particles.begin();
	 it != v.particles.end(); ++it) {
      if (it != first) s << ",";
      s << it->Index();
    }
    s << ")";
  }
  s << " chi2=" << v.Chi2();
  s << "(" << v.Ndof() << " dof)";
  if (v.InLocal()) s << " crd=LOC";
  else s << " crd=GLO";
  s << "]";
  return s;
}





////////////////////////////////////////////////////////////////////////
//
// (slightly hacked and commented) code for the CERNLIB routine 
// PROB(REAL, INTEGER)
// If you want to run this code, replace (HASH) with # 
// and copy it to another file. You will have to find the
// include files and architecture differences etc. yourself
//


/*


(HASH)include "gen/pilot.h"
*     just some preprocessor logic to determine whether to define certain
*     tokens according to others (architecture, OS, sizeof() etc.) (MS)


      FUNCTION PROB(X,N)

*     ******************************************************************
*     ******************************************************************
*     **                                                              **
*     **   Function to return the probability of observing a chi2     **
*     **   greater than, or equal to, X in a fit with N degrees of    **
*     **   freedom.                                                   **
*     **                                                              **
*     **   From CERNLIB  www.cern.ch                                  **
*     **                                                              **
*     **   CALLS: SQRT, DERFC/ERFC, MTLPRT  (MS)                      **
*     **                                                              **
*     ******************************************************************
*     ******************************************************************

(HASH)include "gen/imp64.inc" // some definitions for 64 bit machines (MS)

*     This code does not use the sane IMPLICIT NONE standard. Therefore
*     all un-typed variables are typed by the compiler according to
*     their first letter. (!!!)
*     This include file defines the types of variables starting with A-H,I,O-Z
*     as one of COMPLEX, COMPLEX(2) or COMPLEX*16
*     according to some of the flags set in pilot.h (MS)
      REAL PROB,X
      CHARACTER NAME*(*)
      CHARACTER*80 ERRTXT
      PARAMETER (NAME = 'PROB')
      PARAMETER (R1 = 1, HF = R1/2, TH = R1/3, F1 = 2*R1/9)
      PARAMETER (C1 = 1.12837 91670 95513D0)
      PARAMETER (NMAX = 300)
*     maximum chi2 per df for df >= 2., if chi2/df > chipdf prob=0.
      PARAMETER (CHIPDF = 100.)
      PARAMETER (XMAX = 174.673, XMAX2 = 2*XMAX)


*     Some limits for different architectures and precisions: (MS)
(HASH)if defined(CERNLIB_IBM)
      PARAMETER (XLIM = 13.3)
(HASH)endif
(HASH)if !defined(CERNLIB_IBM)
      PARAMETER (XLIM = 24.)
(HASH)endif
      PARAMETER (EPS = 1D-30)
(HASH)if defined(CERNLIB_DOUBLE)
      GERFC(V)=DERFC(V)
(HASH)endif
(HASH)if !defined(CERNLIB_DOUBLE)
      GERFC(V)= ERFC(V)
(HASH)endif

*     (first example of an evil IMPLICIT complex ;-) (MS)
      Y = X
      U = HF*Y

*     First, some error handling (MS)
      IF (N.LE.0) THEN
         H = 0
         WRITE(ERRTXT,101) N
         CALL MTLPRT(NAME,'G100.1',ERRTXT)
      ELSEIF (Y.LT.0) THEN
         H = 0
         WRITE(ERRTXT,102) X
         CALL MTLPRT(NAME,'G100.2',ERRTXT)

*     The real code starts here (MS)
      ELSEIF (Y.EQ.0.OR.(N/20.GT.Y)) THEN
*        Extremely small chi2/dof -> P=1 (MS)
         H = 1

      ELSEIF (N.EQ.1) THEN
*        One dof -> P = erf(sqrt(chi2/2)) (MS)
         W = SQRT(U)
         IF (W.LT.XLIM) THEN
            H = GERFC(W)
         ELSE
            H = 0
         ENDIF

      ELSEIF (N.GT.NMAX) THEN
*        Lots of d.o.f (>300 in this case). I'm not interested. (MS)
         S = R1/N
         T = F1*S
         W = ((Y*S)**TH - (1-T))
                / SQRT(2*T)
         IF (W.LT.-XLIM) THEN
            H = 1
         ELSEIF (W.LT.XLIM) THEN
            H = HF*GERFC(W)
         ELSE
            H = 0
         ENDIF

      ELSE
*        Most Grover-relevant cases fall here (>1 dof, chi2/dof ~1) (MS)
         M = N/2
         IF (U.LT.XMAX2.AND.(Y/N).LE.CHIPDF) THEN
*           reasonably low chi2 and chi2/dof (for Grover) (MS)
            S = EXP(-HF*U)
            T = S
            E = S
            IF (2*M.EQ.N) THEN
*              always true, surely? I'm confused...(MS)
               FI = 0
               DO I = 1, M-1
                  FI = FI + 1
                  T  = U*T/FI
                  S  = S + T
               ENDDO
               H = S*E

            ELSE
*              always false?  :-( (MS)
               FI = 1
               DO I=1, M-1
                  FI = FI + 2
                  T = T*Y/FI
                  S = S + T
               ENDDO
               W = SQRT(U)
               IF (W.LT.XLIM) THEN
                  H = C1*W*S*E + GERFC(W)
               ELSE
                  H = 0.
               ENDIF
            ENDIF
         ELSE
            H = 0
         ENDIF
      ENDIF


      IF (H.GT.EPS) THEN
         PROB = H
      ELSE
         PROB = 0.
      ENDIF


      RETURN
      END


 */
