//**********************************************************************/
// Created: 23.08.97,   Olya Igonkina   for   MARPLE package
//
// implementation of RcMsOfCesl class
//
//**********************************************************************/
#include <iostream.h>
#include <iomanip.h>
#include <math.h>

#include "marple/marple.hh"
#include <CLHEP/config/CLHEP.h>

#include "arte/ArteMessage.hh"

extern "C"{
 extern void FORTRAN_ROUTINE( rcmesm )(int *, double *, double *);
}

RcMsOfGesl::RcMsOfGesl(const ArtePointer <GESL> gesl,
		       HepDouble ms,
		       HepDouble eLoss,
		       HepDouble zcenter,
		       int insideMagnet )
  : fgesl(gesl), fms(ms) , feLoss(eLoss),
    fz_center(zcenter), finsideMagnet(insideMagnet)
{
  int indx = gesl - *ArteTable<GESL>::begin() +1 ;
  if( indx < 0 || indx > (int)ArteTable<GESL>::size() )
    {
      arteMsg << "MARPLE: RcMsOfGesl: superlayer " << indx 
	   << " is unknown";
      arteMsg.allp();
      return;
    }
  if( fabs(ms)<0.1e-5 )
      {
	double dms, eLoss;
	FORTRAN_ROUTINE( rcmesm)( &indx, &dms, &eLoss);
	setms( dms);
	seteLoss(eLoss);
      }
    if( fabs(zcenter)<0.1e-5  ) setzcenter( ( gesl->z1 + gesl->z2)/2 );  
    if( insideMagnet < 0  )   setMagnet( gesl->mfl);   
}


// check if the point rp is inside of superlayer
int RcMsOfGesl::check( Hep3Vector &rp ) const
{
  if( rp.z() < (fgesl->z1+MARPLE_NEGLIGIBLE) ||
      rp.z() > (fgesl->z2-MARPLE_NEGLIGIBLE) ) 
    { return 0; }

  int i = 0;
  double dz, a1, a2, b1, b2;
  switch ( fgesl->shp )
    {
    case 1 :
      // superlayer is a box 
      if( rp.x() > fgesl->x1 && rp.x() < fgesl->x2 &&
      	  rp.y() > fgesl->y1 && rp.y() < fgesl->y2 )
      	{
      	  i = 1;
      	}
      break;
    case 2 :
      //  superlayer is a X-trapezoid (ie. x  along z)
      dz = ( rp.z() - fgesl->z1 ) / ( fgesl->z2 - fgesl->z1 );
      a1 = fgesl->x1 + fgesl->dx * dz;
      a2 = fgesl->x2 - fgesl->dx * dz;
      if( rp.x() > a1 && rp.x() < a2 &&
      	  rp.y() > fgesl->y1 && rp.y() < fgesl->y2)
      	{
      	  i = 1;
      	}
      break;
    case 3 :
      //  superlayer is a trapezoid (ie. x and y varying along z)
      dz = ( rp.z() - fgesl->z1 ) / ( fgesl->z2 - fgesl->z1 );
      a1 = fgesl->x1 + fgesl->dx * dz;
      a2 = fgesl->x2 - fgesl->dx * dz;
      b1 = fgesl->y1 + fgesl->dy * dz;
      b2 = fgesl->y2 - fgesl->dy * dz;
      if( rp.x() > a1 && rp.x() < a2 &&
      	  rp.y() > b1 && rp.y() < b2)
      	{
      	  i = 1;
      	}
      break;
    case 5 :
      //  superlayer is a tube 
      a1 = fgesl->x2 - fgesl->x1;
      a2 = fgesl->x2 + fgesl->x1;
      b1 = fgesl->y2 - fgesl->y1;
      b2 = fgesl->y2 + fgesl->y1;
      if( rp.x() > a1 && rp.x() < a2 &&
      	  rp.y() > b1 && rp.y() < b2)
      	{
      	  i = 1;
      	}
      break;
    case 9 :
      // superlayer is spherical shell
      a1 = fgesl->x2 - fgesl->x1;
      a2 = fgesl->x2 + fgesl->x1;
      b1 = fgesl->y2 - fgesl->y1;
      b2 = fgesl->y2 + fgesl->y1;
      if( rp.x() > a1 && rp.x() < a2 &&
      	  rp.y() > b1 && rp.y() < b2)
      	{
      	  i = 1;
      	}
      break;
    default :
      arteMsg << "MARPLE: the unknown shape of GESL" << fgesl;
      arteMsg.allp();
      break;
    }
  return i;
}



// printout average Multiple scattering for GESL table

ostream & operator << (ostream & os, const RcMsOfGesl &rms)
{
  os <<  " Multiple Scattering for GESL: "<< setw(2) << rms.gesl();
  os << " MS = " << setw(11)<< rms.ms() ;
  os << " eLOss = " << setw(11) << rms.eLoss();
  os << " Z_GESL = " << setw(6) << rms.zcenter() <<endl;
  return os ;
}

