// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 
// - introduction of contructor from rseg
// - moved many routines from other state classes to here
// - introduced accessors to covmatrix elements
//                                     14-Aug-00 Wouter Hulsbergen

// class rangerState
// base class for trajectory state

#ifndef rangerState_HH
#define rangerState_HH

#ifndef _SYMMatrix_H_
#include "CLHEP/Matrix/SymMatrix.h"
#endif
#ifndef _Vector_H_
#include "CLHEP/Matrix/Vector.h"
#endif
#ifndef __ARTE_HH__
#include "arte/arte.hh"
#endif
#ifndef RANGER_FMODE_HH
#include "rang/FitMode.hh"
#endif

class rangerHitBase;
class rangerStateP;
class rangerStatePI;
class rangerStateL;
class rangerStateLI;
class rangerRadEl;
class rangerScatThin;
class rangerScatThick;

class rangerState {
  // constructors, destructor, assignment
public:
  rangerState(int dim=0) ;
  rangerState(const rangerState& s) ;  // copy constructor
  virtual ~rangerState();
  virtual rangerState& operator=(const rangerState& s) ;  // assignment

  static rangerState* makeState( const rangerFitMode fm); // create empty state
  static rangerState* makeState( const ArtePointer<RHIT>& rhit,const rangerFitMode fm); 
  static rangerState* makeState( const ArtePointer<HITB>& hitb,const rangerFitMode fm); 
  static rangerState* makeState( const ArtePointer<RCCL>& rccl,int charge,const rangerFitMode fm); 
  static rangerState* makeState( const ArtePointer<RSEG>& rseg,const rangerFitMode fm, 
				 unsigned char side); 
  
  // setters and getters
  const double& z() const { return zState; }
  const double& qPEst() const { return qPEstimate; }
  void setZ( double z ) { zState =  z ; }      // update accessors
  void setQpEst( double qp ) { qPEstimate = qp; }
  
  HepVector& vec() { return p ; }
  const HepVector& vec() const { return p ; } 
  HepSymMatrix& cov() { return c ; }
  const HepSymMatrix& cov() const { return c ; }
  int dim() const { return p.num_row() ; } // dimension of state vector

  void copyVec( const rangerState& s) ; // copy state vector from another state (do not copy cov matrix)
  void linExtrapVec( const rangerState& s, float z) ; // extrapolate state vector from another state
  void dump() ;               // dump private members to screen

  // implemented by child class
  virtual rangerState* clone() const=0; // produce copy
  virtual rangerFitMode fitMode() const=0;

  // these are the parameter update and lookup accessors. in principle
  // they could all be purely virtual. that would certainly be most
  // beautiful. however, that means a lot of extra function
  // calls. therefore we use the fact that the first four parameter
  // are the same for all derived states. so this has to be changed if
  // we get classes that want to overrule the defaults!

  const double x()  const { return p(1); }
  const double y()  const { return p(2); }
  const double tx() const { return p(3); }
  const double ty() const { return p(4); }
  virtual const double qP() const = 0;
  virtual const double I()  const = 0;

  void setX(double x)   { p(1) = x ; }
  void setY(double y)   { p(2) = y ; }
  void setTx(double tx) { p(3) = tx ; }
  void setTy(double ty) { p(4) = ty ; }
  virtual void setQp(double qp) = 0 ;
  virtual void setI(double i)   = 0 ;

protected:

  const double covXX()   const { return cov().fast(1,1) ; }
  const double covXY()   const { return cov().fast(2,1) ; }
  const double covXTx()  const { return cov().fast(3,1); }
  const double covXTy()  const { return cov().fast(4,1) ; }
  const double covYY()   const { return cov().fast(2,2) ; }
  const double covYTx()  const { return cov().fast(3,2) ; }
  const double covYTy()  const { return cov().fast(4,2) ; }
  const double covTxTx() const { return cov().fast(3,3) ; }
  const double covTxTy() const { return cov().fast(4,3) ; }
  const double covTyTy() const { return cov().fast(4,4) ; }
  virtual const double covXQp()  const = 0;
  virtual const double covXI()   const = 0;
  virtual const double covYQp()  const = 0;
  virtual const double covYI()   const = 0;
  virtual const double covTxQp() const = 0;
  virtual const double covTxI()  const = 0;
  virtual const double covTyQp() const = 0;
  virtual const double covTyI()  const = 0;
  virtual const double covQpQp() const = 0;
  virtual const double covQpI()  const = 0;
  virtual const double covII()   const = 0;

  void setCovXX(const double v)   { cov().fast(1,1) = v ; }
  void setCovXY(const double v)   { cov().fast(2,1) = v ; }
  void setCovXTx(const double v)  { cov().fast(3,1) = v ; }
  void setCovXTy(const double v)  { cov().fast(4,1) = v ; }
  void setCovYY(const double v)   { cov().fast(2,2) = v ; }
  void setCovYTx(const double v)  { cov().fast(3,2) = v ; }
  void setCovYTy(const double v)  { cov().fast(4,2) = v ; }
  void setCovTxTx(const double v) { cov().fast(3,3) = v ; }
  void setCovTxTy(const double v) { cov().fast(4,3) = v ; }
  void setCovTyTy(const double v) { cov().fast(4,4) = v ; }
  virtual void setCovXQp(const double v) = 0;
  virtual void setCovXI(const double v)  = 0;
  virtual void setCovYQp(const double v) = 0;
  virtual void setCovYI(const double v)  = 0;
  virtual void setCovTxQp(const double v)= 0;
  virtual void setCovTxI(const double v) = 0;
  virtual void setCovTyQp(const double v)= 0;
  virtual void setCovTyI(const double v) = 0;
  virtual void setCovQpQp(const double v)= 0;
  virtual void setCovQpI(const double v) = 0;
  virtual void setCovII(const double v)  = 0;

public:

  void covInit() ;
  virtual int predict(const double& zDestination,
		      const double& qPExp,rangerState*& targetState,
		      HepMatrix*& FTransport) const=0; // perform Kalman
  // filter prediction from this state to the target state
  virtual void project(HepVector& resid,HepMatrix& HH,rangerHitBase* hit)=0;
  // project into measurement space of hit 
  
  void applyScat(const rangerScatThin& scat) ; // apply m.s. correction 
  void applyScat(const rangerScatThick& scat) ;// apply m.s. correction 
  
  void applyRadEl(const rangerRadEl& ) ; // apply radiative energy loss correction 
  void applyDeDx(const float ) ;         // apply continuous energy loss correction based on ms()
  void applyDeDxe(const float ) ;         // apply continuous energy loss correction based on dedx()
  void applyRadPhoton(const float ,const float ) ; // apply resolved radiated photon
  
  void storeRseg(ArtePointer<RSEG> rseg,int firstLast); // store track parameters into RSEG
  void storeRtra(ArtePointer<RTRA> rseg,int firstLast); // store track parameters into RSEG
  void storeRhit(ArtePointer<RHIT> rhit); // store parameters as RHIT
  ArtePointer<RPNT> storeRpnt(ArtePointer<RPNT> ); // store parameters as RPNT

  virtual void mergeRhit(ArtePointer<RHIT> rhit,float weight); // merge parameters to RHIT

public: // Rainer had these protected
  void initRHIT(const ArtePointer<RHIT>& rhit) ;
  void initHITB(const ArtePointer<HITB>& hitb) ;
  void initRCCL(const ArtePointer<RCCL>& rccl, int charge) ;
  void initRSEG(const ArtePointer<RSEG>& rseg, unsigned char side) ;

protected:

  double zState;       // z of reference
  double qPEstimate;   // estimate for momentum parameter Q/p
  HepVector p ;        // parameter vector
  HepSymMatrix c ;     // covariance matrix
} ;

#endif
