// ********************************************************************
//
//  Routine:      rcSelectJPSIEE
//  Author:       T.Oest
//
//  Description:  search for J/psi candidates
//
// symbol:    rcSelectJPSIEE
//
// called by: rcSelectDecays
//
// purpose:   J/Psi -> e+ e- AND detached vertex
//
// hints:     selection results are stored in appropriate tables (RTRA, RVER)
//            cuts might be set in classification if not defaults are used
//            classification bits are set using entries in the tables
//            
//  Modification:
//  -- faster J/psi selection by changing order of cuts  
//                                                      01/12/99 T.Oest
//  -- changed the cuts in the J/psi selection, 
//     now each lepton must be identified either by
//     Bremsstrahlung, RICH or OTR                      05/06/00 D.Samtleben
//
//  -- take care of all ECAL clusters assigned to an RTRA entry
//  -- use VDS-ECAL momentum
//                                                      12/07/00 D.Samtleben
//  -- polished code so that even I can understand it   17/07/00 AG 
//
//  -- check the opposite charge of the tracks
//     from VDS-ECAL momentum and not from RTRA momentum
//                                                      21/07/00 D.Samtleben
//  -- fill RTRA-TOPP relation for Jpsi
//                                                      21/07/00 D.Dujmic
//  -- Make an RTRA entry also for Upsilon selection
//       if invariant mass >6.0 GeV 
//       no identification/confirmation is required
//     Mass without added br is stored in the "leh" field 
//                                                      29/11/00 M.Villa
//  -- changed name to rcSelectJPSIEE                   08/12/00 AG
//  -- Changed vertex finding routine: JPSIvertex.
//     Now identified electrons (loose cuts) are passed to the JPSIvertex
//     routine to search for detached VX.
//     Cut on E_t of cluster reduced to 0.9 GeV/c       12/12/00 M.Villa
//  -- added some debug printout                        15/12/00 AG
//  -- Mods to evaluate identified dielectrons          03/07/01 M.Villa
//  -- Code cleaned by using cmp & clone fields         24/07/01 M.Villa
//  -- removed include reco/rc99Areco.hh                27/09/01 D.Samtleben
// ********************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <math.h>

// ARTE includes
#include "arte/arte.hh"
#include "farm/farm.hh"
#include "arte/RSEG_cmp.hh"
#include "arte/ArteUtil.h"
#include "reco/rcXkinema.h"
#include "reco/JPSIvertex.hh"
#include "analysis/anl_richid.h"

// externals
extern int FADEB; // debug flag steered by runargs=-debug

// globals
float cutJPSIEE     = 2.0; // default di-lepton low mass cut
float cutRECYPS     = 6.0; // default di-lepton high mass cut
float cutRECDVlower = 0.5; // default lower detached vertex cut
float cutRECDVupper = 3.0; // default higher detached vertex cut

// prototypes
float rc99Bremsstrahlung(ArtePointer<RTRA>);
int
JPSIvertex(ArtePointer<RTRA> track1,ArtePointer<RTRA> track2,
	   JPSIVtxType t);

//***************************************************************
// function
//*****************************************************************

int rcSelectJPSIEE()
{
  ARTE_mark( "rcSelectJPSIEE 0", 0 );

  // print
  if( FADEB ){
    cout << "JPSIEE: "
	 << "cutRECYPS="     << cutRECYPS     << " "
	 << "cutJPSIEE="     << cutJPSIEE     << " "
	 << "cutRECDVlower=" << cutRECDVlower << " "
	 << "cutRECDVupper=" << cutRECDVupper << endl;
  }
 
  //==================================================================
  // 1. Cuts
  //==================================================================
  float dyCut         = 3.;
  float pCut          = 5.;
  float ptCut         = 0.0;
  float EtCut         = 0.9;
  float EoverPCut     = 0.36;
  float MassCut       = cutJPSIEE;
  float MassCutUps    = cutRECYPS; 
                        // default to 6.0 (can appear in a kuip cmd)  

  //==================================================================
  // 
  //==================================================================
  float px1, py1, pz1, p1, px2, py2, pz2, p2, pt1, pt2, mom1, mom2;
  float Et1, Et2, Ee1, Ee2, Eb1, Eb2;
  float px, py, pz, p, mee, meecorr, pt;
  float dy1, dy2;

  // RTRA loop 1
  for( ArteTable<RTRA>::iterator rtra1 = ArteTable<RTRA>::begin();
       rtra1 != ArteTable<RTRA>::end(); ++rtra1 ) {
    
    bool OtrOrRich1 = (((*rtra1)->cmp) & (Rsegc::rich | Rsegc::patt))!=0;
      
    // RCCL loop 1 (many RCCL per RTRA)
    for (ArteRelationIterator<RCCL> rccl1=(*rtra1)->rccl.begin();
	 rccl1 != (*rtra1)->rccl.end(); ++rccl1){       
      
      mom1=1./(*rtra1)->pf;
      p1=fabs(mom1);
      // cout << "yes, found it ..1 " << (*rtra1) << " " << mom1 << endl;
      
      if (p1 > pCut && fabs((*rccl1)->esum/fabs(p1)-1.)< EoverPCut){
	// cout << "found a track ..1 T" << *rtra1 << " Cl" << (*rccl1)<<
	// " energy:"<<(*rccl1)->esum<< endl;
	
	pz1 = p1 / sqrt(1.+(*rtra1)->txf*(*rtra1)->txf +
                        (*rtra1)->tyf*(*rtra1)->tyf);
	px1 = (*rtra1)->txf * pz1;
	py1 = (*rtra1)->tyf * pz1;
	pt1 = sqrt(px1*px1+py1*py1);
	Et1 = (*rccl1)->esum * 
	  sqrt((*rccl1)->x*(*rccl1)->x+(*rccl1)->y*(*rccl1)->y)/(*rccl1)->z;
	dy1 = (*rccl1)->y -
	  ((*rtra1)->yf + (*rtra1)->tyf *((*rccl1)->z - (*rtra1)->zf));
	
	// cout<<" 1: pt"<<pt1<<" et:"<<Et1 <<" dy:"<<dy1<<endl;
	//
	// particle 1
	//
	if (pt1 > ptCut && Et1 > EtCut && fabs(dy1) < dyCut){
	  
	  // RTRA loop 2
	  ArteTable<RTRA>::iterator rtra2 = rtra1;
	  for( ++rtra2;
	       rtra2 != ArteTable<RTRA>::end(); ++rtra2 ) {
	    bool OtrOrRich2 = 
	      (((*rtra2)->cmp) & (Rsegc::rich | Rsegc::patt))!=0;	    
	    
	    // RCCL loop 2
	    for(ArteRelationIterator<RCCL> rccl2=(*rtra2)->rccl.begin();
		rccl2!= (*rtra2)->rccl.end();rccl2++){       

	      mom2  = 1./(*rtra2)->pf;
	      p2=fabs(mom2);
	      //   cout << "found it ..2 T" <<(*rtra2),<" p:"<<mom2<< endl; 
	      //
	      // particle 2
	      //
	      if (mom1 * mom2 < 0 
		  && p2 > pCut && 
		  fabs((*rccl2)->esum/fabs(p2)-1.) < EoverPCut){
		
		//cout << "here we go ...2 " << (*rtra2) << " CL"<<(*rccl2) 
		//     <<" energy:"<<(*rccl2)->esum << endl;
		
		if ((*rccl1) != (*rccl2) && 
		    ((*rtra1)->csvd>0 && (*rtra1)->csvd!=(*rtra2)->csvd)  ){
		  
		  pz2 = p2 / sqrt(1.+(*rtra2)->txf*(*rtra2)->txf +
				  (*rtra2)->tyf*(*rtra2)->tyf);
		  px2 = (*rtra2)->txf * pz2;
		  py2 = (*rtra2)->tyf * pz2;
		  pt2 = sqrt(px2*px2+py2*py2);
		  Et2 = (*rccl2)->esum * 
		    sqrt((*rccl2)->x*(*rccl2)->x
			 +(*rccl2)->y*(*rccl2)->y)/(*rccl2)->z;
		  dy2 = (*rccl2)->y -
		    ((*rtra2)->yf 
		     + (*rtra2)->tyf *((*rccl2)->z - (*rtra2)->zf));
		  
		  // cout<<" 2: pt"<<pt2<<" et:"<<Et2 <<" dy:"<<dy2<<endl;
		  //
		  // J/Psi candidate
		  // 
		  if (pt2 > ptCut && Et2 > EtCut && fabs(dy2) < dyCut){
		    
		    px  = px1 + px2;
		    py  = py1 + py2;
		    pz  = pz1 + pz2;
		    p   = sqrt(px*px+py*py+pz*pz);
		    pt  = sqrt(px*px+py*py);
		    
		    //--------------------------------------------------------
		    //   electron electron mass
		    //--------------------------------------------------------
		    float corr1, corr2;
		    
		    Ee1    = p1;
		    Ee2    = p2;
		    Eb1    = rc99Bremsstrahlung((*rtra1));
		    Eb2    = rc99Bremsstrahlung((*rtra2));
		    
		    mee    = sqrt( 2.*(Ee1*Ee2 - px1*px2 - py1*py2 - pz1*pz2));
		    
		    if (Ee1 != 0.0 && Ee2 != 0.0) {
		      corr1     = 1. + Eb1 / Ee1;
		      corr2     = 1. + Eb2 / Ee2;
		      meecorr   = mee * sqrt(corr1 * corr2);
		    } else {
		      meecorr   = 0.0;
		    }

		    //--------------------------------------------------------
		    //   mass cut
		    //--------------------------------------------------------

		    if (meecorr > MassCutUps ||
 			  (meecorr > MassCut &&
			    (Eb1>0||OtrOrRich1)
			   &&(Eb2>0||OtrOrRich2)) ){

		      //------------------------------------------------------
		      //   vertex cut
		      //------------------------------------------------------
		      int rc = 
			JPSIvertex(*rtra1 ,*rtra2, ELECTRONS_ALWAYS);
		      // cout<<" *** JPSIvertex ELECTRONS_ALWAYS "<<rc<<endl;

		      if( rc !=0 ){

			ArtePointer<RTRA> rtra = ArteTable<RTRA>::size();
			//----------------------------------------------------
			//   fill RTRA and vertex bank
			//----------------------------------------------------
			
			rtra->lee = meecorr;
			rtra->leh = mee;
			rtra->cmp = (*rtra1)->cmp | (*rtra2)->cmp;
		        rtra->lemi = 0.0;
		        if( Eb1+Eb2>0. && 
				(Eb1>0. || anl_isaelectron(*rtra1)==2 ) &&
				(Eb2>0. || anl_isaelectron(*rtra2)==2 ) ){
			   rtra->lemi = meecorr;
		        }
                      }
		    } else {

		      int rc = 
			JPSIvertex(*rtra1 ,*rtra2, ELECTRONS_DETACHED);
		      //cout<<" *** JPSIvertex ELECTRONS_DETACHED "<<rc<<endl;

		      if( rc>0 ){
		        ArtePointer<RTRA> rtra = ArteTable<RTRA>::size();
			rtra->lee = meecorr;
			rtra->leh = mee;
			rtra->cmp = (*rtra1)->cmp | (*rtra2)->cmp;
		        rtra->lemi = 0.0;
		        if( Eb1+Eb2>0. && 
				(Eb1>0. || anl_isaelectron(*rtra1)==2 ) &&
				(Eb2>0. || anl_isaelectron(*rtra2)==2 ) ){
			   rtra->lemi = meecorr;
		        }

                      }
                    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  ARTE_mark( "rcSelectJPSIEE 1", 0 );
  return 0; 
}
