// -*-C++-*-
#ifndef _RICHEVENTMC_H_
#define _RICHEVENTMC_H_

#include <map.h>
#include <multimap.h>
#include <math.h>
#include "arte/arte.hh"
#include "rich/ricutil.h"
#include "rich/PhiLambda.h"
#include "CLHEP/Hist/HBookTuple.h"


/////////////////////////  RICH photons ///////////////////////////////

/** Access to MC photon information. 

Example 1: 
for (begin(); valid(); next()) 
will loop over all radiated photons (AFTER MIMP reduction has been applied).

Example 2: 
for (begin(10); valid(); next()) 
will loop over all radiated photons for MC particle 10 (MTRA=10). Function
detected() returns true if the phtons was detected.
*/

class RICHphotons {
 private:
  // mtra=key,  mimp=line in MIMP table
  multimap <unsigned const, ArtePointer<MIMP>, less<unsigned> > photonList;
  multimap <unsigned const, ArtePointer<MIMP>, less<unsigned> >::iterator it;
  multimap <unsigned const, ArtePointer<MIMP>, less<unsigned> >::iterator itl;

 public:
  RICHphotons();

  /// for (begin(); valid(); next()) loops over all photons
  inline void   begin() { 
    it =photonList.begin();
    itl=photonList.end();
  }
  /// for (begin(10); valid(); next()) loops over photons of MC particle imtra
  inline void   begin(unsigned imtra) {
    it =photonList.lower_bound(imtra);
    itl=photonList.upper_bound(imtra);
  }
  /// See begin
  inline void   next()  { ++it;           }
  /// See begin
  inline bool   valid() { return it!=itl; }
  /// Checks if end of photon list
  inline bool   end()   { return it==photonList.end(); }
  /// Checks if MC particle (MTRA=imtra) has photons
  inline bool   find(unsigned imtra) {
    if ((*it).first!=imtra) {
      it=photonList.find(imtra);
      return it!=photonList.end();
    }
    return 1;
  }
  /// Returns the number of photons
  inline int size() { return photonList.size(); }

  /// Returns MTRA number
  inline unsigned          mtra() { return (*it).first; }
  /// Returns MIMP pointer
  inline ArtePointer<MIMP> mimp() { return (*it).second; }
  /// Returns channel id if photon detected, -1 otherwise
  inline int id() { 
    int id=-1;
    for (ArteRelationIterator<DIGB>::iterator idigb=mimp()->digb.begin();
	 idigb!=mimp()->digb.end(); ++idigb) id=(*idigb)->id;
    return id;
  }
  /// Checks if photon was detected
  inline bool detected() { return id() > -1; }
  /// Returns phi coordinate of detected photon
  inline float DetectedPhi() { 
    ArteRelationIterator<HITB>::iterator ihitb=mimp()->hitb.begin();
    return (*ihitb)->q;
  } 
  /// Returns HITB pointer
  inline ArtePointer<HITB> hitb() { return (*mimp()->hitb.begin()); }
  /// Returns lambda coordinate of detected photon
  inline float DetectedLambda() { 
    ArteRelationIterator<HITB>::iterator ihitb=mimp()->hitb.begin();
    return (*ihitb)->r;
  }  
  /// Get total number of 'radiated' photons (MIMPS)
  inline int NumMIMP() { return photonList.size(); }
  /// Prints all photons
  void print();
  /// Prints photons for MC particle imtra
  void print(unsigned imtra);
};




// //////////////////////////// MC RICH rings ////////////////////////////////

struct RichRingData {
  Vec3  x;         // point in RICH
  float nrefr;     // refraction index
  float tx,ty;     // direction
  float radius;    // radius
  int   hitbs;     // detected photons
  int   mimps;     // MC radiated photons (after reduction)
  ArtePointer<TOPP> topp;   // type
  ArtePointer<TOPP> mother; // mother type
  ArtePointer<MVER> vertex; // vertex
};

/** MC information on RICH rings */
class RICHrings {
 private:
  // mtra=key
  map <unsigned const, RichRingData, less<unsigned> > ringList;
  map <unsigned const, RichRingData, less<unsigned> >::iterator it;
  map <unsigned const, RichRingData, less<unsigned> >::iterator itl;
  RICHphotons  *photons;
  RichRingData data;

 public:
  /// Constructor
  RICHrings();
  /// Destructor
  inline ~RICHrings() { delete photons; }
  /// for (begin(); valid(); next()) loops over all RICH rings
  inline void   begin() { 
    it =ringList.begin();
    itl=ringList.end();
  }
  /// See begin
  inline void next()  { ++it;           }
  /// See begin
  inline bool valid() { return it!=itl; }
  /// Checks if MC particle (MTRA=imtra) has photons
  inline bool find(unsigned imtra) { 
    if ((*it).first!=imtra) {
      it=ringList.find(imtra);
      return it!=ringList.end();
    }
    return 1;
  }
  /// Returns number of rings
  inline int size() { return ringList.size(); }

  /// Print all rings
  void print();
  /// Print all photons
  void printPhotons() { photons->print(); }
  void fill(HBookTuple *tuple);
  void fillRSEG();

  /// Returns MTRA pointer
  inline ArtePointer<MTRA> mtra() { return ArtePointer<MTRA>((*it).first); }
  /// Returns average refraction index
  inline float RefractionIndex()  { return (*it).second.nrefr; }
  /// Returns a point in RICH vessel
  inline Vec3  x()       { return (*it).second.x; }
  /// Returns tx=px/pz of track
  inline float tx()      { return (*it).second.tx; }
  /// Returns ty=py/pz of track
  inline float ty()      { return (*it).second.ty; }
  /// Returns phi angle of track
  inline float phi()     { return RICH_LambdaPhi->phi   (tx(),ty()); }
  /// Returns lambda angle of track
  inline float lambda()  { return RICH_LambdaPhi->lambda(tx(),ty()); }
  /// Returns radius of track, as average radiation angle
  inline float radius()  { return (*it).second.radius; }
  /// Returns number of detected photons
  inline int   DetectedPhotons()  { return (*it).second.hitbs; }
  /// Returns number of radiated photons
  inline int   RadiatedPhotons()  { return (*it).second.mimps; }
  /// for (firstPhoton(imtra); validPhoton(); nextPhoton()) loops over all radiated photons for
  /// MC particle (MTRA=imtra)
  inline void  firstPhoton(unsigned imtra) { photons->begin(imtra); }
  /// See firstPhoton
  inline bool  validPhoton() { return photons->valid(); }
  /// See firstPhoton
  inline void  nextPhoton()  { photons->next(); }
  /// Returns momentum
  float  momentum();
  /// Returns particle pointer in TOPP
  inline ArtePointer<TOPP> topp()   { return (*it).second.topp; }
  /// Returns mother pointer in TOPP
  inline ArtePointer<TOPP> mother() { return (*it).second.mother; }
  /// Returns vertex pointer in RVER
  inline ArtePointer<MVER> vertex() { return (*it).second.vertex; }
  /// Access to all photons
  inline RICHphotons *GetPhotons()   { return photons; } 
  /// Returns impact parameter vector
  inline Vec3 T() {
    Vec3   x0=x();
    Vec3   n0=normalize(Vec3(tx(),ty(),1));
    Vec3   x1=x0-n0*x0.z()/n0.z();
    float  delta=norm(cross(x1,x0))/norm(x1-x0);
    float  t=sqrt(norm(x0)*norm(x0)-delta*delta);
    return x0-t*n0;
  }
};




// ////////////////////// RC RICH rings ///////////////////////////////


struct RichRecoData {
  float tx,ty;     // direction
  float radius;    // radius
  float photons;   // reconstructed photons
  float background;// background photons
  float f;         // f number
  ArtePointer<RSEG> rseg; // rseg pointer
};

struct RecoKey {
  unsigned mtra;
  float dtx,dty;   // direction precision
  float dradius;   // radius precision
  float dphotons;  // photon precision
};

struct RichRule {
  bool operator()(const RecoKey k1,const RecoKey k2) const {
    if (k1.mtra==k2.mtra) {
      return (k1.dtx*k1.dtx+k1.dty*k1.dty <
              k2.dtx*k2.dtx+k2.dty*k2.dty);
    }
    else return (k1.mtra<k2.mtra);
  }
};

/** RingSearch information on RICH rings - only stand alone rings found by RISE */
class RICHreco {
 private:
  // mtra=key
  multimap < RecoKey const, RichRecoData, RichRule > recoList;
  multimap < RecoKey const, RichRecoData, RichRule >::iterator it;
  multimap < RecoKey const, RichRecoData, RichRule >::iterator itl;
  RichRecoData data;
  RecoKey      key;
  RICHrings    *mcRings;

 public:
  /// Constructor
  RICHreco();
  /// Destructor
  inline ~RICHreco() { delete mcRings; }
  /// for (begin(); valid(); next()) loops over all reconstructed rings
  inline void   begin() { 
    it =recoList.begin();
    itl=recoList.end();
  }
  /// See begin
  inline void   next()  { ++it;           }
  /// See begin
  inline bool   valid() { return it!=itl; }
  /// Check if MC particle has reconstructed ring.
  /// In principle several reconstructed rings can be
  /// assigned to a single MC particle, to get them all use:
  /// for (find(imtra); valid(); next()) will loop over all reco rings with MC number imtra
  bool find(unsigned imtra);


  /// Print all rings
  void print();
  void fill(HBookTuple *tuple);

  /// Returns MTRA pointer
  inline ArtePointer<MTRA> mtra() { return ArtePointer<MTRA>((*it).first.mtra); }
  /// Returns RSEG pointer
  inline ArtePointer<RSEG> rseg() { return ArtePointer<RSEG>((*it).second.rseg); }
  /// Returns reconstructed tx=px/pz
  inline float tx()      { return (*it).second.tx; }
  /// Returns (reconstructed tx)-(real tx)
  inline float dtx()     { return (*it).first.dtx; }
  /// Returns reconstructed ty=py/pz  
  inline float ty()      { return (*it).second.ty; }
  /// Returns (reconstructed ty)-(real ty)
  inline float dty()     { return (*it).first.dty; }
  /// Returns reconstructed phi angle
  inline float phi()     { return atan(tx()); }
  /// Returns reconstructed lambda angle
  inline float lambda()  { return atan(ty()*cos(phi())); }
  /// Returns reconstructed radius
  inline float radius()  { return (*it).second.radius; }
  /// Returns (reconstructed radius)-(real radius)
  inline float dradius() { return (*it).first.dradius; }
  /// Returns number of reconstructed photons
  inline float photons() { return (*it).second.photons; }
  /// Returns real number of detected photons
  inline float dphotons(){ return (*it).first.dphotons; }
  /// Returns background photons under the peak
  inline float background() { return (*it).second.background; }
  /// Returns f-number (N_up-N_dw)/(N_up+N_dw)
  inline float  f()      { return (*it).second.f; }
};


#endif











