// *****************************************************************************
//
// source:    SLTFillArteBanks.C
//
// type:      source code
//
// created:   16 Nov 2000
//
// author:    Brian Petersen
//            email: aagaard@nbi.dk
//
// purpose:   Fill RSLA and RSLI banks with SLT information
//
// changes:         
// 16 Nov 2000 (BP) Creation
// 22 Nov 2000 (BP) Also fill vertex information into RSLI
// 27 Nov 2000 (AG) KUIP needs to be initialized here
// 30 Jan 2001 (AG) corrected flag in ARTE_mark to 1 for return
// 30 Jan 2001 (BP) Skip tracks not tracked through full detector
//
// *****************************************************************************

// ARTE includes
#include "arte/arte.hh"
#include "farm/farm.hh"

// includes
#include "sltinfo.h"

// prototypes
extern "C" void sltkuip_();

//
// initialize the Xyz KUIP commands
//
static class SLTIniter
{
public:
  SLTIniter() { atregkuip( sltkuip_, 1 ); } 
} _sltIniter;

//
// function code
//
void
SLTFillArteBanks()
{
  ARTE_mark( "SLTFillArteBanks 0", 0 );

  ArteTable<RSLA>::clear();
  ArteTable<RSLI>::clear();

  if (SLTReadDSLT()!=SLT_OK){
    return;
  }

  SLT_STATUS stat;
  SLT_TRIGGER trig;
  SLT_VERTEX vert;

  // Only fill for good events
  if (SLTStatus(&stat)!=SLT_OK){
    ARTE_mark( "SLTFillArteBanks 1", 0 );
    return;
  }
  if (!(stat.status&SLTEVENT_ACCEPT)){
    ARTE_mark( "SLTFillArteBanks 1", 0 );
    return;
  }

  // Find the accepted trigger pairs
  for(int trignum=0;trignum<stat.ntriggers;trignum++) {
    if (SLTTrigInfo(trignum,&trig)!=SLT_OK) continue;
    if (trig.status==SLTTRIG_REJECT)  continue;
    if (SLTVertexInfo(trignum,&vert)!=SLT_OK) continue;    

    SLT_TRACK info[2];
    SLT_SEEDTRACK seed[2];
    SLT_REFITTRACK refit[2];
    SLT_SILITRACK sili[2];
    SLT_SILITRACKCOV cov[2];
    SLT_REFITTRACKCOV refcov[2];
    int i;

    for(i=0;i<trig.ntracks;i++) {
      int trk=trig.trackindex[i];
      if (SLTTrackInfo(trk,&info[i])       != SLT_OK) break;
      if (info[i].status                   != SLTTRACK_ACCEPT) break;
      if( SLTSeedTrackInfo(trk,&seed[i])   != SLT_OK) break;
      if (SLTSiliTrackInfo(trk,&sili[i])   != SLT_OK) break;
      if (SLTRefitTrackInfo(trk,&refit[i]) != SLT_OK) break;
      if (SLTRefitTrackCovInfo(trk,&refcov[i]) != SLT_OK) break;
      if (SLTSiliTrackCovInfo(trk,&cov[i]) != SLT_OK) break;
    }
    if (i!=trig.ntracks) continue; // We failed to get some information

    // Fill in RSLI information

    RSLI rsli;
    
    rsli.itri=trignum;
    rsli.type=trig.type;
    rsli.flag=1; // Currently we just set it to "ACCEPT"
    rsli.ntra=trig.ntracks; 
    rsli.tra[0]=rsli.tra[1]=0;

    // The vertex information is currently not available from the SLT
    rsli.chi= vert.chisq;
    rsli.pos[0]=vert.xv;
    rsli.pos[1]=vert.yv;
    rsli.pos[2]=vert.zv;
    rsli.err[0]=vert.sigxv;
    rsli.err[1]=vert.sigyv;
    rsli.err[2]=vert.sigzv;
    rsli.dxy=vert.dxy;
    rsli.dz=vert.dz;

    // Fill in RSLA information for tracks which are not in already

    for(i=0;i<trig.ntracks;i++) {
      int trk=trig.trackindex[i];
      int found=0;
      // Did we already add this track?
      for( ArteTable<RSLA>::iterator rsla1 = ArteTable<RSLA>::begin();
	   rsla1 != ArteTable<RSLA>::end(); ++rsla1 ) {
	if ((*rsla1)->itra==trk+1) found=(*rsla1).index();
      }
      if (found) {
	rsli.tra[i]=found;
      } else {
	RSLA rsla;
	
	rsla.itra=trk+1;
	rsla.flag=1;
	rsla.flt2[0]=seed[i].x2;
	rsla.flt2[1]=seed[i].y2;
	rsla.flt2[2]=seed[i].z2;
	rsla.flt2[3]=seed[i].x2_err;
	rsla.flt2[4]=seed[i].y2_err;
	rsla.flt1[0]=seed[i].x1;
	rsla.flt1[1]=seed[i].y1;
	rsla.flt1[2]=seed[i].z1;
	rsla.flt1[3]=seed[i].x1_err;
	rsla.flt1[4]=seed[i].y1_err;
	rsla.slox=refit[i].tx;
	rsla.intx=refit[i].bx;
	for(int j=0;j<3;j++) 
	  rsla.covx[j]=refcov[i].xcov[j];
	rsla.sloy=refit[i].ty;
	rsla.inty=refit[i].by;
	for(int j=0;j<3;j++) 
	  rsla.covy[j]=refcov[i].ycov[j];


	rsla.nhit=int( refit[i].nhits ); // AG: Why is refit[i].nhits a float?
	rsla.chi=refit[i].chisq;
	rsla.p=sili[i].pxpz;
	rsla.tx=sili[i].tx;
	rsla.bx=sili[i].bx;
	rsla.hitx=sili[i].nhitx;
	rsla.xchi=sili[i].chix;
	for(int j=0;j<3;j++)
	  rsla.xcov[j]=cov[i].xcov[j];
	rsla.ty=sili[i].ty;
	rsla.by=sili[i].by;
	rsla.hity=sili[i].nhity;
	rsla.ychi=sili[i].chiy;
	for(int j=0;j<3;j++) 
	  rsla.ycov[j]=cov[i].ycov[j];
	rsla.id=seed[i].trackID;

	ArteTable<RSLA>::push_back( rsla );
	rsli.tra[i]=ArteTable<RSLA>::size();
      }
    }
    ArteTable<RSLI>::push_back( rsli );
  }

  ARTE_mark( "SLTFillArteBanks 1", 0 );
  return;
}
      
      
