//======================================================
// Function to select run without coasting beam in OTR.
// Should be run after waxevnt() function.
//
// Example of code in usevnt.C:
//   #include "wax/OTrSelect.h"
//   waxevnt();
//   if (OTrSelect(OTR_STRICT_CUT, 900, 0)) {
//      printf("This event has been rejected\n");
//      return;
//   }
//   // Some ranger reconstruction can be done here...
//
//  Possible values of flag:
//     OTR_STRICT_CUT - to make strict cut on mean TDC count and its RMS
//     OTR_LOOSE_CUT - to make more loose cut
//     OTR_RELAXED_CUT - to make even looser cut
//
//  For nhit1 (lowest number of OTR hits) one can take values 300 - 1000.
//  For run 14154 best value is around 900.
//  This will throw away the majority of events without interaction.
//  
//  For nhit2 (maximum allowed number of OTR hits)
//  no default value is recommended.
//  For alignment studies nhit2=10000 can be reasonable.
//  Value 0 will mean no rejection on nhit2.
//  
//  A.Lanyov, W.Hulsbergen            April-May 2000
//
//======================================================

#include <stdio.h>
#include "wax/DecodeManager.hh"
#include "wax/OTrSelect.h"
#include "wax/HITB.hh"
#include "Otr/OtrEvent.h"
#include <stdlib.h>
#include "gean/ZGESL.h"

// returns 0 in case of success, non-zero otherwise
extern "C" 
int OTrEventTime( float& mean, float& width)
{
  int rc=0;
  if(ArteTable<HITB>::booked()) { 
    // Hit preparation switch should match the one in
    // rc99AOTR.C. Very ugly but it works.
    int hitprepflag=0 ;
    char* f ;
    if ((f= getenv("OtrHitPreparation"))) hitprepflag = atoi(f);
    if ( ArteTable<MCIN>::booked() ) hitprepflag = 2;

    double sum(0),sum2(0),time ;
    int n=0;
    for(ArteTable<HITB>::iterator hitbIt = ArteTable<HITB>::begin() ;
	hitbIt != ArteTable<HITB>::end(); ++hitbIt)
      if(hitbIt->gede->cmp==LGESL_WCH && hitbIt->gede->pit<1) {
	switch(hitprepflag) {
	case 1: //wax
	  time = waxHITB(hitbIt).GetDriftTime() ;
	  //time = waxHITB(hitbIt).GetTdcCount() ;
	  break ;
	case 2: //mc
	  time = hitbIt->y ;
	  break ;
	default:
	  time = hitbIt->y - OtrITdcT0(OtrIFromId(hitbIt->chid)) ;
	  //time = hitbIt->y ;
	}
	sum += time;
	sum2 += time*time ;
	n++ ;
      }
    if(n>0) {
      mean = sum/n ;
      width = sqrt((sum2-sum*sum/n)/n) ;
    } else rc=1;
  } else rc=2 ;
  
  if(rc) mean=width=0;
  return rc;
}

extern "C" int OTrSelect(int flag, int nhit1, int nhit2) {
  float tdcmean, rms;
  int nhit;
  ArtePointer<EVRC> evrc(ArteTable<EVRC>::size());
  if (evrc->ottmean !=0. && evrc->ottwidth != 0. && ArtePointer<EVHI>(LGESL_WCH) != 0) {
#ifdef DEBUG
    printf("Using data from EVRC and EVHI tables: ottmean=%f, ottwidth=%f\n",
	   evrc->ottmean, evrc->ottwidth);
#endif
    tdcmean = evrc->ottmean;
    rms = evrc->ottwidth;
    nhit = ArtePointer<EVHI>(LGESL_WCH)->hit;
  }
  else {
    int rcEvent = OTrEventTime(tdcmean, rms);
    if (rcEvent) {
#ifdef DEBUG
      printf("OTrEventTime returned error: rc=%d\n",rcEvent);
#endif
      return 0;
    }

    nhit=0;
    for(ArteTable<HITB>::iterator hitbIt = ArteTable<HITB>::begin();
	hitbIt != ArteTable<HITB>::end(); ++hitbIt) {
      if(hitbIt->gede->cmp==LGESL_WCH) nhit++;
    }
  }

#ifdef DEBUG
  printf("Nhit =%5d, Mean = %6.2f, Rms = %6.2f\n",nhit,tdcmean,rms);
#endif

  int rc = 0;
  if (nhit<nhit1) {
#ifdef DEBUG
    printf("Rejected by nhit1\n");
#endif
    rc |= OTR_NHIT1_CUT;
  }

  if (nhit2 && nhit>nhit2) {
#ifdef DEBUG
    printf("Rejected by nhit2\n");
#endif
    rc |= OTR_NHIT2_CUT;
  }

  if ((flag & OTR_RELAXED_CUT)  && OTr_cut_relaxed (tdcmean, rms)==0) {
#ifdef DEBUG
    printf("Rejected by relaxed cut\n");
#endif
    rc |= OTR_RELAXED_CUT;
  }

  if ((flag & OTR_LOOSE_CUT)  && OTr_cut_loose (tdcmean, rms)==0) {
#ifdef DEBUG
    printf("Rejected by loose cut\n");
#endif
    rc |= OTR_LOOSE_CUT;
  }

  if ((flag & OTR_STRICT_CUT) && OTr_cut_strict(tdcmean, rms)==0) {
#ifdef DEBUG
    printf("Rejected by strict cut\n");
#endif
    rc |= OTR_STRICT_CUT;
  }

#ifdef DEBUG
  if (rc==0) printf("Accepted\n");
  printf("OTrSelect returns %d\n",rc);
#endif
  return rc;
}

/********************************************************************
Selection of events without coasting beam.

These functions to select a good event without coasting beam
return 1 if event is good, and 0 otherwise.
 
OTr_cut_loose defines a loose cut, OTr_cut_strict defines a more strict cut.
Parameters are mean TDC count and its RMS calculated by all 5-mm OTR cells.

In run 14154 (5 MHz) loose cut selects 27% of all events, strict - 20%.
These numbers are so small since these cuts also throw away empty events:
If one looks only events with nhit>300, the percentages will be 46% and 34%,
and if nhit>1000, then - 76% and 58%, respectively.

In addition a cut nhit>300 can be useful.
Here nhit is total number of TDC counts in the range [16-250] in 5-mm
OTR cells.

*********************************************************************/

/**
  Returns 1 if point (mean,rms) lies inside the good region
  and 0 otherwise.

  The region is specified by a number of points.
 */
extern "C" int OTr_cut_loose(float mean, float rms) {

  if (mean<157.) return 0;
  if (mean>180.) return 0;
  if (rms>40.) return 0;
  return 1;

  /*------------- Not used
#define NCUTS 5
  float amean[6] = { 161.5, 171.3, 185.1, 185.2, 168.1, 161.5 };
  float arms[6]  = {  40.5,  39.6,  29.9,  20.0,  20.0,  40.5 };

  float mean0 = 0.;
  float rms0 = 0.;
  for (int i=0; i<NCUTS; i++) {
    mean0 += amean[i];
    rms0 += arms[i];
  }
  mean0 /= NCUTS;
  rms0 /= NCUTS;

  for (int i=0; i<NCUTS; i++) {
    float r  = (mean -amean[i])*(arms[i+1]-arms[i]) - (amean[i+1]-amean[i])*(rms -arms[i]);
    float r0 = (mean0-amean[i])*(arms[i+1]-arms[i]) - (amean[i+1]-amean[i])*(rms0-arms[i]);
    if (r*r0<0) {
      return 0;
    }
  }
  return 1;
  */
}

/**
  Returns 1 if point (mean,rms) lies inside the good region
  and 0 otherwise.

  The region is specified by rectangular borders.
 */
extern "C" int OTr_cut_relaxed(float mean, float rms) {
  if (mean<155.) return 0;
  if (mean>183.) return 0;
  if (rms>50.) return 0;
  return 1;
}

/**
  Returns 1 if point (mean,rms) lies inside the good region
  and 0 otherwise.

  The region is specified by rectangular borders.
 */
extern "C" int OTr_cut_strict(float mean, float rms) {
  if (mean<157.) return 0;
  if (mean>175.) return 0;
  if (rms>35.) return 0;
  return 1;
}
